/*
   This very simple program illustrates the use of message queues.
   The process performs the following actions:

     -- Create a message queue with apropriate rights.
     -- Fork
     -- Send messages to the queue
     -- The child (resulting from the fork extracts the messges from the queue

   Author: denis BEURIVE
*/









#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>





/* ----------------------------------------------------------------- */
/* Definition of the message to send. Please note that the structure */
/* of the message must follow certain rules.                         */
/* ----------------------------------------------------------------- */

#define MSGSIZE   32

struct msg {
             long mtype;
             char message[MSGSIZE];
           };

/* ----------------------------------------------------------------- */
/* Permission given to the message queue                             */
/* ----------------------------------------------------------------- */

#define PERMISSIONS 0X1C0    /* in binary 111000000 => rwx------ */

/* ----------------------------------------------------------------- */
/* Fathers' entry pooint                                             */
/* ----------------------------------------------------------------- */

int main (int argc, char *argv[])
{
  key_t            ipc_key; 
  int              msgid, mesnum, i, cr, size, child_status;
  pid_t            childpid;
  struct msg       buff;
  struct msqid_ds  msg_info;
  

  /* --------------------------------------------------------------- */
  /* Father's code                                                   */
  /* --------------------------------------------------------------- */

  if (argc != 3)
  {
    fprintf (stderr, "\nusage: msg <IPC public key> <number of messages to send>\n");
    return 1;
  }

  ipc_key = (key_t)atoi(argv[1]);
  mesnum  = atoi(argv[2]);

  /* --------------------------------------------------------------- */
  /* Create the message queue.                                       */
  /*   -- The function returns an error if the message queue already */
  /*      exists                                                     */
  /*   -- We set the permissions here.                               */
  /* --------------------------------------------------------------- */

  msgid = msgget(ipc_key, IPC_CREAT | IPC_EXCL | PERMISSIONS);
  if (msgid == -1)
  {
    fprintf (stderr, "\nFather (pid %d), error while creating message queue: %s\n", (int)getpid(), strerror(errno));
    return 1;
  }

  /* --------------------------------------------------------------- */
  /* Read permissions                                                */
  /* --------------------------------------------------------------- */

  cr = msgctl (msgid, IPC_STAT, &msg_info);
  if (cr == -1)
  {
    fprintf (stderr, "\nFather (pid %d), error while getting message queue info: %s\n", (int)getpid(), strerror(errno));
    return 1;
  }

  fprintf (stdout, "\nMessage queue permissions are: %X\n", msg_info.msg_perm.mode); fflush(stdout);


  /* --------------------------------------------------------------- */
  /* Create the child process                                        */
  /* --------------------------------------------------------------- */

  childpid = fork();

  if (childpid != 0)
  {
    /* ------------------------------------------------------------- */
    /* This is the father's code: send mesages to the queue          */
    /* All messages have the same priority (that is 1)               */
    /* ------------------------------------------------------------- */

    buff.mtype = 1;
    size       = sizeof(buff) - sizeof(buff.mtype);

    for (i=0; i<mesnum; i++)
    {
      sprintf (buff.message, "%d", i);

      cr = msgsnd (msgid, (struct msgbuf*)&buff, size, 0);
      if (msgid == -1)
      {
        fprintf (stderr, "\nFather (pid %d), error while sending: %s\n", (int)getpid(), strerror(errno));
        return 1;
      }
    }

    /* ------------------------------------------------------------- */
    /* Send last message (all zeros)                                 */
    /* ------------------------------------------------------------- */

    memset ((void*)buff.message, 0, MSGSIZE);

    cr = msgsnd (msgid, (struct msgbuf*)&buff, size, 0);
    if (msgid == -1)
    {
      fprintf (stderr, "\nFather (pid %d), error while sending las message: %s\n", (int)getpid(), strerror(errno));
      return 1;
    }

    /* ------------------------------------------------------------- */
    /* Wait for the child to terminate                               */
    /* ------------------------------------------------------------- */

    if (wait (&child_status) == (pid_t)-1)
    {
      fprintf (stderr, "\nFather (pid %d), error while waiting for child termination: %s\n", (int)getpid(), strerror(errno));
      return 1;
    }

    fprintf (stdout, "\nFather (pid %d), exit now\n", (int)getpid());

    /* ------------------------------------------------------------- */
    /* Destry the message queue                                      */
    /* ------------------------------------------------------------- */

    if (msgctl (msgid, IPC_RMID, NULL) == -1)
    {
      fprintf (stderr, "\nFather (pid %d), error while deleting the message queue: %s\n", (int)getpid(), strerror(errno));
      return 1;
    }

    return 0;
  }

  /* --------------------------------------------------------------- */
  /* This is the child's code                                        */ 
  /* --------------------------------------------------------------- */

  /* --------------------------------------------------------------- */
  /* Get the message queue                                           */
  /* --------------------------------------------------------------- */

  msgid = msgget(ipc_key, IPC_CREAT);
  if (msgid == -1)
  {
    fprintf (stderr, "\nChild (pid %d), error while getting message queue: %s\n", (int)getpid(), strerror(errno));
    return 1;
  }

  /* --------------------------------------------------------------- */
  /* Read messages from the queue                                    */
  /* --------------------------------------------------------------- */

  buff.mtype = 1;
  size = sizeof(buff) - sizeof(buff.mtype);

  do {
       cr = msgrcv (msgid, (struct msgbuf*)&buff, size, 0, 0);
       if (cr == -1)
       {
         fprintf (stderr, "\nChild (pid %d), error while receiving message: %s\n", (int)getpid(), strerror(errno));
         return 1;
       }

       fprintf (stdout, "\n  => extacted [%s]", buff.message); fflush(stdout);

     } while ( buff.message[0] != 0 );

  fprintf (stdout, "\nChild (pid %d), exit now\n", (int)getpid()); 

  return 0;
}
