/*! \file array.h
    Header file for array.c
 */

#include <stdlib.h>

/*! \brief Data structure that defines an array handler.
 */

struct s_array {
                 /*! \brief Pointer to the first element of the array. */
                 void *first_element;

                 /*! \brief Size of the array (maximum number of elements). */
                 int  array_size;

                 /*! \brief Current number of elements in the array. */
                 int  number_of_elements;

                 /*! \brief User defined function used to copy data from/to the array.
                     \param src Pointer to the data to copy.
                     \param dest Pointer to the destination.
                     \return Upon successful completion the function returns 0, otherwise it returns 1.
                     \warning <UL>
                                 <li><i>data</i> is copied from <i>src</i> to <i>dest</i>.
                                 <li>The function <i>copy_data()</i> <FONT COLOR="#FF0000"><B>MUST take care of memory management</B></FONT>
                                     (if required).
                              </UL>
                  */
                 int  (*copy_data)(void *src, void *dest);

                 /*! \brief User defined function used to remove an element from the array (it may be necessary to free memory for example).
                     \param target Pointer to the element to remove.
                     \return Upon successful completion the function returns 0, otherwise it returns 1.
                  */
                 int  (*remove_data)(void *target);

                 /*! \brief User defined function used to compare two elements of the array.
                     \param d1 Pointer to the first element.
                     \param d2 Pointer to the second element.
                     \return The function must return:
                             <UL>
                               <li>-1 if <i>d1 < d2</i>
                               <li>0  if <i>d1 = d2</i>
                               <li>+1 if <i>d1 > d2</i>
                               <li>-2 if an error occured.
                             </UL>
                  */
                 int  (*compare)(void *d1, void *d2);

                 /*! \brief User defined function that returns a pointer to the element which index is <i>idx</i>.
                     \param start Pointer to the first element of the array.
                     \param idx Index of the element.
                     \return The function returns a pointer to the element which index is <i>idx</i>.
                     \warning The gap between two elements of the array depends on the type of data stored.
                  */
                 void* (*index)(void *start, int idx);

                 /*! \brief User defined function used to initialize data (may be a NULL pointer if no initialization is required).
                     \param data Pointer to the data to initialize.
                     \return Upon successful completion the function returns 0, otherwise it returns 1
                  */
                 int (*init_data)(void *data);

                 /*! \brief Number of elements scheduled for removing.
                  */
                 int element_to_remove;

                 /*! \brief Pointer to an array of integers that will receive the index to remove from the array.
                            This array is allocated during the initialization of the array. You must free the memory using 
                            the function array_free().
                  */
                 int *to_remove;
               };

/*! \brief C data type for the <i>array</i> handler.
 */

typedef struct s_array Array;

/*! \Brief This macro returns the size of the array (that is, the maximum number of elements that can be stored in the array).
 */ 

#define array_size(array) (array.array_size)

/*! \Brief This macro returns the current number of element in the array.
 */ 

#define array_number_of_elements(array) (array.number_of_elements)

/*! \Brief This macro return a pointer on the first element in the array.
 */

#define array_first_elem(array) (array.first_element)

/*! \Brief This macro returns the number of element to remove.
 */

#define array_to_remove(array) (array.element_to_remove)


int array_init (
                 Array *array,
                 void  *first_element,
                 int   size,
                 int   (*copy_data)(void *src, void *dest),
                 int   (*remove_data)(void *target),
                 int   (*compare)(void *d1, void *d2),
                 void* (*index)(void *start, int idx),
                 int   (*init_data)(void *data)
               );

void array_free (Array *array);

int array_init_element (Array *array);

int array_add (
                Array *array,
                void  *element
              );

int array_rem (
                Array *array,
                int   idx
              );

int array_insert (
                   Array *array,
                   int   idx,
                   void  *data
                 );

int array_schedule_for_rem (Array *array, int idx);

int array_rem_all (Array *array);

int array_find (
                   Array *array,
                   void  *data
               );
