/*! \file date.h
    Header file for all date's utilities.
 */

#ifdef __cplusplus
extern "C" {
#endif

#ifndef DATE_HPP

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include "strings_utils.h"

/*! \brief Maximum number of bytes for the environment variable &quot;TZ&quot; (ex: &quot;TZ=GMT +0&quot;).
 */

#define TZSET_MAX_SIZE  256

/*! \brief Maximum number of bytes for the tag that defines the time zone (MET, CEST, ...).
 */

#define TIMEZONE_SIZE   50

/*! \brief Maximum number of bytes for the tag that defines the time daylight saving mode (DST, ...).
 */

#define DAYLIGHT_SIZE   50

/*! \brief This defines the maximum number of bytes for a TSM formated date (ex: Thu Aug 30 01:00:48 MET DST
           2001).
 */

#define MAX_TSM_DATE_SIZE  256

/*! \brief This defines the maximum number of lines in the time zones configuration file. One line is a set of 3 vales:
           (time zone, daylight saving mode, time shift) - for example (MET DST -3600).
 */

#define MAX_TZ_DEF        300

/*! \brief Return value for the function load_timezones(). Too many time zones definitions in the time zones configuration
           file. This maximum number is represented by the constant MAX_TZ_DEF.
 */

#define TIME_ZONES_DEF_OVERFLOW        -1

/*! \brief Return value for the function load_timezones(). This means that the program is running out of memory.
 */

#define TIME_ZONES_NO_MEM              -2

/*! \brief Return value for the function load_timezones() (or via the time zone configuration parser). This means that
           the time zone configuration file is not valid. One time zone tag is too long (too many characters).
 */

#define TIME_ZONES_TIMEZONE_OVERFLOW   -3

/*! \brief Return value for the function load_timezones() (or via the time zone configuration parser). This means that
           the time zone configuration file is not valid. One daylight saving tag is too long.
 */

#define TIME_ZONES_DAYLIGHT_OVERFLOW   -4

/*! \brief Return value for the function load_timezones() (or via the time zone configuration parser). This means that
           the time zone configuration file is not valid. The parser found a syntax error.
 */

#define TIME_ZONES_PARSE_ERROR         -5

/*! \brief Return value for the function load_timezones(). No time zones configuration file specified. Please read the 
           documentation for the function load_timezones().
 */

#define TIME_ZONES_NO_FILE             -6

/*! \brief Return value for the function load_timezones(). This means that the time zones configuration file is empty
           (no configuration found).
 */

#define TIME_ZONES_NO_DEF_FOUND        -7

/*! \brief Return value for the function load_timezones(). Can not open the time zones configuration file.
 */

#define TIME_ZONES_CAN_NOT_OPEN_FILE   -8

/*! \brief Return value for the function load_timezones(). This means that the function loaded the time zones configuration
           file successfully.
 */

#define TIME_ZONES_LOAD_OK             -9


/*! \brief If the last argument of the function date2timestamp() is set to AUTOMATIC_TIME_SHIFT_COMPUTATION, then
           the function date2timestamp() will use the external timezone configuration file to find out the correct
           time shift. please note that AUTOMATIC_TIME_SHIFT_COMPUTATION = 24*3600 (one day shift).
 */

#define AUTOMATIC_TIME_SHIFT_COMPUTATION 86400

/*! \brief Data structure that defines a time zone (for example "MET DST -3600").
 */

struct timezones {

                   /*! \brief Tag that defines the time zone (ex: MET). */

                   char   timezone[TIMEZONE_SIZE];    /* ex: MET    */

                   /*! \brief Tag that defines the daylight saving mode (ex: DST). */

                   char   daylight[DAYLIGHT_SIZE];    /* ex: DST    */

                   /*! \brief Time shift (ex: -3600). */

                   int    shift;                      /* ex: -3600  */
                 };

/*! \brief Data structure that contains all the time zone configuration extracted from the time zones configuration file.
           This structure is set up by the function load_timezones() (calling the time zones configuration file parser).
 */

struct tz_tab {

                /*! \brief Array of data structures that define time zones. */

                struct timezones tz[MAX_TZ_DEF];

                /*! \brief Number of elements in the array 'tz'. */

                int              entry_number;
              };



/*! \brief Maximum size for a reference date (load the timestamps reference parser).
 */
#define TTREF_MAX_DATE_SIZE        1020

/*! \brief Maximum size for a reference timestamp (load the timestamps reference parser).
 */
#define TTREF_MAX_TIMESTAMP_SIZE   100

/*! \brief Maximum number of entries in the reference timestamp configuration file.
 */
#define MAX_TTREF_ENTRIES          300

/*! \brief The data structure defines a couple (date, timestamp) used  as reference.
 */
struct tstamp_ref {
                    /*! \brief Date used as reference. */

                    char   date[TTREF_MAX_DATE_SIZE];

                    /*! \brief Timestalo used as reference. */

                    char   timestamp[TTREF_MAX_TIMESTAMP_SIZE];
                  }; 


/*! \brief The data structure defines the list of couples (date, timestamp) loaded from the timestamps reference file).
 */ 
struct ttref_tab  {
                    /*! \brief List of all couples (date, timestamp) used  as reference. */

                    struct tstamp_ref ttref[MAX_TTREF_ENTRIES];

                    /*! \brief Number of couples (date, timestamp), in the array 'tstamp_ref'. */

                    int               entry_number;
                  };

/*! \brief Return value for the timestamps reference parser. 
 */
#define TTREF_TOO_MANY_REF  -1

/*! \brief Return value for the timestamps reference parser.
 */
#define TTREF_NO_MEM        -2

/*! \brief Return value for the timestamps reference parser.
 */
#define TTREF_SYNTAX_ERROR  -3

/*! \brief Return value for the timestamps reference parser.
 */
#define TTREF_DATE_TOO_LONG -4

/*! \brief Return value for the timestamps reference parser.
 */
#define TTREF_REF_TOO_LONG  -5

/*! \brief Return value for the function test_tz_conf(). This means that this function could not open the configuration
           file.
 */
#define TTREF_CAN_NOT_OPEN_CONF_FILE -6

/*! \brief Return value for the function test_tz_conf(). This means that one of the input date was invalid (the input
           date comes from the timestamps reference configuration file.
 */
#define TTREF_INVALID_DATE  -7

/*! \brief Return value for the function test_tz_conf(). This means that the timestamps' computation was correct.
 */
#define TTREF_OK            0 

/*! \brief Return value for the function test_tz_conf(). This means that the timestamps' computation was not correct.
 */
#define TTREF_KO            1

/*********************************************************/

/*! \brief Configuration value for the function test_tz_conf(). This means that we request the time zone auto configuration
           mode.
 */
#define TZ_AUTOCONF_ON    0

/*! \brief Configuration value for the function test_tz_conf(). This means that we do NOT request the time zone auto
           configuration mode.
 */
#define TZ_AUTOCONF_OFF   1

/*********************************************************/

/*! \brief This defines fields' indexes. The day of the week is the field number 0 (the first one) in the TSM's date.
 */

#define DAY_OF_THE_WEEK    0

/*! \brief This defines fields' indexes. The month is the field numbre 1 (the second one) in the TSM's date.
 */

#define MONTH              1

/*! \brief This defines fields' indexes. The day of the month is the field number 2 (the third one) in the TSM's date.
 */

#define DAY_OF_THE_MONTH   2

/*! \brief This defines fields' indexes. The clock time is the field number 3 (the fourth one) in the TSM's date.
 */

#define CLOCK              3

/*! \brief This defines fields' indexes. The hour is the field number 0 (the first one) in the TSM's clock time.
 */

#define HOUR               0

/*! \brief This defines fields' indexes. The minute is the field number 1 (the second one) in the TSM's clock time.
 */

#define MINUTES            1

/*! \brief This defines fields' indexes. The second is the field number 2 (the third one) in the TSM's clock time.
 */

#define SECONDS            2

/*! \brief This defines fields' indexes. The tag "MET" should be the field number 4 (the fifth one) in the TSM's date.
 */

#define TAG1               4

/*! \brief This defines fields' indexes. The tag "DST" should be the field number 5 (the sixth one) in the TSM's date.
 */

#define TAG2               5 

/*! \brief This defines fields' indexes. The year should be the field number 6 (the seventh one) in the TSM's date.
 */

#define YEAR               6

/*! \brief The maximum number of bytes used to save the current date.
 */

#define MAX_TIMER_SIZE     256

int set_local_TZ_to_GMT();
char* print_tz_conf();
int test_tz_conf (char* config_file, int auto_conf);
int load_timezones (char *filename);
int tzfound (struct timezones *tzz);
int tzcmp (struct timezones *tz1, struct timezones *tz2);
int get_day (char *day);
int get_month (char *month);
char* date2timestamp (char* date, char* timestamp, int nbcar, int shift);
char* dater();
time_t get_utc_timestamp();
char *get_tsm_date (time_t timestamp, int debug);

#define DATE_HPP
#endif


#ifdef __cplusplus
}
#endif

