/*! \file list.h
    Header file for list.c
 */

#ifdef __cplusplus
extern "C" {
#endif


#ifndef LIST_H
#define LIST_H

#include <stdlib.h>

  /* ----------------------------------------------------------------------- */
  /* Data structures                                                         */
  /* ----------------------------------------------------------------------- */

  /*! \brief This structure defines one element of the linked list.
   */

  typedef struct ListElmt_
  {
    /*! \brief Pointer to the element's data. */
    void  *donnee;

    /*! \brief Pointer to the next element in the linked list. */
    struct ListElmt_   *suivant;

    #ifdef DOUBLE_LINK

      /*! \brief Pointer to the previous element in the linked list. */ 
      struct ListElmt_   *precedent;

    #endif

  } ListElmt;

  /*! \brief This structure defines a linked list.
   */

  typedef struct List_
  {
    /*! \brief Number of elements in the linked list. */
    int    taille;
  
    /*! \brief Pointer to the user defined function used to compare two elements. */
    int    (*corresp)(const void *val1, const void *val2);

    /*! \brief Pointer on to the user defined function used to free the element's data memory. */
    void   (*detruire)(void *donnee);

    /*! \brief Pointer to the beginning of the linked list. */
    ListElmt   *tete;

    /*! \brief Pointer to the end of the linked list. */
    ListElmt   *queue;
  } List;

  /* ----------------------------------------------------------------------- */
  /* library's API                                                           */
  /* ----------------------------------------------------------------------- */

  void list_init(List *liste, void (*detruire)(void *donnee), int (*corresp)(const void *val1, const void *val2));

  void list_destroy(List *liste);

  int list_ins_next(List *liste, ListElmt *element, const void *donnee);

  int list_rem_next(List *liste, ListElmt *element, void **donnee);

  ListElmt* list_find (List *liste, ListElmt* from, void* data, ListElmt **prec);

  /* ----------------------------------------------------------------------- */
  /* Library's macros                                                        */
  /* ----------------------------------------------------------------------- */

  /*! \brief Return the number of elements in a linked list.
      \param liste Pointer to the list structure.
   */

  #define list_size(liste) ((liste)->taille)

  /*! \brief Return a pointer to the first element of the linked list.
      \param liste Pointer to the list structure.
   */

  #define list_head(liste) ((liste)->tete)

  /*! \brief Return a pointer to the last element of the linked list.
      \param liste Pointer to the list structure.
   */

  #define list_tail(liste) ((liste)->queue)

  /*! \brief Test if a given element is the first element of a linked list.
      \param liste Pointer to the list structure.
      \param element Pointer to the linked list element.
      \return This macro returns 1 if <i>element</i> point to the beginning of <i>liste</i>. Otherwise it returns 0.
   */

  #define list_is_head(liste, element)((element) == (liste)->tete ? 1 : 0)

  /*! \brief Test if a given element is the last element of a linked list.
      \param liste Pointer to the list structure.
      \param element Pointer to the linked list element.
      \return This macro returns 1 if <i>element</i> point to the end of <i>liste</i>. Otherwise it returns 0.
   */

  #define list_is_tail(element)((element)->suivant == NULL ? 1 : 0)

  /*! \brief Return a pointer to the data hold by an element.
      \param element Pointer to the linked list element.
      \return This macro returns a pointer to the data associted with <i>element</i>.
   */

  #define list_data(element) ((element)->donnee)

  /*! \brief Return a pointer to the next element in the linked list.
      \param element Pointer to the linked list element.
      \return This macro returns a pointer to the element next to <i>element</i>.
   */
 
  #define list_next(element) ((element)->suivant)

  #ifdef DOUBLE_LINK

    /*! \brief Return a pointer to the previous element in the linked list.
        \param element Pointer to the linked list element.
        \return This macro returns a pointer to the element previous to <i>element</i>.
     */

    #define list_prev(element) ((element)->precedent)
  #endif

#endif



#ifdef __cplusplus
}
#endif

