/*! \file my_smtp.h
    Header file for the smtp.c
 */

#ifdef __cplusplus
extern "C" {
#endif

#ifndef MY_SMTP_HEADER

  #include "my_sockets.h"

  /* ------------------------------------------------- */
  /* Configuration values and data structures          */
  /* ------------------------------------------------- */

  /*! \brief Maximum number of characters for the client identification (HELO command).
   */

  #define SMTP_ID_SIZE        1034

  /*! \brief Maximum number of characters for an e-mail address.
   */

  #define SMTP_EMAIL_SIZE_MAX 1024

  /*! \brief Maximum number of recipients for a given mail.
   */

  #define SMTP_RCPT_MAX       128

  /*! \brief This structure defines the information used by the SMTP client to send an e-mail.
   */

  struct smtp_mail {
                      /*! \brief The e-mail address of the sender. */
                      char from[SMTP_EMAIL_SIZE_MAX];

                      /*! \brief The e-mail addresses of the recipients. */
                      char rcpt[SMTP_RCPT_MAX][SMTP_EMAIL_SIZE_MAX];

                      /*! \brief The number of recipients. */
                      int  rcpt_number;

                      /*! \brief Mail body to be sent. This must point to a zero terminated string of
                                 characters. */
                      char *data;
                   };

  /*! \brief Maximum number of characters for a host name.
   */

  #define SMTP_HOSTNAME_SIZE   512

  /*! \brief This structure defines the information used by the SMTP client to connect to the server.
   */

  struct smtp_connexion_config {
                                 /*! \brief The client identification (HELO command). */
                                 char id[SMTP_ID_SIZE];

                                 /*! \brief TCP port used to the server (default is 25). */
                                 int port;

                                 /*! \brief Host name of the server that runs the SMTP server. */
                                 char host[SMTP_HOSTNAME_SIZE];

                                 /*! \brief Connexion timeout in seconds. */
                                 long  connect_timeout_sec;

                                 /*! \brief Connexion timeout in micro seconds. */
                                 long connect_timeout_micro;

                                 /*! \brief Read timeout in seconds. */
                                 long  read_timeout_sec;

                                 /*! \brief Read timeout in micro seconds. */
                                 long  read_timeout_micro;
                               };

  /* ------------------------------------------------- */
  /* Data structure used to store the status of an     */
  /* operation.                                        */
  /* ------------------------------------------------- */

  /*! \brief This structure defines the information returned by all client SMTP functions.
   */

  struct smtp_res {
                    /*! \brief Error code. */
                    int  error_code;

                    /*! \brief Ellapsed time in milli-seconds. */
                    long ellapsed;
                  };

  /* ------------------------------------------------- */
  /* Data structure used to store information about    */
  /* the answer of a SMTP server.                      */
  /* ------------------------------------------------- */

  /*! \brief This structure contains the information included in the answer from a SMTP server.
   */

  struct smtp_response {
                         /* Not used yet, but defined for future improvement */
                       };

  /* ------------------------------------------------- */
  /* Library API                                       */
  /* ------------------------------------------------- */

  int client_smtp_connect (
                            struct smtp_connexion_config *conf,
                            struct smtp_res *result
                          );

  void smtp_disconnect();

  int smtp_send_email (
                        struct smtp_connexion_config  *conf,
                        struct smtp_mail              *email,
                        struct smtp_res               *result
                      );

  /* ------------------------------------------------- */
  /* Error codes                                       */
  /*                                                   */
  /*                  ### WARNING ###                  */
  /*                                                   */
  /* This library uses the library 'libmy_sockets.a'.  */
  /* The returned error codes for the SMTP library     */
  /* must be negative and lower that the lowest return */
  /* value for the library 'libmy_sockets.a' (see file */
  /* "my_sockets.h").                                  */
  /* ------------------------------------------------- */

  /*! \brief Return balues for all the SMTP API. This means that the system call write() failed.
   */

  #define SMTP_WRITE_ERROR         -100

  /*! \brief Return balues for all the SMTP API. This means that the response from the SMTP server was
             not valid.
   */

  #define SMTP_BAD_ANSWER          -101

  /*! \brief Return balues for the function smtp_send_email(). This means that one of the input arguments
             "from" or "rcpt" is too long.
   */

  #define   SMTP_BUFFER_OVERFLOW   -102

  /* ------------------------------------------------- */
  /* Internal buffer size used to receive/send data    */
  /* from/to the SMTP server.                          */
  /* ------------------------------------------------- */

  /*! \brief Size of the buffer used to receive/send data from/to the SMTP server.
   */

  #define SMTP_BUFFER_SIZE   1024

  /* ------------------------------------------------- */
  /* SMTP status returned by the SMTP server           */
  /* ------------------------------------------------- */

  /*! \brief Return value for the function response_parser(). The SMTP server returned "220 ...". */

  #define SMTP_SERVER_IDENTIFICATION_OK    220

  /*! \brief Return value for the function response_parser(). The SMTP server returned "250 ...". */

  #define SMTP_CLIENT_IDENTIFICATION_OK    250

  /*! \brief Return value for the function response_parser(). The SMTP server returned "250 ...". */

  #define SMTP_FROM_OK                     250

  /*! \brief Return value for the function response_parser(). The SMTP server returned "250 ...". */

  #define SMTP_RCPT_OK                     250

  /*! \brief Return value for the function response_parser(). The SMTP server returned "354 ...". */

  #define SMTP_DATA_OK                     354

  /*! \brief Return value for the function response_parser(). The SMTP server returned "250 ...". */

  #define SMTP_END_OF_DATA_OK              250

  /*! \brief Return value for the function response_parser(). The SMTP server an invalid response. */

  #define SMTP_INVALID_ANSWER              -1

  /* ------------------------------------------------- */
  /* Constants used to format output to the SMTP       */
  /* server.                                           */
  /* ------------------------------------------------- */

  #define SMTP_FROM_TAG_START    "MAIL FROM:<"
  #define SMTP_FROM_TAG_STOP     ">\r\n"

  #define SMTP_RCPT_TAG_START    "RCPT TO:<"
  #define SMTP_RCPT_TAG_STOP     ">\r\n"

  #define MY_SMTP_HEADER
#endif

#ifdef __cplusplus
}
#endif

