/*! \file option.h
    This file defines data structures for the option reader.
 */


#ifdef __cplusplus
extern "C" {
#endif


#ifndef OPTION_H
#define OPTION_H
#endif

/*! \brief This data structure contains the definition for one option.
 */

typedef struct Option_
        {
          /*! \brief Pointer to a zero terminated string of characters that represents the option flag.
                     Please note that an option's flag should verify the following rules:
                     <UL>
                         <LI>At least 2 characters.
                         <LI>Begin with a '-'.
                         <LI>No spaces.
                         <LI>Not all '-'.
                     </UL>
          */
          char   *flag;

          /*! \brief This flag defines the number of argument required by the option. The option may require
                     zero (ex: --help) or one argument.
                     <UL>
                         <LI>If you set this argument to the value 0, it means that the option does not
                             require any arguemnt (ex: --help).
                         <LI>Any other value means that the option requires one (and only one) argument.
                     </UL>
          */
          int    arg;

          /*! \brief The tag defines the maximum number of characters for the option's argument.
          */
          int    max_size;

          /*! \brief Pointer to a zero terminated string of characters that represents the regular expression
                     that matches the option's argument.
          */
          char   *reg_exp;

          /*! \brief Address to a pointer that will receive the address of the sero terminated string of 
                     characters that represents en option's argument.
          */ 
          char   **pointer;

          /*! \brief Pointer to a zero terminated string of characters that contains a literal description of
                     the option.
          */
          char   *desc;
        } Option;

/*! \brief This data structure is used to store all option definitions for the process.
 */
typedef struct Option_set_
        {
          /*! \brief Pointer to an array of <I>Option</I> data structures that define all the process' options.
          */
          Option *opt;

          /*! \brief Number of elements in the array pointed by <I>opt</I>.
          */
          int    count;

          /*! \brief This integer represents the index (in the array of strings that represents the command
                     line) of the last item (option or argument) parsed. Following arguments start at index
                     <I>last_index</I> + 1.
          */
          int    last_index; 
        } Option_set;

int init_options (Option_set *opt);
int parse_command_line (char *argv[], int argc, Option_set *opt, char **err, char **flag, char **arg);
int get_argument (Option_set *opt, char *flag, char **arg);
int is_argument_required (Option_set *opt, char *flag);

/*! \brief Return the index (in the array of strings that represents the command ine) of the last item 
           (option or argument) parsed.
 */

#define last_argv_index(options)        (options.last_index)

#ifdef __cplusplus
}
#endif
