/*! \file array.c
    This file implements basic array manipulation tools.
 */


#include "array.h"
#include "sort.h"

/*! \brief Initialize an array handler.
    \param array Pointer to the array handler to initialize.
    \param first_element pointer to the first element of the array.
    \param size Size of the array (that is: the maximum number of elements in the array).
    \param copy_data User defined function used to copy data from/to the array.
                     Upon successful completion the function returns 0, otherwise it returns 1.
                     See the description of <i>struct s_array</i> for a complete description.
                     <FONT COLOR="#FF0000"><B>WARNING:</B></FONT> The function copy_data <B>MUST</B> take care off memory 
                     allocation/desallocation (if required).
    \param remove_data User defined function used to remove an element from the array.
                       Upon successful completion the function returns 0, otherwise it returns 1.
                       See the description of <i>struct s_array</i> for a complete description.
                       <i>remove_data</i> can be a NULL pointer if no special operation is required to remove an element.
    \param compare User defined function used to compare two elements of the array.
                   The function must return:
                   <UL>
                     <li>-1 if <i>d1 < d2</i>
                     <li>0  if <i>d1 = d2</i>
                     <li>+1 if <i>d1 > d2</i>
                     <li>-2 if an error occured.
                   </UL>
                   See the description of <i>struct s_array</i> for a complete description.
    \param index User defined function that returns a pointer to the element which index is <i>idx</i>.
                 See the description of <i>struct s_array</i> for a complete description.
    \param init_data User defined function used to initialize data. You can specify a NULL pointer if initialization is not required.
                     See the description of <i>struct s_array</i> for a complete description.
    \return Upon successful completion the function returns 0, otherwise it returns -1.
 */

int array_init (
                  Array *array,
                  void  *first_element,
                  int   size,
                  int   (*copy_data)(void *src, void *dest),
                  int   (*remove_data)(void *target),
                  int   (*compare)(void *d1, void *d2),
                  void* (*index)(void *start, int idx),
                  int   (*init_data)(void *data)
               )
{
  array->first_element      = first_element;
  array->array_size         = size;
  array->number_of_elements = 0;
  array->copy_data          = copy_data;
  array->remove_data        = remove_data;
  array->compare            = compare;
  array->index              = index;
  array->init_data          = init_data;
  array->element_to_remove  = 0;
  array->to_remove          = (int*)malloc(array->array_size * sizeof(int));

  if (array->to_remove == NULL) { return -1; }

  return 0;
}

/*! \example test_array.c
    This program shows how to use the array API
 */

/*! \brief Free all memory allocated for the array structure.
    \warning This function <B>DOES NOT</B> free memory allocated for the user's data (this is the responsability of the user).
 */

void array_free (Array *array) { free(array->to_remove); return; }

/*! \example test_array.c
    This program shows how to use the array API
 */

/*! \brief Schedule an element of the array for removing.
    \param array Pointer to an array data structure.
    \param idx Index of the element to remove.
    \return Upon successful completion the function returns 0, otherwise it returns -1.
 */

int array_schedule_for_rem (Array *array, int idx)
{
  int i;

  /* ------------------------------------------------- */
  /* First make sure that this index is not already    */
  /* scheduled.                                        */
  /* ------------------------------------------------- */

  for (i=0; i<array->element_to_remove; i++)
  { if (array->to_remove[i] == idx) { return 0; } }

  /* ------------------------------------------------- */
  /* Then, make sure that we can add an element to     */
  /* the list of elements to remove.                   */
  /* ------------------------------------------------- */

  if (array->element_to_remove >= array->array_size) { return -1; }

  /* ------------------------------------------------- */
  /* Add the index to the list of index to remove      */
  /* ------------------------------------------------- */

  array->to_remove[array->element_to_remove] = idx;

  /* ------------------------------------------------- */
  /* Increment the number of index to remove and exit  */
  /* ------------------------------------------------- */

  array->element_to_remove += 1;

  return 0;
}

/*! \example test_array.c
    This program shows how to use the array API
 */


/* ------------------------------------------------------------ */
/* Compare two integers -- Not part of public API               */
/* ------------------------------------------------------------ */

static int compare_int(const void *int1, const void *int2)
{
  if (*(const int *)int1 > *(const int *)int2) { return 1; }
  if (*(const int *)int1 < *(const int *)int2) { return -1; }
  return 0;
}

/*! \brief Remove a list of previouly scheduled element.
    \param array Pointer to an array data structure.
    \return Upon successful completion the function returns 0, otherwise it returns -1.
 */

int array_rem_all (Array *array)
{
  int i;

  /* ------------------------------------------------- */
  /* Is there something to remove ?                    */
  /* ------------------------------------------------- */

  if (array->element_to_remove == 0) { return 0; }

  /* ------------------------------------------------- */
  /* Sort the idexes of the elements to remove         */
  /* ------------------------------------------------- */

  issort((void*)array->to_remove, array->element_to_remove, sizeof(int), compare_int);

  /* ------------------------------------------------- */
  /* Modify indexes, so we take care of the left shift */
  /* ------------------------------------------------- */

  for (i=0; i<array->element_to_remove; i++) { array->to_remove[i] -= i; }

  /* ------------------------------------------------- */
  /* Now remove the elements                           */
  /* ------------------------------------------------- */
  
  for (i=0; i<array->element_to_remove; i++)
  { if (array_rem(array, array->to_remove[i]) == -1) { return -1; } }

  /* ------------------------------------------------- */
  /* Reset the number of element to remove             */
  /* ------------------------------------------------- */

  array->element_to_remove = 0;

  return 0;
}

/*! \brief Initialize all elements of an array.
    \param array Pointer to an array data structure.
    \return Upon successful completion the function returns 0, otherwise it returns -1.
    \warning Depending on what you put in the function <i>init_data()</i>, you may call this function one or more times.
 */

int array_init_element (Array *array)
{
  void *location;
  int  i;

  for (i=0; i<array->array_size; i++)
  {
    location = array->index(array->first_element, i);
    if (array->init_data(location) == 1) { return -1; }
  }

  return 0;
}

/*! \example test_array.c
    This program shows how to use the array API
 */

/*! \brief Add an element to the array (the element id added after the last element).
    \param array Pointer to the array handler to initialize.
    \param element Pointer to the element to add to the array.
    \return Upon successful completion the function returns 0, otherwise it returns -1.
 */

int array_add (
                Array *array,
                void  *element
              )
{
  void *location;

  /* ------------------------------------------------- */
  /* First, make sure that we can add an element       */
  /* ------------------------------------------------- */

  if (array->number_of_elements >= array->array_size) { return -1; }

  /* ------------------------------------------------- */
  /* Get a pointer to the first available location     */
  /* (That is, the end of the array)                   */
  /* Indexes start at 0. The index of the Nth element  */
  /* is N-1.                                           */
  /* ------------------------------------------------- */

  location = array->index(array->first_element, array->number_of_elements);

  /* ------------------------------------------------- */
  /* Copy the element to the available location        */
  /* ------------------------------------------------- */

  if (array->copy_data (element, location) == 1) { return -1; }
  
  /* ------------------------------------------------- */
  /* Increment the number of elements and exit         */ 
  /* ------------------------------------------------- */

  array->number_of_elements += 1;

  return 0;

}

/*! \example test_array.c
    This program shows how to use the array API
 */

/*! \brief Remove an element from an array. The elements next to the one removed are shifted to the left (so there is no gaps, and the
           elements' order does not change).
    \param array Pointer to the array handler to initialize.
    \param idx Index of the element to remove.
    \return Upon successful completion the function returns 0, otherwise it returns -1.
 */

int array_rem (
                Array *array,
                int   idx
              )
{
  void *location, *src, *dst;
  int  i;

  /* ------------------------------------------------- */
  /* First, make sure that 'idx' is not out of the     */
  /* array. Indexes start at 0. The index of the Nth   */
  /* element is N-1                                    */
  /* ------------------------------------------------- */

  if (idx > array->number_of_elements-1) { return -1; }

  /* ------------------------------------------------- */
  /* Get a pointer to the location pointed by 'idx'    */
  /* ------------------------------------------------- */

  location = array->index(array->first_element, idx);

  /* ------------------------------------------------- */
  /* Remove the element and shift the next elements    */
  /* ------------------------------------------------- */

  if (array->remove_data != NULL)
  { if (array->remove_data(location) == 1) { return -1; } }

  /* ------------------------------------------------- */
  /* Now, the difficult part: shift all the elements   */
  /* ------------------------------------------------- */

  for (i=idx; i<array->number_of_elements-1; i++)
  {
    dst = array->index(array->first_element, i);
    src = array->index(array->first_element, i+1);

    if (array->copy_data (src, dst) == 1) { return -1; }
  }

  /* ------------------------------------------------- */
  /* Remove the last element. Keep in mind that the    */
  /* function 'copy_data' doest not remove the source  */
  /* ------------------------------------------------- */

  if (array->remove_data != NULL)
  {
    location = array->index(array->first_element, array->number_of_elements-1);
    if (array->remove_data(location) == 1) { return -1; }
  }

  /* ------------------------------------------------- */
  /* Decrement the number of element in the array and  */
  /* exit                                              */
  /* ------------------------------------------------- */

  array->number_of_elements -= 1;

  return 0;
}

/*! \example test_array.c
    This program shows how to use the array API
 */

/*! \brief Insert an element in an array.
    \param array Pointer to an array data structure.
    \param idx Index in the array where to insert the new elememt.
    \param data Pointer to a user defined data (that represents the content of the new element).
    \return Upon successful completion the function returns 0, otherwise it returns -1.
 */

int array_insert (
                   Array *array,
                   int   idx,
                   void  *data
                 )
{

  void *src, *dst, *location;
  int  i;

  /* ------------------------------------------------- */
  /* First, make sure that we can add an element       */
  /* ------------------------------------------------- */

  if (array->number_of_elements >= array->array_size) { return -1; }

  /* ------------------------------------------------- */
  /* The make shure that 'idx' is a valid position     */
  /* Note: the index of the last element is            */
  /*       (array->number_of_elements - 1)             */
  /* ------------------------------------------------- */

  if (idx >= array->number_of_elements) { return -1; } 

  /* ------------------------------------------------- */
  /* Shift all elements which index is greater that    */
  /* 'idx' to the right.                               */
  /* ------------------------------------------------- */

  for (i=array->number_of_elements; i>idx; i--)
  {
    src = array->index(array->first_element, i-1);
    dst = array->index(array->first_element, i);

    if (array->copy_data (src, dst) == 1) { return -1; }
  }

  /* ------------------------------------------------- */
  /* Now copy 'data' at index 'idx'.                   */
  /* ------------------------------------------------- */

  location = array->index(array->first_element, idx);
  if (array->copy_data (data, location) == 1) { return -1; }

  /* ------------------------------------------------- */
  /* Increment the number of elements and exit         */
  /* ------------------------------------------------- */

  array->number_of_elements += 1;

  return 0;
}

/*! \example test_array.c
    This program shows how to use the array API
 */

/*! \brief Find an given element in the array.
    \param array Pointer to the array data structure.
    \param data Pointer to the user defined element to find.
    \return If the element is found, then the function returns the index of the element. Otherwise the function
            returns -1.
 */

int array_find (
                   Array *array,
                   void  *data
               )
{
  int  idx;
  void *location;

  for (idx=0; idx<array->number_of_elements; idx++)
  {
    location = array->index(array->first_element, idx);
    if (array->compare(location, data) == 0) { return idx; }
  }

  return -1;
}

/*! \example test_array.c
    This program shows how to use the array API
 */


