/*! \file cgi.c
    This file contains utilities for CGIs.
 */


#include <stdlib.h>
#include <stdio.h>
#include <strings.h>


/*! \brief Extract the login and the domain from a given email address (login@domain).
    \param address Pointer to a zero terminated string of characters that represents the email address.
    \param login Pointer to a memory location that will be used to store a zero terminated string of characters
                 that is the login.
    \param domain Pointer to a memory location that will be used to store a zero terminated string of characters
                  that is the domain.
    \return Upon successful completion, the function returns the value 0. Otherwise the function returns 1, this
            means that the given address does not contain any character '@'.
    \warning No test is done regarding the size of the memory chuncks pointed by login and domain. It is your
             responsability to allocate enough memory.
 */

int get_domain (char *address, char *login, char *domain)
{
  char *pt, c;

  pt = strstr(address, "@");
  if (pt == NULL) { return 1; }

  c = *pt;

  strcpy (domain, pt+1);

  *pt = 0;

  strcpy (login, address);

  *pt = c;

  return 0;
}

/*! \brief URL encode a string of characters.
    \param string Pointer to a zero terminated string that contains the string to encode.
    \return The function returns a pointer to the URL encoded string. If an error occured, then the function
            returns NULL.
    \warning Note that the returned string has been allocated by the function. Therefore you must free the memory.
 */ 

char* url_escape(char *string)
{
   int            alloc;
   char           *ns;
   unsigned char  in;
   int            newlen;
   int            index;



   newlen = alloc = strlen(string)+1;
   ns = (char*) malloc((size_t)alloc);
   if (ns == NULL) { return NULL; }
   index = 0;

   while((*string) != 0)
   {
      in = *string;
      if(' ' == in) { ns[index++] = '+'; }
      else if(!(in >= 'a' && in <= 'z') &&
              !(in >= 'A' && in <= 'Z') &&
              !(in >= '0' && in <= '9'))
           {
             /****************************************************/
             /* encode it:                                       */
             /* the size grows with two, since this'll become a  */
             /* %XX.                                             */
             /****************************************************/
	         
	         newlen += 2;
	         
	         if(newlen > alloc)
	         {
	           alloc *= 2;
	           ns = realloc(ns, alloc);
	           if (ns == NULL) { return NULL; }
	         }
	         
	         sprintf(&ns[index], "%%%02X", in);
	         index+=3;
           }
           else
           { ns[index++]=in; }

           string++;
   }
 
   ns[index]=0;
   return ns;
}

/*! \example test_url_encoding.c
    This file shows how to use the function url_escape().
 */

/*! \brief URL decode a string of characters.
    \param string Pointer to a zero terminated string that contains the string to decode.
    \param length number of character to decode (should be the size of the string).
    \return The function returns a pointer to the URL decoded string. If an error occured, then the function
            returns NULL.
    \warning Note that the returned string has been allocated by the function. Therefore you must free the memory.             */

char* url_unescape(char *string, int length)
{
   int alloc; 
   char *ns;
   unsigned char in;
   int index;
   int hex;
   char querypart;       /* everything to the right of a '?' letter is
                            the "query part" where '+' should become ' '.
                            RFC 2316, section 3.10 */

   alloc = (length?length:strlen(string))+1;
   ns = malloc(alloc);
   if (ns == NULL) { return NULL; }
   index = 0;
   querypart = 0;
  
   while(--alloc > 0)
   {
      in = *string;
      if ((querypart == 1) && ('+' == in)) { in = ' '; }
      else if ((querypart == 0) && ('?' == in))
      {
        /* we have "walked in" to the query part */
        querypart = 1;
      }
      else if('%' == in)
           {
             /* encoded part */
             if(sscanf(string+1, "%02X", &hex))
             {
               in = hex;
               string+=2;
               alloc-=2;
             }
           }

      ns[index++] = in;
      string++;
   }
   
   ns[index]=0; /* terminate it */
   return ns;
}  

/*! \example test_url_encoding.c
    This file shows how to use the function url_unescape().
 */


