/*! \file chtbl.c
    Hash table implementation
 */

#include <stdlib.h>
#include <string.h>
#include "list.h"
#include "chtbl.h"



/*! \brief Initialize the hash table. This function must be called first.
    \param htab Pointer to a hash table data structure.
    \param conteneurs Number of linked list in the hash table.
    \param h Pointer to the function used to hash the input key.
    \param corresp Pointer to the function used to compare user's defined data. This function is specific to the type of data stored into
                   the hash table. This function must return if the user defined data are identical, 0 otherwise.
    \param detruire Pointer to the function used to free all allocated memory for the user's defined data. This function is specific to the
                    type of data stored into the hash table.
    \return Upon successful completion the function returns 0, otherwise the function returns 1.
 */

int chtbl_init(
                CHTbl *htab,
                int   conteneurs, 
                int   (*h)(const void *cle), 
                int   (*corresp)(const void *cle1, const void *cle2), 
                void  (*detruire)(void *donnee)
              )
{
  int  i;

  /* ------------------------------------------------------------------------- */
  /* Allocate linked lists for hash table                                      */
  /* ------------------------------------------------------------------------- */

  if ((htab->table = (List *)malloc(conteneurs * sizeof(List))) == NULL) { return -1; }

  /* ------------------------------------------------------------------------- */
  /* Initialize the linked lists                                               */
  /* ------------------------------------------------------------------------- */

  htab->conteneurs = conteneurs;
  for (i = 0; i < htab->conteneurs; i++) { list_init(&htab->table[i], detruire, corresp); }

  /* ------------------------------------------------------------------------- */
  /* Initialize the hash table                                                 */
  /* ------------------------------------------------------------------------- */

  htab->h = h;
  htab->corresp = corresp;
  htab->detruire = detruire;
  htab->taille = 0;

  return 0;
}

/*! \example test_hash_table.c
    This test program shows how to use the hash table's API.
 */

/*! \brief Free all memory allocated for the hash table.
    \param htab Pointer to the hash table data structure to destroy.
 */

void chtbl_destroy(CHTbl *htab)
{
  int                i;
 
  /* ------------------------------------------------------------------------- */
  /* Delete all linked list                                                    */
  /* ------------------------------------------------------------------------- */

  for (i = 0; i < htab->conteneurs; i++) { list_destroy(&htab->table[i]); }

  /* ------------------------------------------------------------------------- */
  /* Free all memory used by the hash table data structure                     */
  /* ------------------------------------------------------------------------- */

  free(htab->table);

  /* ------------------------------------------------------------------------- */
  /* Clean up the structure (just to look better)                              */
  /* ------------------------------------------------------------------------- */

  memset(htab, 0, sizeof(CHTbl));

  return;
}

/*! \example test_hash_table.c
    This test program shows how to use the hash table's API.
 */

/*! \brief Insert data into the hash table.
    \param htab Pointer to the hash table data structure.
    \param donnee Pointer to the user defined data to insert.
    \return The function may return one of the following value:
            <UL>
              <li>0: The data was not in the hash table and it has been successfuly inserted.
              <li>1: The data already was in the hash table (this is not an error).
              <li>-1: An error occured (this means that the program is running out of memory).
            </UL>
    \warning Data are <B>NOT</B> copied into the hash table. The hash table only contains pointers to user's data. Therefore you should
             not free the memory allocated fot the data yourself.
 */

int chtbl_insert(CHTbl *htab, const void *donnee)
{
  void *temp;
  int   conteneur, retval;

  /* ------------------------------------------------------------------------- */
  /* Make sure that the data is not already into the hash table                */
  /* ------------------------------------------------------------------------- */

  temp = (void *)donnee;
  if (chtbl_lookup(htab, &temp) == 0) { return 1; }

  /* ------------------------------------------------------------------------- */
  /* Hash the input key                                                        */
  /* ------------------------------------------------------------------------- */

  conteneur = htab->h(donnee) % htab->conteneurs;

  /* ------------------------------------------------------------------------- */
  /* Insert data into the hash table                                           */
  /* ------------------------------------------------------------------------- */

  if ((retval = list_ins_next(&htab->table[conteneur], NULL, donnee)) == 0)
  { htab->taille++; }

  return retval;
}

/*! \example test_hash_table.c
    This test program shows how to use the hash table's API.
 */

/*! \brief Remove a record from the hash table.
    \param htab Pointer to a hash table data structure.
    \param donnee Pointer to a data structure pointer that will point to the data of the element that have been removed. Keep in mind that
                  the user data is <B>NOT</B> freed by the function chtbl_remove(). You must free it yourself.
    \return Upon successful completion the function returns 0, otherwise the function returns -1. Please note that the return value -1
            is not really an error. It really means that the data was not found in the hash table.
    \warning You must keep in mind that the data section of the removed element is <B>NOT</B> freed by the function. A pointer to the
             data section is returned (via <i>donnee</i>). This allows you to free it yourself.
 */

int chtbl_remove(CHTbl *htab, void **donnee)
{
  ListElmt           *element, *prec;
  int                conteneur;

  /* ------------------------------------------------------------------------- */
  /* Hash the input key                                                        */
  /* ------------------------------------------------------------------------- */

  conteneur = htab->h(*donnee) % htab->conteneurs;

  /* ------------------------------------------------------------------------- */
  /* Search the linked list for the data                                       */
  /* ------------------------------------------------------------------------- */

  element = list_find (&htab->table[conteneur], NULL, *donnee, &prec); 
  if (element != NULL)
  {
    /* ----------------------------------------------------------------------- */
    /* Data found, remove it                                                   */
    /* Note that 'prec' could be NULL, but that's OK.                          */
    /* ----------------------------------------------------------------------- */

    if (list_rem_next(&htab->table[conteneur], prec, donnee) == 0)
    {
      htab->taille--;
      return 0;
    }
  }

  /* ------------------------------------------------------------------------- */
  /* The data was not found                                                    */
  /* ------------------------------------------------------------------------- */

  return -1;

}

/*! \example test_hash_table.c
    This test program shows how to use the hash table's API.
 */

/*! \brief Searh the hash table for a given data.
    \param htab Pointer to a hash table structure.
    \param donnee Pointer to a user defined data. This pointer will be used:
                  <UL>
                     <li>To specify the data to search for.
                     <li>To point to the element's data (that has been found).
                  </UL>
    \return If the element was in the hash table, that the function returns 0 and <i>donnee</i> points to the element's data. Otherwise, 
            the function return -1.
 */

int chtbl_lookup(const CHTbl *htab, void **donnee)
{
  ListElmt           *element;
  int                conteneur;
 
  /* ------------------------------------------------------------------------- */
  /* Hash the input key                                                        */
  /* ------------------------------------------------------------------------- */

  conteneur = htab->h(*donnee) % htab->conteneurs;

  /* ------------------------------------------------------------------------- */
  /* Search for the jey in the linked list                                     */
  /* ------------------------------------------------------------------------- */

  for (element = list_head(&htab->table[conteneur]); element != NULL; element = list_next(element))
  {
    if (htab->corresp(*donnee, list_data(element)))
    {
      /* --------------------------------------------------------------------- */
      /* The record has been found                                             */
      /* Assign 'donnee' to the element's data.                                */
      /* --------------------------------------------------------------------- */

      *donnee = list_data(element);
      return 0;
    }
  }

  /* ------------------------------------------------------------------------- */
  /* The data was not found                                                    */
  /* ------------------------------------------------------------------------- */

  return -1;
}

/*! \example test_hash_table.c
    This test program shows how to use the hash table's API.
 */

/*! \brief Cover the entire hash table in search for an given element.
    \param htab Pointer to a hash table structure.
    \param donnee Pointer to a user defined data. This pointer will be used:
                  <UL>
                     <li>To specify the data to search for.
                     <li>To point to the element's data (that has been found).
                  </UL>
    \return If the element was in the hash table, that the function returns 0 and <i>donnee</i> points to the element's data. Otherwise, the function return -1.
    \warning This function lookup compare each element in the hash table with a given element (<i>donnee</i>).
             It may take long ... do not use it if you don't have to!
 */

int chtbl_find (const CHTbl *htab, void **donnee)
{
  int       ht;
  List      *linked_list;
  ListElmt  *element;

  /* ------------------------------------------------------------------------- */
  /* For each linked list, look for the data                                   */
  /* ------------------------------------------------------------------------- */

  for (ht=0; ht<htab->conteneurs; ht++)
  {
    linked_list = &(htab->table[ht]);
    element     = list_head (linked_list);

    while (element != NULL)
    {
      if (htab->corresp(list_data(element), *donnee))
      {
        *donnee = list_data(element);
        return 0;
      }
      element = list_next(element);
    }
  }

  return -1;
}

/*! \example test_hash_table.c
    This test program shows how to use the hash table's API.
 */

/*! \brief Returns the next element of a given linked list.
    \param htab Pointer to a hash table structure.
    \param rewind Flag used to set the hash pointer to the first element of the hash. The value of this
                  parameter can be:
                  <UL>
                     <LI>REWIND_HASH: Set the hash pointer to the first element of the hash. 
                     <LI>CURRENT_HASH: Set the hash pointer to the current element in the hash. 
                  </UL>
    \return The function returns a pointer to the current element in the hash, or NULL if the current element
            is the last one.
    \warning Each time you call the function, it returns the next element in the hash (until the last one where
             the function returns NULL).
             <UL>
                <LI>When you <B>start walking</B> through a hash table, you <B>must</B> set the parameter
                    <I>rewind</I> to the value REWIND_HASH !. If you don't do that, the baheviour of the function
                    is unpredictable. You may generate a segmentation fault.
                <LI>Do <B>NOT</B> modify (add or remove elements) the hash while you are walking
                    through it! If you don't do that, the baheviour of the function is unpredictable.
                    You may generate a segmentation fault.
             </UL>
 */

void* chtbl_next (CHTbl *htab, int rewind)
{
  List      *linked_list;
  ListElmt  *element;

  /* ------------------------------------------------------------------------- */
  /* Set the hash pointer to the first element if necessary                    */
  /* ------------------------------------------------------------------------- */

  if (rewind == REWIND_HASH)
  {
    linked_list = &(htab->table[0]);
    (htab->pointer).current_list    = 0;
    (htab->pointer).current_element = list_head(linked_list);
  }
  else
  {
    linked_list = &(htab->table[(htab->pointer).current_list]);
  } 

  /* ------------------------------------------------------------------------- */
  /* Find the next element                                                     */
  /* ------------------------------------------------------------------------- */

  while ((htab->pointer).current_element == NULL)
  {
    /* ----------------------------------------------------------------------- */
    /* Go to the next linked list                                              */
    /* ----------------------------------------------------------------------- */

    (htab->pointer).current_list += 1;
    if ((htab->pointer).current_list >= htab->conteneurs) { return NULL; }

    linked_list = &(htab->table[(htab->pointer).current_list]);
    (htab->pointer).current_element = list_head(linked_list);
  }

  element = (htab->pointer).current_element;

  (htab->pointer).current_element = list_next(element);

  return list_data(element);
}

/*! \example test_hash_table.c
    This test program shows how to use the hash table's API.
 */


