/*! \file config.c
          Configuration parser interface.
 */

#include <stdio.h>
#include <errno.h>
#include "chtbl.h"

#define PRIVATE_CODE_ONLY
#include "config.h"
#undef PRIVATE_CODE_ONLY


/*! \brief This flag indicates wether the memory allocated for the hash table is freed or not.
 */

static int allochtab = 0;

/*! \brief Hash function used by the hash table module.
    \remark See example for the hash table module.
 */

int config_hash_function (const void *key)
{
  struct user_data *data;
  int              index, i;

  data = (struct user_data*)key;
  index = 0;

  for (i=0; i<strlen(data->tag); i++) { index *= (int)(data->tag[i]); }

  if (index < 0) { index *= - 1; }

  return index;
}

/*! \brief Function used to free allocated memory in the hash table.
    \remark See example for the hash table module.
 */

void config_free_user_data (void *d)
{
  struct user_data *data;

  data = (struct user_data*)d;
  free (data->tag);
  free (data->value);
  free (data);

  return;
}

/*! \brief Function used to compare 2 elements in the hash table.
    \remark See example for the hash table module.
 */

int config_compare_user_data (const void *d1, const void *d2)
{
  struct user_data *data1, *data2;

  data1 = (struct user_data*)d1;
  data2 = (struct user_data*)d2;

  return !strcmp(data1->tag, data2->tag);
}

/*! \brief Handler to the hash table.
 */

CHTbl htab;

/*! \brief Flex generated parser.
 */

extern int configlex();

/*! \brief Flex input file descriptor.
 */

extern FILE *configin;

/*! \brief Load a given configuration file.
    \param path Path to the configuration file to load.
    \param err Pointer to an integer that will be used to store the value of 'errno' if an error occured.
    \return Upon successful completion, the function returns tha value PARSE_CONFIGURATION_OK.
            Otherwise the function may return one of the following values:
            <UL>
               <LI>PARSE_CONFIGURATION_NO_MEM: Could not allocate memory.
               <LI>PARSE_CONFIGURATION_FILE_NOT_FOUND: Could not open configuration file.
               <LI>PARSE_CONFIGURATION_SYNTAX_ERROR: Syntax error in configuration file.
               <LI>PARSE_CONFIGURATION_DUPLICATED_TAG: Configuration contains a duplicated tag.
               <LI>PARSE_CONFIGURATION_OK: Success.
            </UL>
    \warning <UL>
               <LI>If the returned value is PARSE_CONFIGURATION_NO_MEM, the module does not free
                   allocated memory. You should abort the program.
               <LI>If the returned value is PARSE_CONFIGURATION_FILE_NOT_FOUND, PARSE_CONFIGURATION_SYNTAX_ERROR
                   or PARSE_CONFIGURATION_DUPLICATED_TAG, the module free all allocated memory.
             </UL>
 */

int parse_configuration_file (char *path, int *err)
{
  int rc, retval;

  *err   = 0;
  retval = PARSE_CONFIGURATION_OK;

  /* ------------------------------------------------ */
  /* Initialize the hash table                        */
  /* ------------------------------------------------ */

  if (
       chtbl_init (
                    &htab,
                    10,
                    config_hash_function,
                    config_compare_user_data,
                    config_free_user_data
                  ) == -1
     )
  { return PARSE_CONFIGURATION_NO_MEM; }

  allochtab = 1;
 
  /* ------------------------------------------------ */
  /* Parse configuration file                         */
  /* ------------------------------------------------ */

  configin = fopen (path, "r");
  if (configin == NULL) { *err = errno; return PARSE_CONFIGURATION_FILE_NOT_FOUND; }

  rc = configlex();
  switch (rc)
  {
    case CONF_PARSER_SYNTAX_ERROR:   retval = PARSE_CONFIGURATION_SYNTAX_ERROR;
                                     free_all_configuration_data();
                                     break;
    case CONF_PARSER_DUPLICATED_TAG: retval = PARSE_CONFIGURATION_DUPLICATED_TAG;
                                     free_all_configuration_data();
                                     break;
    case CONF_PARSER_NO_MEM:         retval = PARSE_CONFIGURATION_NO_MEM; 
                                     break;
  }

  fclose (configin);

  return retval;
}

/*! \brief Free all data allocated for the hash table.
 */

void free_all_configuration_data() { if (allochtab == 1) { chtbl_destroy (&htab); allochtab = 0; } }

/*! \brief Return the value associated with a given tag.
    \param tag Pointer to a zero terminated string of characters that represents the tag to look for.
    \param value Pointer that will point to the corresponding value (if the tag was found).
    \return <UL>
              <LI>If the tag could not be find, the function returns the value 0. 'value' is NULL.
              <LI>If the tag has been found, the function returns the value 1.
              <LI>If the function ran out of memory, the function returns the value -1. 'value' is NULL.
            </UL> 
 */

int get_value (char *tag, char **value)
{
  struct user_data *ud, sud;
  int              cr, retval;

  /* ------------------------------------------------ */
  /* Initialise values                                */
  /* ------------------------------------------------ */

  *value  = NULL;
  retval  = 0;
 
  /* ------------------------------------------------ */
  /* Prepare user data for lookup                     */
  /* ------------------------------------------------ */

  sud.tag   = tag;
  sud.value = NULL;

  ud = &sud;

  /* ------------------------------------------------ */
  /* Lookup hash table                                */
  /* ------------------------------------------------ */

  cr = chtbl_lookup (&htab, (void**)&ud);

  if (cr == 0) { retval = 1; *value = ud->value; }

  return retval;
}









