/*! \file cyclic_buffer.c
    This file contains a basic implementation of cyclic buffer.
 */

#include "cyclic_buffer.h"


/*! \brief Create a new cyclic buffer. 
    \param size Maximum number of elements of the buffer.
    \return The function return a handle on the newly vreated cyclic buffer, or NULL if an error occured (out of memory).
    \warning The function create_cyclic_buffer() allocates memory (using malloc()). Therefore, when you don't need the
             buffer anymore, you should destry it using destroy_cyclic_buffer().
 */

cyclic_buffer create_cyclic_buffer (unsigned long size)
{
   cyclic_buffer buffer;
   unsigned long i;

   /* ======================================================= */
   /* Allocate new data structure                             */
   /* ======================================================= */

   buffer = (cyclic_buffer)malloc(sizeof(struct s_cyclic_buffer));
   if (buffer == NULL) { return NULL; }

   buffer->buffer = (void**)malloc(size*sizeof(void*));
   if (buffer->buffer == NULL)
   {
     free (buffer);
     return NULL;
   }

   /* ======================================================= */
   /* Initialize data structure                               */
   /* ======================================================= */

   for (i=0; i<size; i++) { (buffer->buffer)[i] = NULL; }
   buffer->size         = size;
   buffer->start        = 0;
   buffer->stop         = 0;
   buffer->current      = 0;
   buffer->current_size = 0;
   buffer->n            = 0;

   return buffer;
}

/*! \example test_cyclic_buffer.c
    This file shows how to use the function create_cyclic_buffer().
 */

/*! \brief Destroy a cyclic buffer (free all allocated memory).
    \param buffer Handle to the cyclic buffer to destroy.
 */

void destroy_cyclic_buffer (cyclic_buffer buffer)
{
   unsigned long i;

   /* ======================================================= */
   /* Free all alocated memory                                */
   /* ======================================================= */

   for (i=0; i<buffer->size; i++) 
   {
     if ((buffer->buffer)[i] != NULL) { free ((buffer->buffer)[i]); }
   }

   free (buffer->buffer);
   free (buffer);
}

/*! \example test_cyclic_buffer.c
    This file shows how to use the function destroy_cyclic_buffer().
 */

/*! \brief Add a new element to the cyclic buffer.
    \param buffer Handle to the cyclic buffer that will be used to receive the new element.
    \param element Pointer to the element to add to the cyclic buffer.
    \param element_size Size (in bytes) of the element to add.
    \return If succees, the function returns a pointer to the element in the cyclic buffer. Otherwise, if an error occured,
            the function returns NULL.
    \warning Please note that the element is <B>copied</B> into the cyclic buffer. This implies that memory is allocated.
 */

void* cyclic_buffer_add_element (cyclic_buffer buffer, void* element, unsigned element_size)
{
  void *elem;

  /* ======================================================== */
  /* Delete element if necessary                              */ 
  /* ======================================================== */

  if ((buffer->buffer)[buffer->stop] != NULL)
  {
    free ((buffer->buffer)[buffer->stop]);
    (buffer->buffer)[buffer->stop] = NULL;
  }

  /* ======================================================== */
  /* Allocate new element                                     */
  /* ======================================================== */

  (buffer->buffer)[buffer->stop] = (void*)malloc(element_size);
  if ((buffer->buffer)[buffer->stop] == NULL) { return NULL; }

  /* ======================================================== */
  /* Copy element                                             */
  /* ======================================================== */

  memcpy ((buffer->buffer)[buffer->stop], element, element_size);
  elem = (buffer->buffer)[buffer->stop];

  /* ======================================================== */
  /* update indexes                                           */
  /* ======================================================== */

  buffer->stop = (buffer->stop + 1) % buffer->size; 
  if (buffer->stop == buffer->start) { buffer->start = (buffer->stop + 1) % buffer->size; }

  /* ======================================================== */
  /* One more element                                         */
  /* ======================================================== */

  if (buffer->current_size < buffer->size) { buffer->current_size += 1; }

  return elem;
}

/*! \example test_cyclic_buffer.c
    This file shows how to use the function add_element().
 */


/*! \brief Reset the current buffer's pointer (used to extract element from the buffer).
    \param buffer Handle to the cyclic buffer.
 */

void rewind_cyclic_buffer (cyclic_buffer buffer)
{
  buffer->current = buffer->start;
  buffer->n       = 0;
}

/*! \example test_cyclic_buffer.c
    This file shows how to use the function rewind_cyclic_buffer().
 */

/*! \brief Returns the current element of the cyclic buffer.
    \param buffer Handle to the cyclic buffer.
    \return The function returns a pointer to the current buffer's element, or NULL if the buffer's pointer reached the end
            of the buffer.
    \warning Before calling get_element(), you should call rewind_cyclic_buffer() first.
 */

void* cyclic_buffer_get_element(cyclic_buffer buffer)
{
  void    *elem; 

  if (buffer->n >= buffer->current_size) { buffer->n = 0;  return NULL; }

  elem = (buffer->buffer)[buffer->current];
  buffer->current = (buffer->current + 1) % buffer->size;
  (buffer->n)++;

  return elem;
}

/*! \example test_cyclic_buffer.c
    This file shows how to use the function get_element().
 */

/*! \brief Return the index of the first element in the buffer.
    \param buffer Handle to the cyclic buffer.
    \return The function returns the index of the first element in the buffer.
 */

unsigned long cyclic_buffer_get_start(cyclic_buffer buffer) { return buffer->start; }

/*! \example test_cyclic_buffer.c
    This file shows how to use the function get_start().
 */

/*! \brief Return the index of the last element in the buffer.
    \param buffer Handle to the cyclic buffer.
    \return The function returns the index of the last element in the buffer.
 */

unsigned long cyclic_buffer_get_stop(cyclic_buffer buffer) { return buffer->stop; }

/*! \example test_cyclic_buffer.c
    This file shows how to use the function get_stop().
 */

/*! \brief Return the number of elements in the cyclic buffer.
    \param buffer Handle to the cyclic buffer.
    \return The function returns the number of elements in the cyclic buffer.
 */ 

unsigned long cyclic_buffer_get_current_size(cyclic_buffer buffer) { return buffer->current_size; }

/*! \example test_cyclic_buffer.c
    This file shows how to use the function get_current_size().
 */


