/*! \file date.c
    This file contains various functions used to manupulate dates.
 */

#include "date.h"


/*! \brief Current line in the time zones definitions file.
 */

int             date_line_num;

/*! \brief This structure contains the date 'timezone', 'daylight', 'time shift'.
 */

struct tz_tab   tzs;

/*! \brief Parser's entry point (created by flex).
 */

int             dateyylex();

/*! \brief Parser's input stream (used by dateyylex()).
 */

extern FILE     *dateyyin;


/*! \brief This array should contain &quot;MET&quot; and &quot;MET DST&quot; for TZ=MET.
 */

extern char *tzname[2];

/*! \brief The external time_t variable altzone  contains  the  difference, in seconds, between Coordinated Universal
           Time and the alternate time zone.
 */
extern time_t timezone;

/*! \brief The external variable daylight indicates whether time should  reflect  daylight  savings  time.  Both timezone and
           altzone default to 0 (UTC). The external  variable  daylight is  non-zero if an alternate time zone exists.
 */
extern int daylight;

/*! \brief The function date2timestamp() lookups the structure 'tzs' to find the correct timezone configuration for the 
           current date actually processed. This variable points to the current time zone configuration that aplies to
           the currently scaned date.
 */
struct timezones *current_tz;

/*! \brief Set the local time zone to GTM (which is equivalent to UTC).
    \return The function returns 0 if success, 1 otherwise.
 */

int set_local_TZ_to_GMT()
{
  static char current_tz[TZSET_MAX_SIZE];


  strcpy (current_tz, "TZ=GMT +0");

  if (putenv(current_tz) != 0) { return 1; }
  tzset();
  return 0;
}

/*! \brief Load time zones definitions.
    \param filename Pointer to a NULL terminated string of characters that contains the name on the file that contains the
           time zones definitions. If 'filename' is NULL, then the function looks for the environmanr variable 
           "TIMEZONES_DEFINITION".
    \return If the operation succed, the function returns TIME_ZONES_LOAD_OK. Otherwise, the function may return:
<UL>
  <li>TIME_ZONES_NO_FILE: No file name specified (ie; 'filename' is NULL, and TIMEZONES_DEFINITION is not set).
  <li>TIME_ZONES_CAN_NOT_OPEN_FILE: Can not open definition file.
  <li>TIME_ZONES_NO_DEF_FOUND: Definition file is empty.
  <li>TIME_ZONES_DEF_OVERFLOW: Too many timezones definitions.
  <li>TIME_ZONES_NO_MEM: Running out of memory.
  <li>TIME_ZONES_TIMEZONE_OVERFLOW: the timezone flag is too long.
  <li>TIME_ZONES_DAYLIGHT_OVERFLOW: the daylight flag is too long.
  <li>TIME_ZONES_PARSE_ERROR: parser error - check config file.
</UL>

 */

int load_timezones (char *filename)
{
  char *fname;
  int  parser_status;


  if (filename == NULL)
  {
    fname = getenv("TIMEZONES_DEFINITION");
    if (fname == NULL) { return TIME_ZONES_NO_FILE; }
  }
  else (fname = filename);

  dateyyin = fopen (fname, "r");
  if (dateyyin == NULL) { return TIME_ZONES_CAN_NOT_OPEN_FILE; }
  
  parser_status = dateyylex();

  if (parser_status != 0) { return parser_status; }

  if (tzs.entry_number == 0) { return TIME_ZONES_NO_DEF_FOUND; }

  return TIME_ZONES_LOAD_OK;
}

/*! \example test_load_tz.c
    This file shows how to use the function load_timezones().
 */


/*! \brief Compare 2 timezone structures.
    \param tz1 Pointer to the first timezone to compare.
    \param tz2 Pointer to the second timezone to compare.
    \return The function returns 1 if the two structure are identical, otherwise the function returns 0.
 */

int tzcmp (struct timezones *tz1, struct timezones *tz2)
{
  if (strcmp (tz1->timezone, tz2->timezone) != 0) { return 0; }
  if (strcmp (tz1->daylight, tz2->daylight) != 0) { return 0; }

  return 1;
}

/*! \brief Find a time zone data structure in the time zone configuration file.
    \param tzz Pointer to a time zone data structure. This structure must have the fields 'timezone' and 'daylight'
           correctly initialized (NULL terminated string of characters).
    \return If the function found a configuration that matches the time zone defined by 'tzz', it returns 1. Otherwise,
            the function returns 0. Please note that the structure 'tzz' will contain the correct value for the field
            'shift'.
 */

int tzfound (struct timezones *tzz)
{
  int i;

  current_tz = NULL;
  for (i=0; i<tzs.entry_number; i++)
  {
    if (tzcmp (&((tzs.tz)[i]), tzz) == 1)
    {
      current_tz = &((tzs.tz)[i]);
      tzz->shift = (tzs.tz)[i].shift;
      return 1;
    }
  }

  return 0;
}

/*! \example test_load_tz.c
    This file shows how to use the function tzfound().
 */

/*! \brief Contains the names of the 7 days of the week.
 */

extern char *days[7];

/*! \brief Contains the names of the 12 months of the year.
 */

extern char *months[12];

/*! \brief Returns the index of the day (starting at 0, so the returned value is between 0 and 6). The first day (index 0)
           is 'Sunday'.
    \param day Pointer to a string of characters that contains the day of the week.
    \return The function returns the index of the day (starting at 0, so the returned value is between 0 and 6). If an
            error occurs, the function returns -1.
 */

int get_day (char *day)
{
  int  i;

  for (i=0; i<7; i++)
  { if (strcmp(day, days[i]) == 0) { return i; } }

  return -1;
}

/*! \brief Returns the index of the month (starting at 0, so the returned value is between 0 and 11). The first month
           (index 0) is 'january'.
    \param month Pointer to a string of characters that contains the month of the year.
    \returns The function returns the index of the month (starting at 0, so the returned value is between 0 and 11).
             If an error occurs, the function returns -1.
 */ 

int get_month (char *month)
{
  int  i;

  for (i=0; i<12; i++)
  { if (strcmp(month, months[i]) == 0) { return i; } }

  return -1;
}

/*! \brief Returns a zero terminated string of characters that represents the time zones configuration.
    \return If the function succed, it returns a non NULL pointer. Otherwise, it returns NULL.
    \warning The returned pointer has been allocated within the function, make sure to free it (using free()).
 */

char* print_tz_conf()
{
  int    n, size, i;
  char   *res;

  n = tzs.entry_number;

  size = n * (TIMEZONE_SIZE + DAYLIGHT_SIZE + TTREF_MAX_TIMESTAMP_SIZE + 50);
  res  = (char*)malloc((size+1)*sizeof(char));
  if (res == NULL) { return NULL; }

  *res = 0;

  for (i=0; i<n-1; i++)
  {
    strcat (res, "(");
    strcat (res, (tzs.tz)[i].timezone);
    strcat (res, ", ");
    strcat (res, (tzs.tz)[i].daylight);
    strcat (res, ", ");
    sprintf (res+strlen(res), "%d) - ", (tzs.tz)[i].shift);
  } 

  strcat (res, "(");
  strcat (res, (tzs.tz)[i].timezone);
  strcat (res, ", ");
  strcat (res, (tzs.tz)[i].daylight);
  strcat (res, ", ");
  sprintf (res+strlen(res), "%d)", (tzs.tz)[i].shift);

  return res;
}



/*! \brief Calculate the timestamp of a ticket (The number of seconds since the first of January 1970).
    \param date Pointer to a string of characters that contains the date as it is read from the TSM's log. the date looks
           something like ' Tue Aug 28 01:02:24 MET DST 2001'.
    \param timestamp Pointer to a string that will receive the timestamp.
    \param nbcar Size of the memory location pointed by 'timestamp'.
    \param shift This value represents the number of seconds between the current date and the UTC date. If the value of
           shift is AUTOMATIC_TIME_SHIFT_COMPUTATION, then the function will use the time zones configuration file
           to find out the correct time shift.
    \return The function returns a pointer to 'timestamp'. If an error occurs, the function returns NULL. This can mean that
            the program is running out of memory, or that the date is not well formated (not "MET" "DST" ...).
    \warning <UL>
               <li>The first thing to do when using all this function of this library is to call set_local_TZ_to_GMT().
                   Otherwise, timestamps will depend on the system's local time zone configuration.
               <li>You MUST call the function load_timezones() before using this function !!!! If you do not do that you may
                   produce a segmentation fault. 
               <li>The input date 'date' will be modified. If you need to used it after, then make a copy!.
             </UL>

 */

char* date2timestamp (char* date, char* timestamp, int nbcar, int shift)
{
  struct fields    f;
  struct fields    t;
  int              nb_fields;
  struct tm        tt;
  int              seconds;            /* 0    -  59   */
  int              minutes;            /* 0    -  59   */
  int              hours;              /* 0    -  23   */
  int              day_of_the_month;   /* 1    -  31   */ 
  int              month;              /* 0    -  11   */
  int              year;               /* year - 1900  */
  int              day_of_the_week;    /* 0    -  6    */
  int              day_of_the_year;    /* 0    -  365  */
  int              decalage;           /* any number - see doc */ 
  int              year_tag;           /* the position of the year depends on the number of field */
  char             tstamp[200];
  char*            timep;
  time_t           v;
  struct timezones tzz;

  #ifdef DEBUG
  int i;
  #endif

  if (nbcar <= 0) { return NULL; }

  /***********************************************************/
  /* Split the date                                          */
  /* You can have 6 or 7 fields exactly.                     */
  /***********************************************************/


  nb_fields = s_split(date, " ", &f);

  #ifdef DEBUG
  fprintf (stdout, "\n\ndate2timestamp() nb_fields = %d\n", nb_fields); fflush(stdout);
  #endif

  if ((nb_fields != 7) && (nb_fields != 6))
  {
    free_fields (&f);
    timestamp[0] = 0;
    return NULL;
  }

  /***********************************************************/
  /* Find the correct timezone and daylight                  */
  /* So you get the time shift.                              */
  /***********************************************************/

  year_tag = YEAR;

  if (nb_fields == 7)
  {
    if (strlen(f.tabs[TAG1]) >= TIMEZONE_SIZE)
    {
      free_fields (&f);
      timestamp[0] = 0;
      return NULL;
    }

    if (strlen(f.tabs[TAG2]) >= DAYLIGHT_SIZE)
    {
      free_fields (&f);
      timestamp[0] = 0;
      return NULL;
    }

    strcpy (tzz.timezone, f.tabs[TAG1]);
    strcpy (tzz.daylight, f.tabs[TAG2]);
  }

  if (nb_fields == 6)
  {
    if (strlen(f.tabs[TAG1]) >= TIMEZONE_SIZE)
    {
      free_fields (&f);
      timestamp[0] = 0;
      return NULL;
    }

    strcpy (tzz.timezone, f.tabs[TAG1]);
    (tzz.daylight)[0] = 0;

    /* One field missing => the year tag is one tag before !!!!!! */

    year_tag = YEAR - 1;
  }

  #ifdef DEBUG
  fprintf (stdout, "\n\ndate2timestamp(): tzz done\n"); fflush(stdout);
  #endif

  /***********************************************************/
  /* lookup timezone configuration                           */
  /***********************************************************/

  current_tz = NULL;

  if (tzfound(&tzz) == 0)
  {
    free_fields (&f);
    timestamp[0] = 0;
    return NULL;
  }

  #ifdef DEBUG
  fprintf (stdout, "\n\ndate2timestamp(): tzfound() passed\n"); fflush(stdout);
  #endif

  /***********************************************************/
  /* Time zone configuration found - set global variable     */
  /* current_tz                                              */
  /***********************************************************/

  /* set into tzfound() ... */

  /***********************************************************/
  /* Set time shift according to timezone configuration      */
  /***********************************************************/

  if (shift == AUTOMATIC_TIME_SHIFT_COMPUTATION) { shift = tzz.shift; }

  /***********************************************************/
  /* Split the clock time                                    */
  /***********************************************************/

  timep = (char*)malloc((strlen(f.tabs[CLOCK])+1)*sizeof(char));
  if (timep == NULL)
  {
    free_fields (&f);
    timestamp[0] = 0;
    return NULL;
  }

  strcpy (timep, f.tabs[CLOCK]);

  nb_fields = s_split(timep, ":", &t);
  if (nb_fields != 3)
  {
    free_fields (&t);
    free_fields (&f);
    free (timep);
    timestamp[0] = 0;
    return NULL;
  }

  free (timep);

  #ifdef DEBUG
  fprintf (stdout, "\nDate:");
  for (i=0; i<f.number_of_fields; i++)
  { fprintf (stdout, "\n -> %s", f.tabs[i]); }
  fprintf (stdout, "\n"); fflush(stdout);
  #endif
 
  #ifdef DEBUG
  fprintf (stdout, "\nHour:");
  for (i=0; i<t.number_of_fields; i++)
  { fprintf (stdout, "\n -> %s", t.tabs[i]); }
  fprintf (stdout, "\n"); fflush(stdout);
  #endif
 
  /***********************************************************/
  /* Compute UTC timestamp                                   */
  /***********************************************************/

  seconds            = atoi(t.tabs[SECONDS]);
  minutes            = atoi(t.tabs[MINUTES]);
  hours              = atoi(t.tabs[HOUR]);
  day_of_the_month   = atoi(f.tabs[DAY_OF_THE_MONTH]);
  month              = get_month(f.tabs[MONTH]);
  year               = atoi(f.tabs[year_tag]) - 1900;
  day_of_the_week    = get_day(f.tabs[DAY_OF_THE_WEEK]);
  day_of_the_year    = 0;

  decalage           = 0; 

  tt.tm_sec   = seconds;
  tt.tm_min   = minutes;
  tt.tm_hour  = hours;
  tt.tm_mday  = day_of_the_month;
  tt.tm_mon   = month;
  tt.tm_year  = year;
  tt.tm_wday  = day_of_the_week;
  tt.tm_yday  = day_of_the_year;  /* unused */
  tt.tm_isdst = decalage;

  #ifdef DEBUG
  fprintf (stdout, "\nsec:               %d", seconds);
  fprintf (stdout, "\nmin:               %d", minutes);
  fprintf (stdout, "\nhour:              %d", hours);
  fprintf (stdout, "\nday_of_the_month:  %d", day_of_the_month);
  fprintf (stdout, "\nmonth:             %d", month);
  fprintf (stdout, "\nyear:              %d", year);
  fprintf (stdout, "\nday_of_the_week:   %d", day_of_the_week);
  fprintf (stdout, "\nday_of_the_year:   %d", day_of_the_year);
  fprintf (stdout, "\n");
  #endif

  v = mktime(&tt);

  /***********************************************************/
  /* Add time shift                                          */
  /***********************************************************/

  sprintf (tstamp, "%lu", (unsigned long int)v + shift);

  /***********************************************************/
  /* Copy the result into output buffer and then clean all   */
  /***********************************************************/

  if (nbcar-1 < strlen(tstamp))
  { 
    free_fields (&t);
    free_fields (&f);
    timestamp[0] = 0;
    return NULL;
  }

  strcpy (timestamp, tstamp);

  free_fields (&f);
  free_fields (&t);
  return timestamp;
}

/*! \example test_date2timestamp.c
    This file shows how to use the function 'test_date2timestamp()'.
 */

/*! \brief Return the current date as Apache's standard.
    \return The function returns a pointer to a string of characters that contains the current date.
    \warning The returned pointer points to a static string of characters. Therefore you do NOT have to free it (do NOT free
             it!).
 */

char* dater()
{
  static char  strtime[MAX_TIMER_SIZE];
  time_t       tt;

  tt = time(NULL);
  strftime (strtime, 256, "%Y-%m-%d %k:%M:%S", localtime(&tt));

  return strtime;
}

/*! \example test_date2timestamp.c
    This file shows how to use the date functions.
 */


/*! \brief Current line number being parsed by the time zones reference parsed.
 */
int                ttref_line_num = 0;

/*! \brief Data structure that contains all the couples (date, timestamp) used as reference.
 */ 
struct ttref_tab   ttrefs;

/*! \brief Lexical parser build by the flex utility (for parsing the timestamps references configuration file).
 */
extern int ttreflex();

/*! \brief Pointer to the stream descriptor used by the timestamps references parser.
 */
extern FILE* ttrefin;


/*! \brief Check the time zone configuration data against well known couple (date, timestamps).
    \param config_file Pointer to a zero terminated string of characters that represents the name of the timestamps
           reference configuration file. This file is a list of couple (date, timestamps).
    \param auto_conf Timezone autoconfiguration flag. This value can be:
           <UL>
              <li>TZ_AUTOCONF_ON: Auto configuration mode requested.
              <li>TZ_AUTOCONF_OFF: Auto configuration mode not requested.
           </UL>
    \return If the time zone configuration is OK, then the function returns TTREF_OK. Otherwise the function may return
            one of the following codes:
            <UL>
              <li>TTREF_TOO_MANY_REF; Too many references in the timestamps reference config file.
              <li>TTREF_NO_MEM: Not enough memory.
              <li>TTREF_SYNTAX_ERROR: Syntax error found in the timestamps reference config file.
              <li>TTREF_DATE_TOO_LONG: Date too long in the timestamps reference config file.
              <li>TTREF_REF_TOO_LONG: timestamp too long in the timestamps reference config file.
              <li>TTREF_CAN_NOT_OPEN_CONF_FILE Can not open the timestamps reference config file.
              <li>TTREF_INVALID_DATE: Invalid date format in the timestamps reference config file.
              <li>TTREF_KO: Invalid time zone configuration - check the time zone configuration file!
            </UL>
    \warning You must call load_timezones() before you can call this function!.
 */

int test_tz_conf (char* config_file, int auto_conf)
{
  int    cr;
  int    i;
  char   *pt;
  char   tt[TTREF_MAX_TIMESTAMP_SIZE];
  char   ref_date[TTREF_MAX_DATE_SIZE];

  time_t ref, conf;
  int    err, correction;

  /***********************************************************/
  /* Make sure we have not already loaded it !               */
  /***********************************************************/

  if (ttrefs.entry_number == 0)
  {

    /***********************************************************/
    /* Open the timestamps references parser's input file      */
    /***********************************************************/

    ttrefin = fopen(config_file, "r");
    if (ttrefin == NULL) { return TTREF_CAN_NOT_OPEN_CONF_FILE; }

    /***********************************************************/
    /* Load timestamps references                              */ 
    /***********************************************************/

    cr = ttreflex();
    if (cr != 0) { fclose (ttrefin); return cr; }

    fclose (ttrefin); 
  }

  /***********************************************************/
  /* Now compute known timestamps                            */
  /***********************************************************/

  for (i=0; i<ttrefs.entry_number; i++)
  {
    strcpy (ref_date, ((ttrefs.ttref)[i]).date);

    pt = date2timestamp(ref_date, tt, TTREF_MAX_TIMESTAMP_SIZE, AUTOMATIC_TIME_SHIFT_COMPUTATION);
    if (pt == NULL) { return TTREF_INVALID_DATE; }

    #ifdef DEBUG
    fprintf (stdout, "\n-> %s", current_tz->timezone);
    fprintf (stdout, "\n-> %s", current_tz->daylight);
    fprintf (stdout, "\n-> %d", current_tz->shift);
    fprintf (stdout, "\n");
    fflush (stdout);
    #endif

    if (strcmp(tt, ((ttrefs.ttref)[i]).timestamp) != 0)
    {
      if (auto_conf == TZ_AUTOCONF_OFF) { return TTREF_KO; }

      /*******************************************************/
      /* Time zone auto configuration mode activated ...     */
      /* Compute the correct time shift according to the     */
      /* couples (date, timestamp) considered as references. */
      /*******************************************************/

      ref         = string2time_t(((ttrefs.ttref)[i]).timestamp, &err);
      conf        = string2time_t(tt, &err);
      correction  = (int)((unsigned int)ref - (unsigned int)conf);

      current_tz->shift += correction;    
    }
  }

  return TTREF_OK; 
}

/*! \example test_load_ttref.c
    This file shows how to use test_tz_conf().
 */

/*! \brief Return the number of seconds since 00:00:00 UTC, January 1, 1970.
    \return The function returns the number of seconds since 00:00:00 UTC, January 1, 1970.
 */

time_t get_utc_timestamp ()
{
  time_t tt;

  /***********************************************************/
  /* Get the number of seconds since 00:00:00 UTC, January   */
  /* 1, 1970                                                 */
  /***********************************************************/

  tt = time(NULL);
  return tt;
}

/*! \example test_get_utc_timestamp.c
    This file shows how to use the function get_utc_timestamp ().
 */

/*! \brief Format a TSM date (ex: &quot;Tue Sep 17 11:48:39 MET 2002&quot;) from a given timestamp.
    \param timestamp Number of seconds since 00:00:00 UTC, January 1, 1970 (as returned by the POSIX
           function time()).
    \param debug Debug flag (0: no debug, otherwize debug output).
    \return Upon successful completion, the function returns a pointer to a string of characters that
            represents the date associated with the given timestamp. This date respects the TSM's format.
            If an error occured, the function returns the value NULL.
    \warning The returned pointer points to a static memory area (declared whithin the function). Therefore
             you do not have to free it!
 */

char *get_tsm_date (time_t timestamp, int debug)
{
  static char  date[MAX_TSM_DATE_SIZE];
  static char  current_tz[TZSET_MAX_SIZE];
  static char  original_tz[TZSET_MAX_SIZE];
  char         buffer[TZSET_MAX_SIZE];
  char         *s;
  struct tm    *dd;

  

  if (debug)
  {
    s = getenv("TZ");

    fprintf (stdout, "\n------------------------");
    fprintf (stdout, "\nInitial timezone setting:\n");
    fflush (stdout);

    if (s != NULL) { fprintf (stdout, "\n       - TZ:         %s", s); fflush (stdout);}
    else           { fprintf (stdout, "\n       - TZ:         NULL pointer (getenv(\"TZ\"))"); fflush (stdout); }

    s = tzname[0];
    if (s != NULL) { fprintf (stdout, "\n       - tzname[0]:  %s", s); fflush (stdout);}
    else           { fprintf (stdout, "\n       - tzname[0]:  NULL pointer"); fflush (stdout); }

    s = tzname[1];
    if (s != NULL) { fprintf (stdout, "\n       - tzname[1]:  %s", s); fflush (stdout);}
    else           { fprintf (stdout, "\n       - tzname[1]:  NULL pointer"); fflush (stdout); }

    fprintf (stdout, "\n");
    fflush (stdout);
  }

  /* -------------------------------------------- */
  /* Save current timezone setting                */
  /* WARNING!! See "man putenv"                   */
  /* -------------------------------------------- */

  s = getenv("TZ");
  if (s == NULL) { return NULL; }
  strcpy (buffer, "TZ=");
  if (strlen(s) >= TZSET_MAX_SIZE-3) { return NULL; } 
  strcat (buffer, s);
  strcpy (original_tz, buffer);

  /* -------------------------------------------- */
  /* Now set current timezone to MET              */
  /* -------------------------------------------- */

  strcpy (current_tz, "TZ=MET");
  if (putenv(current_tz) != 0) { return NULL; }
  tzset();

  if (debug)
  {
    s = getenv("TZ");

    fprintf (stdout, "\nSetting process timezone to MET:\n");
    fflush (stdout);

    if (s != NULL) { fprintf (stdout, "\n       - TZ:         %s", s); fflush (stdout);}
    else           { fprintf (stdout, "\n       - TZ:         NULL pointer (getenv(\"TZ\"))"); fflush (stdout); }

    s = tzname[0];
    if (s != NULL) { fprintf (stdout, "\n       - tzname[0]:  %s", s); fflush (stdout);}
    else           { fprintf (stdout, "\n       - tzname[0]:  NULL pointer"); fflush (stdout); }

    s = tzname[1];
    if (s != NULL) { fprintf (stdout, "\n       - tzname[1]:  %s", s); fflush (stdout);}
    else           { fprintf (stdout, "\n       - tzname[1]:  NULL pointer"); fflush (stdout); }

    fprintf (stdout, "\n");
    fflush (stdout);
  }

  /* -------------------------------------------- */
  /* Format TSM's date                            */
  /* -------------------------------------------- */

  if (strcmp(tzname[0], "MET")     != 0) { return NULL; }
  if (strcmp(tzname[1], "MET DST") != 0) { return NULL; }

  dd = localtime(&timestamp);

  sprintf (date, "%s %s %02d %02d:%02d:%02d %s %d",
                 days[dd->tm_wday],
                 months[dd->tm_mon],
                 dd->tm_mday,
                 dd->tm_hour,
                 dd->tm_min,
                 dd->tm_sec,
                 (dd->tm_isdst ? "MET": "MET DST"),
                 1900 + dd->tm_year);

  /* -------------------------------------------- */
  /* Restore original timezone setting            */
  /* -------------------------------------------- */

  if (putenv(original_tz) != 0) { return NULL; } 
  tzset();

  if (debug)
  {
    s = getenv("TZ");

    fprintf (stdout, "\nRestore initial process timezone:\n");
    fflush (stdout);

    if (s != NULL) { fprintf (stdout, "\n       - TZ:         %s", s); fflush (stdout);}
    else           { fprintf (stdout, "\n       - TZ:         NULL pointer (getenv(\"TZ\"))"); fflush (stdout); }

    s = tzname[0];
    if (s != NULL) { fprintf (stdout, "\n       - tzname[0]:  %s", s); fflush (stdout);}
    else           { fprintf (stdout, "\n       - tzname[0]:  NULL pointer"); fflush (stdout); }

    s = tzname[1];
    if (s != NULL) { fprintf (stdout, "\n       - tzname[1]:  %s", s); fflush (stdout);}
    else           { fprintf (stdout, "\n       - tzname[1]:  NULL pointer"); fflush (stdout); }

    fprintf (stdout, "\n");
    fflush (stdout);
  }


  return date;
}

/*! \example test_get_tsm_date.c
    This example shows how to use the function get_tsm_date().
 */









