/*! \file dir.c
    This file contains various utilities to manipulate directory.
 */

#include <stdio.h>
#include "my_dir.h"


/*! \brief List the content of a directory (compatible SUN/LINUX).
    \param dir_name Pointer to a zero terminated string of characters that represents the name of the directory to list.
    \param flist Pointer to a 'file_list' structure that will be used to store directory's content.
    \return If the function succed, it returns a positive number that represents the number of entries in the directory.
            Oterwise, the function returns -1.
    \warning Some fields in the structure 'flist' has been allocated within the function. You must call free_file_list() 
             of 'flist' to free all the allocated memory.
 */

int read_directory (char *dir_name, struct file_list *flist)
{
  DIR           *dirp;
  struct dirent *dp;
  int           nb_file, total_file, fsize;

  flist->list        = NULL;
  flist->total_alloc = 0;
  flist->num         = 0;

  dirp = opendir(dir_name);
  if (dirp == NULL) { return -1; }

  total_file = FILE_NUMBER;
  flist->list = (char**)malloc(total_file*sizeof(char*));
  if (flist->list == NULL)
  {
    closedir(dirp);
    return -1;
  }
  flist->total_alloc = total_file;


  nb_file = 0;
  while ((dp = readdir(dirp)) != NULL)
  {
    if (nb_file >= total_file)
    {
      total_file *= 2;
      flist->list = (char**)realloc(flist->list, total_file*sizeof(char*));
      if (flist->list == NULL)
      {
        closedir(dirp);
        return -1;
      }
      flist->total_alloc = total_file;
    }
  
    fsize = strlen(dp->d_name) + 1 + strlen(dir_name); 
    (flist->list)[nb_file] = (char*)malloc((fsize+1)*sizeof(char));
    if ((flist->list)[nb_file] == NULL)
    {
      closedir(dirp);
      return -1;
    }

    sprintf ((flist->list)[nb_file], "%s/%s", dir_name, dp->d_name);
    nb_file++;
    flist->num = nb_file;
  }

  closedir(dirp);
  return nb_file; 
}

/*! \example test_dir.c
    this file shows how to use read_directory() and free_file_list().
 */

/*! \brief List the content of a directory, and keeps only the directories.
    \param dir_name Pointer to a zero terminated string of characters that represents the name of the directory
           to list.
    \param flist_out Pointer to a 'file_list' structure that will be used to store directory's content.
    \return If the function succed, it returns a positive number that represents the number of entries in the
            directory. Oterwise, the function returns -1.
    \warning Some fields in the structure 'flist' has been allocated within the function. You must call
             free_file_list() of 'flist' to free all the allocated memory.
 */

int read_directory_only (char *dir_name, struct file_list *flist_out)
{
  int              nb_files, i, fsize;
  struct file_list flist;
  struct stat      stat_buf;

  /* ------------------------------------------------------- */
  /* Read all the entries of the firectory                   */
  /* ------------------------------------------------------- */

  nb_files = read_directory (dir_name, &flist);
  if (nb_files == -1) { return -1; }

  /* ------------------------------------------------------- */
  /* Allocate memory for 'flist_out'. We know for sure that  */
  /* we have no more that 'nb_files' directories.            */
  /* ------------------------------------------------------- */

  flist_out->list = (char**)malloc(nb_files*sizeof(char*));
  if (flist_out->list == NULL)
  {
    free_file_list (&flist);
    return -1;
  }
  flist_out->total_alloc = nb_files;

  /* ------------------------------------------------------- */
  /* Now look at each entry to keep only the directories     */
  /* ------------------------------------------------------- */

  flist_out->num = 0;
  for (i=0; i<flist.num; i++)
  {
    if (stat ((flist.list)[i], &stat_buf) == -1)
    {
      free_file_list (&flist);
      return -1;
    }

    if (S_ISDIR(stat_buf.st_mode))
    {
      fsize = strlen((flist.list)[i]+1);
      
      (flist_out->list)[flist_out->num] = (char*)malloc((fsize)*sizeof(char));

      if ((flist_out->list)[flist_out->num] == NULL)
      {
        free_file_list (&flist);
        return -1;
      }

      strcpy ((flist_out->list)[flist_out->num], (flist.list)[i]); 

      flist_out->num += 1;
    }
  }

  /* ------------------------------------------------------- */
  /* Free all allocated memory                               */ 
  /* ------------------------------------------------------- */

  free_file_list (&flist);

  return flist_out->num;
}

/*! \example test_dir.c
    this file shows how to use read_directory() and free_file_list().
 */

/*! \brief Free all allocated memory after you called read_directory().
    \param flist Pointer to a structure 'file_list' that should have been filled by read_directory().
    \warning ALWAYS call this function after read_directory().
 */

void free_file_list(struct file_list *flist)
{
  int i;

  for (i=0; i<flist->num; i++)
  { free ((flist->list)[i]); }
  if (flist->total_alloc > 0) { free (flist->list); }
}

/*! \example test_dir.c
    this file shows how to use read_directory() and free_file_list().
 */


/*! \brief Extract the directory base name from a file name.
    \param dirname Pointer to a NULL terminated that represents the file name.
    \param res Pointer to a pointer that will receive the directory base name.
    \return If the function found the directory base name, then it returns BASE_DIR_FOUND and in this case 'res'
            contains the directory base name. If the function can not extract the directory base name it returns
            BASE_DIR_NOT_FOUND and in this case 'res' is a NULL pointer. If the function returns BASE_DIR_NO_MEM, it
            means that the function can not allocate memory (in this case 'res' is NULL).
    \warning 'res' has been allocated within the function. You must delete it yoursef (using free()). But before 
             you free it, make sure to check that 'res' is not NULL (if (res != NULL) { free (res); }).
 */

int base_dir_name (char *dirname, char **res)
{
  char   *c, save;
  int    size;

  *res = NULL;
  c = rindex (dirname, '/');
  if (c == NULL) { return BASE_DIR_NOT_FOUND; }

  save = *c;
  *c = 0;

  size = strlen(dirname);
  *res = (char*)malloc((size+1)*sizeof(char));
  if (*res == NULL)
  {
    *c = save;
    return BASE_DIR_NO_MEM;
  }

  strcpy (*res, dirname);
  *c = save;

  return BASE_DIR_FOUND;
}

/*! \example test_basename.c
    This file shows how to use the function base_dir_name().
 */

