/*! \file dmalloc.c
    This file implements the debug tools for malloc.
 */


#include "dmalloc.h"

#define DMALLOC_ADDRESS_SIZE 100

/*! \brief Allocate memory (using malloc), and log a message into a log file.
    \param size Number of bytes to allocate.
    \param message Pointer to a zero terminated string of characters that contains the message to log.
    \param file Name of the log file.
    \return This function returns the pointer allocated (via malloc), or NULL if the system is running
            out of memory.
 */

void* dmalloc(size_t size, char *message, char *file)
{
  void *p;
  char *iobuff, address[DMALLOC_ADDRESS_SIZE];
  int  buff_size;

  p = malloc (size);
  if (p == NULL) { return NULL; }

  snprintf (address, DMALLOC_ADDRESS_SIZE-1, "\n%X", (unsigned int)p);

  buff_size = strlen(address)+strlen(": ")+strlen(message)+1;

  iobuff = (char*)malloc(buff_size);
  if (iobuff == NULL) { return NULL; }

  my_logger (
              file,
              iobuff,
              buff_size,
              "sss",
              address,
              ": ",
              message
            );

  free (iobuff);
  return p;
}

/*! \example test_dmalloc.c
    This file shows how to use the functions of the dmalloc library.
 */

/*! \brief Free allocated memory (using free), and log a message into a log file.
    \param p Pointer to the start of the memory location to free.
    \param message Pointer to a zero terminated string of characters that contains the message to log.
    \param file Name of the log file.
    \return The function returns 0 if success, otherwise the function returns 1 (this means that the system
            is running out of memory).
 */

int dfree(void* p, char *message, char *file)
{
  char *iobuff, address[DMALLOC_ADDRESS_SIZE];
  int  buff_size;

  snprintf (address, DMALLOC_ADDRESS_SIZE-1, "\n%X", (unsigned int)p);
  free (p);
  buff_size = strlen(address)+strlen(": ")+strlen(message)+1;

  iobuff = (char*)malloc(buff_size);
  if (iobuff == NULL) { return 1; }

  my_logger (
              file,
              iobuff,
              buff_size,
              "sss",
              address,
              ": ",
              message
            );

  free (iobuff);

  return 0;
}

/*! \example test_dmalloc.c
    This file shows how to use the functions of the dmalloc library.
 */
