/*! \file dstring.c
    This file implements dynamic string manipulation functions.
 */

#include <stdlib.h>
#include <string.h>
#include "dstring.h"


#include <stdio.h>

/*! \brief Initialize a dynamic string.
    \param s Pointer to a given dynamic string.
    \param init_size Initial size of the dynamic string.
    \return Upon successful completion, the function returns 0. Otherwise the function returns 1 (this means
            that the process is running out of memory.
    \warning You <b>MUST</b> initialize a dynamic string (using dstring_init()), before you can do any
             operation on it.
 */

int dstring_init (dstring *s, size_t init_size)
{
  s->data   = (char*)malloc(init_size);
  if (s->data == NULL) { return 1; }
  memset ((void*)s->data, 0, init_size);
  s->size   = init_size;
  s->nb_car = 0;

  return 0;
}

/*! \example test_dstring.c
    This file shows how to use the entire API for dynamic string manipulation.
 */

/*! \brief Free all memory allocated for a dynamic string.
    \param s Pointer to a given dynamic string.
 */

void dstring_free (dstring *s)
{
  free (s->data);
  s->data   = NULL;
  s->size   = 0;
  s->nb_car = 0;
}

/*! \example test_dstring.c
    This file shows how to use the entire API for dynamic string manipulation.
 */

/*! \brief Add characters into a given dynamic string.
    \param s Pointer to a given dynamic string.
    \param buff Pointer to a buffer that contains the characters to add to the dynamic string.
    \param nb_elem Number of characters to add.
    \return Upon successful completion, the function returns 0. Otherwise the function returns 1 (this means
            that the process is running out of memory.
    \warning You <b>MUST</b> initialize a dynamic string (using dstring_init()), before you can add data into it.
 */

int dstring_add (dstring *s, char *buff, size_t nb_elem)
{

  /* --------------------------------------------- */
  /* We may need to allocate memory                */
  /* --------------------------------------------- */

  while ((s->nb_car + nb_elem) > s->size)
  {     
    s->size = 2 * s->size;
    s->data = (char*)realloc((void*)s->data, s->size);
    if (s->data == NULL) { return 1; }
  }

  memcpy ((void*)(s->data + s->nb_car), (void*)buff, nb_elem);

  s->nb_car = s->nb_car + nb_elem;

  return 0;
}

/*! \example test_dstring.c
    This file shows how to use the entire API for dynamic string manipulation.
 */

/*! \brief Return the content of a dynamic string.
    \param s Pointer to a given dynamic string.
    \param size pointer to a <i>size_t</i> value that will be used to store the number of bytes extracted
           from the dynamic string.
    \return Upon successful completion, the function returns a pointer to a dynamically allocated buffer
            that contains the dynamic string's data. The value pointed by <i>size</i> represents the number
            of bytes allocated for the returned string (including the final 0). If the function could not
            allocate memory, then it returns NULL.
    \warning The returned pointer points to a memory location that has been allocated within the function.
             Therefore you should free it (using free()) when you don't need it anymore.
 */

char* dstring_get_data (dstring *s, size_t *size)
{
  char *st;

  *size = s->nb_car + 1;

  st = (char*)malloc(*size);
  if (st == NULL) { return NULL; }

  memcpy ((void*)st, (void*)s->data, s->nb_car);

  /* --------------------------------------------- */
  /* Add a zero at the end ...                     */
  /* --------------------------------------------- */

  *(st + s->nb_car) = 0;

  return st;
}

/*! \example test_dstring.c
    This file shows how to use the entire API for dynamic string manipulation.
 */

