/*! \file files.c
    This file contains all the file's related functions.
 */

#include "files.h"

/*! \brief Wait for a file to appear.
    \param file_name Name of the file.
    \param period The function test if the file exists periodically. If the file does not exist, then the process is put
           to sleep for a certain amount of time, and then the function tests the file existence again ... and so on. Between
           each test, the process is put to sleep for 'period' seconds.
    \param stop_file If the file that name is defined by 'stop_file' exists, then the function exists and returns 1.
    \return  If the file represented by 'file_name' exists, the function return 0. If the file represented by
             'stop_file' exists, the the function returns 1.
    \warning This function uses the system call access() to test the existence of a file. According to the documentation,
             this system call can be interrupted by a signal. But in this case the return value is -1, which is also the
             return value if the file does not exist. Therefore we do not have to perform any special action in case the
             system call is interrupted ... we just try again.
 */

int wait_for_file (char* file_name, int period, char* stop_file)
{
  while (access (file_name, F_OK) == -1)
  {
     if (stop_file != NULL)
     {
       if (access (stop_file, F_OK) == 0) { return 1; }
     }

     sleep (period);
  }
  return 0;
} 

/*! \example test_wait_for_file.c
    This file shows how to use the function wait_for_file().
 */


/*! \brief Wait for a file (within a list of files) to exist.
    \param file_names This variable represents the list of file names to test. It is an array of strings, each string 
           represents a file name. The last element of the list MUST be set to NULL.
    \param period The function test if the file exists periodically. If the file does not exist, then the process is put
           to sleep for a certain amount of time, and then the function tests the file existence again ... and so on. Between
           each test, the process is put to sleep for 'period' seconds.
    \param stop_file If the file that name is defined by 'stop_file' exists, then the function exists and returns 1.
    \return If one of the files specified in the list of files exists, the function returns a non NULL pointer that points
            a the name of the file. If the file represented by 'stop_file' exists, the the function returns NULL.
    \warning The last element of the list of file names MUST be set to NULL. If you want to wait on 3 files, then the
             variable 'file_names' must be an array of 4 elements: the 3 file names and the last NULL pointer.
 */

char* wait_for_multiple_file (char **file_names, int period, char* stop_file)
{
  char *name;
  int  pos;

  while (1==1)
  {
     pos  = 0;
     name = file_names[pos];

     while (name != NULL)
     {
       if (access (name, F_OK) == 0) { return name; }
       pos++;
       name = file_names[pos];
     }

     if (stop_file != NULL)
     { if (access (stop_file, F_OK) == 0) { break; } }

     sleep(period);
  }

  return NULL;
}

/*! \example test_wait_for_multiple_files.c
    This file shows how to use the function wait_for_multiple_file.
 */

/*! \brief Load the content of a file into an in-memory buffer.
    \param file_name Pointer to a zero terminated file that represents the name of the file to load..
    \param buff Pointer to a pointer that will be used to save the file's content.
    \return If the function succed, the function returns a positive interger that represents the number of bytes extracted
            from the file. If the function failes, then it returns a negative integer. This integer could be:
            <UL>
              <li>FILES_UTIL_MALLOC_ERR
              <li>FILES_UTIL_READ_ERR
            </UL>
    \warning The memory pointed by 'buff' has been allocated within the function, therefore you must free it yourself
             (using free()).
 */ 

int read_file (char *file_name, char **buff)
{
  int   c, lu, taille;
  FILE  *fd;
 
  fd = fopen (file_name, "r");
  if (fd == NULL) { return FILES_UTIL_OPEN_ERR; }

 
  taille = FILES_UTIL_INIT_BUFF_SIZE;
  *buff = (char*)malloc(taille);
  if (*buff == NULL)
  {
     fclose(fd);
     return FILES_UTIL_MALLOC_ERR;
  }

  lu=0;
  c = fgetc(fd);

  while (feof(fd) == 0)
  {
    if (c == EOF)
    {
       fclose(fd);
       return FILES_UTIL_READ_ERR;
    }
    lu++;
    
    while (lu > (taille-1))
    {
      taille *= 2;
      *buff = (char*)realloc((void*)(*buff), taille);
      if (*buff == NULL)
      { 
         fclose(fd);
         return FILES_UTIL_MALLOC_ERR;
      }
    }

    *((*buff) + (lu-1)) = (char)c;
    c = fgetc(fd);    
  }
  *((*buff) + lu) = '\0';

  fclose(fd);
  return lu;
}

/*! \example test_read_file.c
    This file shows how to use the function read_file().
 */

/*! \brief Set the timeout file that contains the current number of seconds since January 1970 00:00:00.
    \param filename Pointer to a zero terminated string of characters that represents the path to the file.
    \return The function returns one of the following values:
            <UL>
                <li>SET_TIMEOUT_FILE_OPEN_ERROR: Can not open the target file 'filename'.
                <li>SET_TIMEOUT_FILE_WRITE_ERROR: Error while writing in the target file.
                <li>SET_TIMEOUT_FILE_OK: The function succed.
            </UL>
 */

int set_timeout_file (char *filename)
{
  time_t  current_time;
  FILE    *fd;
  int     cr;

  fd = fopen(filename, "w");
  if (fd == NULL) { return SET_TIMEOUT_FILE_OPEN_ERROR; }

  current_time = time(NULL);

  cr = fwrite ((void*)&current_time, sizeof(time_t), 1, fd);
  if (cr != 1)
  {
    fclose (fd);
    return SET_TIMEOUT_FILE_WRITE_ERROR;
  }

  fclose (fd);
  return SET_TIMEOUT_FILE_OK;
}

/*! \example test_timeout_file.c
    This file shows how to use the functions set_timeout_file() and test_timeout_file().
 */

/*! \brief Look at a <i>timeout file</i> to see if the calling process should wait or go ahead.
    \param filename Pointer to a zero terminated string of characters that represents the timeout file.
    \param timeout Integer value that represents the timeout in seconds.
    \return The function returns one of the following values:
            <UL>
                <li>TEST_TIMEOUT_FILE_TIMEOUT_ELLAPSED: The timeout is passed.
                <li>TEST_TIMEOUT_FILE_READ_ERROR: An error occured while reading the timeout file.
                <li>TEST_TIMEOUT_FILE_TIMEOUT_NOT_ELLAPSED: The timeout is not passed yet.
                <li>TEST_TIMEOUT_FILE_UNLINK_FAILED: An error occured while removing the timeout file.
            </UL>
    \warning The timeout file should have been set by using the function set_timeout_file(). 
 */

int test_timeout_file (char *filename, int timeout)
{
  time_t  current_time, file_time, delta;
  FILE    *fd;
  int     cr;

  /* ============================================= */
  /* Try to open the timeout file - If it does not */
  /* exist, it means that there was no timeout ... */
  /* ============================================= */

  fd = fopen(filename, "r");
  if (fd == NULL) { return TEST_TIMEOUT_FILE_TIMEOUT_ELLAPSED; }

  /* ============================================= */
  /* Read timestamp written into the file          */
  /* ============================================= */

  cr = fread ((void*)&file_time, sizeof(time_t), 1, fd);
  if (cr != 1)
  {
    fclose (fd);
    return TEST_TIMEOUT_FILE_READ_ERROR;
  }
  current_time = time(NULL);

  /* ============================================= */
  /* Calculate the ellapsed time                   */
  /* ============================================= */

  delta = current_time - file_time;
  if (delta < timeout)
  { 
    fclose (fd);
    return TEST_TIMEOUT_FILE_TIMEOUT_NOT_ELLAPSED;
  }
 
  fclose (fd);

  /* ============================================= */
  /* Delete timeout file if the timeout expired    */
  /* ============================================= */

  if (unlink (filename) != 0) { return TEST_TIMEOUT_FILE_UNLINK_FAILED; }

  return TEST_TIMEOUT_FILE_TIMEOUT_ELLAPSED;
}

/*! \example test_timeout_file.c
    This file shows how to use the functions set_timeout_file() and test_timeout_file().
 */



/*! \brief Extract the base name of a given file name.
    \param src Pointer to a zero terminated string of characters that represents the name of the file.
    \param dst Pointer to a memory location that will be sued to store a zero terminated string of
               characters that is the base name of the file.
    \warning No test is done regarding the size of the memory chunck pointed by dst. It is your responsability
             to allocate enough memory for the resulting string of characters!
 */

void basename (char *src, char *dst)
{
  char  *pt;

  pt = strrchr (src, '/');
  if (pt == NULL) { pt = src; }
  else { pt += 1; }

  strcpy (dst, pt);
}


