/*! \file flock.c
    This file contains funtions used to lock/unlock files.
 */

#include "flock.h"
#include <stdio.h>

#define DEBUG_TRACE(x) fprintf (stderr, x)

/*! \brief Lock a file.
    \param fd File descriptor that identifies the file to lock.
    \return If the lock succed, the function returns 0. Otherwise the function returns 1.
 */

int lock_file(int fd)
{
  int cr;

  #ifdef SUN
    struct flock lck;
  #endif

  #ifdef LINUX
    cr = flock (fd, LOCK_EX);
    if (cr == -1)
    {
      DEBUG_TRACE("\n\nflock failed\n\n");
      return 1;
    }
  #endif

  #ifdef SUN
    lck.l_start   = 0;        /* from the beginning of thr file ... */
    lck.l_len     = 0;        /* to the end if the file.            */
    lck.l_type    = F_WRLCK;  /* exclusive lock                     */
    lck.l_whence  = SEEK_SET; /* lock from the beginning            */
    cr = fcntl (fd, F_SETLKW, &lck);
    if (cr == -1)
    {
      DEBUG_TRACE("\n\nfcntl failed\n\n");
      return 1;
    }
  #endif

  return 0;
}

/*! \brief Unlock a file.
    \param fd File descriptor that identifies the file to unlock.
    \return If the operation succed, the function returns 0. Otherwise the function returns 1.
 */

int unlock_file(int fd)
{
  int cr;

  #ifdef SUN
    struct flock lck;
  #endif

  #ifdef LINUX
    cr = flock (fd, LOCK_UN);
    if (cr == -1)
    {
      DEBUG_TRACE("\n\nflock failed\n\n");
      return 1;
    }
  #endif

  #ifdef SUN
    lck.l_start   = 0;        /* from the beginning of thr file ... */
    lck.l_len     = 0;        /* to the end if the file.            */
    lck.l_type    = F_UNLCK;  /* exclusive lock                     */
    lck.l_whence  = SEEK_SET; /* lock from the beginning            */
    cr = fcntl (fd, F_SETLKW, &lck);
    if (cr == -1)
    {
      DEBUG_TRACE("\n\nfcntl failed\n\n");
      return 1;
    }
  #endif

  return 0;
}


/*! \brief Maximum number of characters for an integer.
 */

#define INT_MAX_CHAR    32

/*! \brief Write data into the log file.
    \param file_name Pointer to a NULL terminated string of characters that contains the name of the log file.
           Please note that if file_name[0] is equal to 0, then the function does nothing. This may be used to
           manage debug logging.
    \param buffer Pointer to a memory location used by the function to format the data to print.
    \param buffsize Size of the memory location 'buffer' in bytes.
    \param format Pointer to a NULL terminated string of characters that describes the format of the output. 's' means
           'char*', 'd' means 'interger' and 'u' means 'unsigned int'.
    \return The function return LOG_OK if the operation succed. Otherwise the function returns LOG_BUFFER_OVERFLOW,
            LOG_WRITE_ERROR, LOG_OPEN_ERROR, LOG_CLOSE_ERROR or LOG_LOCK_ERROR.
 */


int my_logger (char *file_name, char *buffer, int buffsize, char *format, ...)
{
   int          flock_fd;
   int          cr;
   int          done;
   char         buff[INT_MAX_CHAR];

   va_list      ap;
   char         *s;
   int          d;
   unsigned int u;
  

   /*****************************/
   /* If the name of the file   */
   /* empty, then do nothing.   */
   /*****************************/

   if (file_name[0] == 0) { return LOG_OK; }
 
   /*****************************/
   /* Open and lock the file    */
   /*****************************/

   flock_fd = open (file_name, O_WRONLY | O_CREAT | O_APPEND, S_IRWXU);
   if (flock_fd == -1)
   {
     DEBUG_TRACE("\n\nopen failed\n\n");
     return LOG_OPEN_ERROR;
   }

   cr = lock_file (flock_fd);
   if (cr == 1)
   {
     DEBUG_TRACE("\n\nlock_file failed\n\n");
     close(flock_fd); return LOG_LOCK_ERROR;
   }

   /*****************************/
   /* Extract stuff to write    */
   /*****************************/

   va_start (ap, format);
   done      = 0;
   memset ((void*)buffer, 0, buffsize);

   while (*format)
   {
     switch  (*(format++))
     {
       case 's': {
            s = va_arg (ap, char *);
            if (s != NULL)
            {
              done += strlen(s);
              if (done > buffsize-1)
              {
                unlock_file (flock_fd);
                close (flock_fd);
                return LOG_BUFFER_OVERFLOW;
              }
              strcat (buffer, s);
            }
       }; break;

       case 'd': {
            d = va_arg (ap, int);
            sprintf (buff, "%d", d);
            done += strlen(buff);
            if (done > buffsize-1)
            {
              unlock_file (flock_fd);
              close (flock_fd);
              return LOG_BUFFER_OVERFLOW;
            }
            strcat (buffer, buff);
       }; break;

       case 'u': {
            u = va_arg (ap, unsigned int);
            sprintf (buff, "%u", u);
            done += strlen(buff);
            if (done > buffsize-1)
            {
              unlock_file (flock_fd);
              close (flock_fd);
              return LOG_BUFFER_OVERFLOW;
            }
            strcat (buffer, buff);
       }; break;
     }
   }

   /*****************************/
   /* Write data to log file    */
   /*****************************/

   cr = write (flock_fd, buffer, strlen(buffer));
   if (cr == -1)
   {
      unlock_file (flock_fd);
      close (flock_fd);
      return LOG_WRITE_ERROR;
   }

   /*****************************/
   /* Unlock and close the file */
   /*****************************/

   cr = unlock_file (flock_fd);
   if (cr == 1)
   {
     DEBUG_TRACE("\n\nunlock_file failed\n\n");
     close(flock_fd); return LOG_LOCK_ERROR;
   }

   cr = close (flock_fd);
   if (cr == -1)
   {
     DEBUG_TRACE("\n\nclose failed\n\n");
     return LOG_CLOSE_ERROR;
   }

   return LOG_OK;
}

/*! \example test_flock.c
    This file shows how to ise the function my_logger().
 */


/*! \brief Append a file to another - The destination file is locked.
    \param src Pointer to a zero terminated string of characters that represents the name of the source file.
    \param dst Pointer to a zero terminated string of characters that represents the name of the destination file.
    \return The function returns COPY_OK if the operation succed. Otherwise, the function may return:
            <UL>
              <li>COPY_SRC_OPEN_ERROR Can not open the source file.
              <li>COPY_DST_OPEN_ERROR Can not open the destination file.
              <li>COPY_LOCK_ERROR Lock on the destination file failed.
              <li>COPY_STAT_ERROR Can not stat() the srouce file.
              <li>COPY_READ_ERROR Error while reading the source file.
              <li>COPY_WRITE_ERROR Error while writing to the destination file.
              <li>COPY_UNLOCK_ERROR Error while unlocking the destinatination file.
              <li>COPY_SRC_CLOSE_ERROR Error while closing the source file.
              <li>COPY_DST_CLOSE_ERROR Error while closing the destination file.
            </UL>
 */

int append_file_with_lock (char *src, char *dst)
{
   int         src_fd, dst_fd;
   int         cr;
   off_t       fsize, i, nb_blcoks, reminder;
   char        iobuff[COPY_IOBUFF_SIZE];
   struct stat s_stat;
   size_t      rw;


   /**********************************/
   /* Open the source file           */
   /**********************************/
  
   src_fd = open (src, O_RDONLY, S_IRWXU);
   if (src_fd == -1)
   {
     return COPY_SRC_OPEN_ERROR;
   }

   /**********************************/
   /* Open the destination file and  */
   /* lock it.                       */
   /**********************************/

   dst_fd = open (dst, O_WRONLY | O_CREAT | O_APPEND, S_IRWXU);
   if (dst_fd == -1)
   {
     close (src_fd);
     return COPY_DST_OPEN_ERROR;
   }

   cr = lock_file (dst_fd);
   if (cr == 1)
   {
     close(src_fd);
     close(dst_fd);
     return COPY_LOCK_ERROR;
   }

   /**********************************/
   /* Get file's size                */
   /**********************************/

   cr = stat (src, &s_stat);
   if (cr != 0)
   {
     close(src_fd);
     unlock_file (dst_fd);
     close(dst_fd);
     return COPY_STAT_ERROR;
   }

   /**********************************/
   /* Calculate number of full       */
   /* block.                         */
   /**********************************/

   fsize = s_stat.st_size;
   nb_blcoks = fsize / COPY_IOBUFF_SIZE;
   reminder = fsize - (nb_blcoks*COPY_IOBUFF_SIZE);

   /**********************************/
   /* Copy full blocks               */
   /**********************************/

   for (i=0; i<nb_blcoks; i++)
   {
     rw = read (src_fd, (void*)iobuff, COPY_IOBUFF_SIZE);
     if (rw != COPY_IOBUFF_SIZE)
     {
       close(src_fd);
       unlock_file (dst_fd);
       close(dst_fd);
       return COPY_READ_ERROR;
     }

     rw = write (dst_fd, (void*)iobuff, COPY_IOBUFF_SIZE);
     if (rw != COPY_IOBUFF_SIZE)
     {
       close(src_fd);
       unlock_file (dst_fd);
       close(dst_fd);
       return COPY_WRITE_ERROR;
     }
   }

   /**********************************/
   /* Copy reminder                  */
   /**********************************/

   rw = read (src_fd, (void*)iobuff, (size_t)reminder);
   if (rw != (size_t)reminder)
   {
     close(src_fd);
     unlock_file (dst_fd);
     close(dst_fd);
     return COPY_READ_ERROR;
   }

   rw = write (dst_fd, (void*)iobuff, (size_t)reminder);
   if (rw != (size_t)reminder)
   {
     close(src_fd);
     unlock_file (dst_fd);
     close(dst_fd);
     return COPY_WRITE_ERROR;
   }

   /**********************************/
   /* Unlock destination file        */
   /**********************************/

   cr = unlock_file (dst_fd);
   if (cr == 1)
   {
     close(src_fd); 
     close(dst_fd);
     return COPY_UNLOCK_ERROR;
   }

   /**********************************/
   /* Close all files                */
   /**********************************/

   cr = close(src_fd);
   if (cr == -1)
   {
     close(dst_fd);
     return COPY_SRC_CLOSE_ERROR;
   }

   cr = close(dst_fd);
   if (cr == -1)
   {
     return COPY_DST_CLOSE_ERROR;
   }

   return COPY_OK;
}

/*! \example test_copy_lock.c
    This file shows how to use the function append_file_with_lock().
 */
