/*! \file issort.c
    This file implements sorting by insersion
 */


#include <stdlib.h>
#include <string.h>
#include "sort.h"




/*! \brief Sort data using the insertion method.
    \param donnee Pointer to the beginning of the array to sort.
    \param nb_elts Number of elements in the array.
    \param taille_elt Size in bytes of one element.
    \param compare User defined function used to compare two elements of the array.
           If you want to sort according to increasing values, then the function <i>compare()</i> should return:
           <UL>
              <li>1 if <i>cle1 &gt; cle2</i>
              <li>0 if <i>cle1 == cle2</i>
              <li>-1 if <i>cle1 &lt; cle2</i>
           </UL>
           If you want to sort according to decreasing values, then the function <i>compare()</i> should return:
           <UL>
              <li>1 if <i>cle1 &lt; cle2</i>
              <li>0 if <i>cle1 == cle2</i>
              <li>-1 if <i>cle1 &gt; cle2</i>
           </UL>
    \return Upon successful completion the function returns 0, otherwise it returns -1
    \warning This function is not suitable if you need to sort an inportant number of element.
 */

int issort (
              void *donnee,
              int  nb_elts,
              int  taille_elt, 
              int  (*compare)(const void *cle1, const void *cle2)
           )
{
  char  *tab = donnee;
  void  *cle;
  int   i, j;

  /* ------------------------------------------------------------ */
  /* Allocate memory                                              */
  /* ------------------------------------------------------------ */

  if ((cle = (char *)malloc(taille_elt)) == NULL) { return -1; }

  /* ------------------------------------------------------------ */
  /* Insert a key                                                 */
  /* ------------------------------------------------------------ */

  for (j = 1; j < nb_elts; j++)
  {
     memcpy(cle, &tab[j * taille_elt], taille_elt);
     i = j - 1;

    /* ---------------------------------------------------------- */
    /* Where to insert the key                                    */
    /* ---------------------------------------------------------- */

     while (i >= 0 && compare(&tab[i * taille_elt], cle) > 0)
     {
       memcpy(&tab[(i + 1) * taille_elt], &tab[i * taille_elt], taille_elt);
       i--;
     }

     memcpy(&tab[(i + 1) * taille_elt], cle, taille_elt);
  }

  /* ------------------------------------------------------------ */
  /* Free all allocated memory                                    */
  /* ------------------------------------------------------------ */

  free(cle);

  return 0;
}

/*! \example test_sort.c
    This file shows how to use the function issort().
 */

/*! \brief Extended version of the function <i>issort()</i>. This can be used to handle complex data.
    \param donnee Pointer to the beginning of the array to sort.
    \param nb_elts Number of elements in the array.
    \param compare User defined function used to compare two elements of the array.
           If you want to sort according to increasing values, then the function <i>compare()</i> should return:
           <UL>
              <li>1 if <i>cle1 &gt; cle2</i>
              <li>0 if <i>cle1 == cle2</i>
              <li>-1 if <i>cle1 &lt; cle2</i>
           </UL>
           If you want to sort according to decreasing values, then the function <i>compare()</i> should return:
           <UL>
              <li>1 if <i>cle1 &lt; cle2</i>
              <li>0 if <i>cle1 == cle2</i>
              <li>-1 if <i>cle1 &gt; cle2</i>
           </UL>
    \param copy_data User defined function used to copy data. <B>The function must take care of memory management.</B>
                     This function must return:
                     <UL>
                        <li>0 upon successful completion.
                        <li>1 if an error occured.
                     </UL>
    \param init_elem User defined function used to initialize an element. You can specify a NULL pointer if you do not need to initialize
                     elements. This function must return:
                     <UL>
                        <li>0 upon successful completion.
                        <li>1 if an error occured.
                     </UL>
                     This function is called whenever a new element is allocated.
    \return Upon successful completion the function returns 0, otherwise it returns -1
    \warning This function is not suitable if you need to sort an inportant number of element.
 */



int extended_issort (
                      void  *donnee,
                      int   nb_elts,
                      int   taille_elt,
                      int   (*compare)(const void *cle1, const void *cle2),
                      int   (*copy_data)(void *src, void *dest),
                      int   (*init_elem)(void *elm)
                    )
{
  char  *tab = donnee;
  void  *cle;
  int   i, j;

  /* ------------------------------------------------------------ */
  /* Allocate memory                                              */
  /* ------------------------------------------------------------ */

  if ((cle = (char *)malloc(taille_elt)) == NULL) { return -1; }
  if (init_elem != NULL) { if (init_elem(cle) == 1) { free(cle); return -1; } }

  /* ------------------------------------------------------------ */
  /* Insert a key                                                 */
  /* ------------------------------------------------------------ */

  for (j = 1; j < nb_elts; j++)
  {
     if (copy_data (&tab[j * taille_elt], cle) == 1) { free(cle); return -1; } 
     i = j - 1;

    /* ---------------------------------------------------------- */
    /* Where to insert the key                                    */
    /* ---------------------------------------------------------- */

     while (i >= 0 && compare(&tab[i * taille_elt], cle) > 0)
     {
       if (copy_data (&tab[i * taille_elt], &tab[(i + 1) * taille_elt]) == 1) { free(cle); return -1; }
       i--;
     }

     if (copy_data (cle, &tab[(i + 1) * taille_elt]) == 1) { free(cle); return -1; }
  }

  /* ------------------------------------------------------------ */
  /* Free all allocated memory                                    */
  /* ------------------------------------------------------------ */

  free(cle);

  return 0;
}

/*! \example test_ext_sort.c
    This file shows how to use the function extended_issort().
 */

