/*! \file list.c
    This code implements linked lists.
 */


#include <stdlib.h>
#include <string.h>
#include "list.h"

/*! \brief Initialize the linked list. This function should be called first.
    \param liste Pointer to a linked list structure (that will be initialized).
    \param detruire Pointer to a function that will be called to free the memory used to store one link of the list. Please note that:
           <UL>
              <li>This function depends on the type of data stored into the linked list.
              <li><i>detruire</i> could be a NULL pointer (if user's data are not dtnamically allocated).
           </UL>
    \param corresp Pointer to a function that will be called to compare to elements. Please note that this function depends on the type
                   of data stored into the linked list. The function must returns 1 if the elements' datas are identical; 0 otherwise.
 */

void list_init(List *liste, void (*detruire)(void *donnee), int (*corresp)(const void *val1, const void *val2))
{
  liste->taille   = 0;
  liste->detruire = detruire;
  liste->corresp  = corresp;
  liste->tete     = NULL;
  liste->queue    = NULL;

  return;
}

/*! \example test_list.c
    This file shows how to use the function list_init().
 */

/*! \brief Destroy the linked list. This function free all allocated memory used by the list.
    \param liste Pointer to a linked list structure (that you need to destroy).
 */

void list_destroy(List *liste)
{
  void  *donnee;

  while (list_size(liste) > 0)
  {
    if ((list_rem_next(liste, NULL, (void **)&donnee) == 0) && (liste->detruire !=  NULL))
    { liste->detruire(donnee); }
  }

  memset(liste, 0, sizeof(List));

  return;
}

/*! \example test_list.c
    This file shows how to use the function list_destroy().
 */

/*! \brief Insert a new element into the linked list.
    \param liste Pointer to a linked list structure.
    \param element Pointer to an (existing) element of the linked list. The new element will be inserted <B>AFTER</B> the element pointed
                   by <i>element</i>.
    \param donnee Pointer to the data to insert. Please note that the data is <B>NOT</B> copied into the linked list.
    \return Upon successful completion the function return the value 0. Otherwise the function returns -1.
    \warning The inserted data is <B>NOT</B> copied into the linked list! Therefore you should not free it unless you know exactly what you
             are doing.
 */

int list_ins_next(List *liste, ListElmt *element, const void *donnee)
{
  ListElmt *nouv_element;

  /* --------------------------------------------------- */
  /* Allocate memory for the new element                 */
  /* --------------------------------------------------- */

  if ((nouv_element = (ListElmt *)malloc(sizeof(ListElmt))) == NULL) { return -1; }

  /* --------------------------------------------------- */
  /* Set new element's data                              */
  /* --------------------------------------------------- */

  nouv_element->donnee = (void *)donnee;

  if (element == NULL)
  {
    /* ------------------------------------------------- */
    /* Insert to the beginning of the list               */
    /*   - in this case there is no previous element.    */
    /*   - the new element becomes the head of the list  */
    /* ------------------------------------------------- */

     if (list_size(liste) == 0) { liste->queue = nouv_element; }

     nouv_element->suivant   = liste->tete;

     #ifdef DOUBLE_LINK
       nouv_element->precedent = NULL;
     #endif

     liste->tete = nouv_element;
  }
  else
  {
    /* ------------------------------------------------- */
    /* Somewhere into the list, but not the beginning    */
    /* ------------------------------------------------- */

     if (element->suivant == NULL) { liste->queue = nouv_element; }

     nouv_element->suivant = element->suivant;

     #ifdef DOUBLE_LINK
       nouv_element->precedent = element;
     #endif

     element->suivant = nouv_element;
  }

  /* --------------------------------------------------- */
  /* increment the number of elements of the list        */
  /* --------------------------------------------------- */

  liste->taille++;

  return 0;
}

/*! \example test_list.c
    This file shows how to use the function list_ins_next().
 */


/*! \brief Remove an element from the linked list.
    \param liste Pointer to a linked list structure.
    \param element Pointer to an (existing) element of the linked list. The element to delete is the element <B>AFTER</B> the element
                   pointed by <i>element</i>. If you want to remove the first element of the list, then you must specify NULL as second
                   argument for the function list_rem_next().
    \param donnee Pointer that vill point to the data stored by the removed element.
    \return Upon successful completion the function return the value 0. Otherwise the function returns -1.
    \warning This function does <B>NOT</B> free the element's data (it only frees the element's structure). Therefore you should use
             <i>donnee</i> to free the element's data yourself.
 */

int list_rem_next(List *liste, ListElmt *element, void **donnee)
{
  ListElmt  *ancien_element;

  /* --------------------------------------------------- */
  /* Make sure that the list is not empty                */
  /* --------------------------------------------------- */

  if (list_size(liste) == 0) { return -1; }

  /* --------------------------------------------------- */
  /* Removing the element                                */
  /* --------------------------------------------------- */

  if (element == NULL)
  {
    /* ------------------------------------------------- */
    /* Removing at the beginning of the list             */
    /*   - The second element becomes the first (which   */
    /*     has no previous element now).                 */
    /* ------------------------------------------------- */

    *donnee = liste->tete->donnee;
    ancien_element = liste->tete;
    liste->tete = liste->tete->suivant;

    if (list_size(liste) == 1) { liste->queue = NULL; }

    #ifdef DOUBLE_LINK
      else { liste->tete->precedent = NULL; }
    #endif
  }
  else
  {
    /* ------------------------------------------------- */
    /* Somewhere into the list, but not the beginning    */
    /* (And not at the end of the list of course).       */
    /* ------------------------------------------------- */

    if (element->suivant == NULL) { return -1; }

    *donnee = element->suivant->donnee;
    ancien_element = element->suivant;
    element->suivant = element->suivant->suivant;

    if (element->suivant == NULL) { liste->queue = element; }

    #ifdef DOUBLE_LINK
      else { element->suivant->precedent = element; }
    #endif
  }

  /* --------------------------------------------------- */
  /* Free the allocated memory (for the element)         */
  /* --------------------------------------------------- */

  free(ancien_element);

  /* --------------------------------------------------- */
  /* Decrement the number of element of the list         */
  /* --------------------------------------------------- */

  liste->taille--;

  return 0;
}

/*! \example test_list.c
    This file shows how to use the function list_rem_next().
 */


/*! \brief Search an element into the linked list.
    \param liste Pointer to the linked list structure.
    \param from Pointer to the linked list's element where the search begins. If 'point' in NULL, then the search begins at the beginning
                of the linked list.
    \param data Pointer to a data a user defined structure that contains the data to search for.
    \param prec Pointer to a pointer that will receive the address of the element before the found element.
    \return <UL>
              <li>If the function finds an element that matches <i>data</i>, then it returns a pointer to this element. In this case,
                  <i>prec</i> points to the previous element (unless the found element was the first element in the linked list ... in this
                  the value of <i>prec</i> is NULL).
              <li>If there is no element that matches <i>data</i>, then the function returns NULL.
            </UL>
 */

ListElmt* list_find (List *liste, ListElmt *from, void* data, ListElmt **prec)
{
  *prec = NULL; 

  /* --------------------------------------------------- */
  /* If 'from' is NULL pointer, then start the search    */
  /* from the beginning of the linked list.              */
  /* --------------------------------------------------- */

  if (from == NULL) { from = list_head(liste); }

  /* --------------------------------------------------- */
  /* Compare each element with 'data'                    */
  /* --------------------------------------------------- */

  while (from != NULL)
  {
    if (liste->corresp(data, list_data(from))) { return from; }
    *prec = from;
    from = list_next(from);
  }

  return NULL;
}

/*! \example test_list.c
    This file shows how to use the function list_find().
 */


