/*! \file option.c
    This file implements the options' reader.
 */

#include <strings.h>
#include <sys/types.h>
#include <regex.h>
#include "option.h"


#ifdef LINUX
#include <stdio.h>
#endif

/*! \brief Pointer used to set the argument of the options that do not require arguments (ex: --help).
 */

static char  self_option[] = "";

/*! \brief Returns a pointer to the <I>Option</I> data structure that represents a given option.
    \param opt Pointer to a <I>Option_set</I> data structure that defines all options.
    \param flag Pointer to a zero terminated string of characters that represents the option's flag to 
                locate.
    \return If the option represented by the string pointed by <I>flag</I> is found, then the function returns
            a non NULL pointer to the <I>Option</I> data structure that defines the option. Otherwize (the
            option was not found), the function returns the value NULL.
 */

static Option* find_option (Option_set *opt, char *flag)
{
  Option *opt_found;
  int    i;

  opt_found = opt->opt;
  for (i=0; i<opt->count; i++)
  {
    if (strcmp(opt_found->flag, flag) == 0) { return opt_found; }
    opt_found++;
  }

  return NULL;
}

/*! \brief Returns a pointer to the argument associated with a given option.
    \param opt Pointer to a <I>Option_set</I> data structure that defines all options.
    \param flag Pointer to a zero terminated string of characters that represents the option's flag that you
                want to get the argument.
    \param arg If the option is valid (defined in the list of options), then the content of <I>arg</I> pointes
               to the argument associated with the option. If the content of <I>arg</I> is equal to NULL, this
               means that the option is not activated in the command line.
    \return If the option is found (that is: the option is defined in the <I>Option_set</I> data structure
            pointed by <I>opt</I>), the function returns the value 0. In this case:
            <UL>
                <LI>if <I>arg</I> is NULL, it means that the option was not activated.
                <LI>if <I>arg</I> is not NULL, it means that the option was activated.
                    <UL>
                        <LI>If the option requires one argument, then <I>arg</I> pointes to a zero terminated
                            string of characters that represents the argument.
                        <LI>If the option does not require any argument, then you don't care about the string
                            of characters pointed by <I>arg</I> (this string exists, but it does not mean
                            anything).
                    </UL>
            </UL>
            If the option can not be found (the option is not defined in the <I>Option_set</I> data structure
            pointed by <I>opt</I>), the function returns the value 1. Please note that a return value of 1 means
            that you've made an programming error (in your code, you are testing for options that you did not
            define earlier).
 */

int get_argument (Option_set *opt, char *flag, char **arg)
{
  Option *opt_found;

  *arg = NULL;
  opt_found = find_option (opt, flag);

  if (opt_found == NULL) { return 1; }
  *arg = *(opt_found->pointer);

  return 0;
}

/*! \example test_options.c
    This file shows how to use all the function of the library libmy_option.
 */

/*! \brief This function returns a value that represents the number of arguments (0 or 1 argument) required by
           for a given option.
    \param opt Pointer to a <I>Option_set</I> data structure that contains the definitions of all options.
    \param flag Pointer to a zero terminated string of characters that represents the option's flag to test.
    \return The function may return one of the following values:
            <UL>
               <LI>-1: An error occured. The option's flag pointed by <I>flag</I> does not represent any
                       pre-defined options. This is a programming error. You are trying to test an option that
                       you did not define previously.
               <LI>0:  The option does not require any argument.
               <LI>1:  The option requires one argument.
            </UL>
 */

int is_argument_required (Option_set *opt, char *flag)
{
  Option *opt_found;

  opt_found = find_option (opt, flag);
  if (opt_found == NULL) { return -1; }

  if (opt_found->arg != 0) { return 1; }

  return 0;
}

/*! \example test_options.c
    This file shows how to use all the function of the library libmy_option.
 */

/*! \brief Initialize the array of option definitions.
    \param opt Pointer to a <I>Option_set</I> data structure that contains the definitions of all options.
    \return Upon successful completion the function returns the value 0. Otherwise it returns the value 1.
    \warning Please note that an option's flag should verify the following rules:
             <UL>
                <LI>At least 2 characters.
                <LI>Begin with a '-'.
                <LI>No spaces.
                <LI>Not all '-'.
             </UL>
*/

int init_options (Option_set *opt)
{
  int     i, j, s;
  Option  *o;
  char    *pt;

  /* ------------------------------------------------------------- */
  /* Initialize options' array                                     */
  /* ------------------------------------------------------------- */

  opt->last_index = -1;
  for (i=0; i<opt->count; i++) { *((opt->opt + i)->pointer) = NULL; }

  /* ------------------------------------------------------------- */
  /* Check the syntax of all option's tags                         */
  /*  - At least 2 characters.                                     */
  /*  - Begin with a '-'.                                          */
  /*  - No spaces.                                                 */
  /*  - Not all '-'.                                               */
  /*                                                               */
  /* Note: the algo is not fast, but who cares. Command line is    */
  /*       not very long and we parse it only once. At least, this */
  /*       algo is pretty simple.                                  */
  /* ------------------------------------------------------------- */

  for (i=0; i<opt->count; i++)
  {
    o  = opt->opt + i;
    pt = o->flag;
    s  = strlen(pt);

    if ((s < 2) || (*pt != '-')) { return 1; }
    for (j=0; j<s; j++) { if (*(pt+j) == ' ') { return 1; } }
    for (j=0; j<s; j++) { if (*(pt+j) != '-') { break;    } }
    if (j == s) { return 1; }
  }

  return 0;
}

/*! \example test_options.c
    This file shows how to use all the function of the library libmy_option.
 */

/*! \brief Parse the command line and set up the option's data structure.
    \param argv Array of sero terminated strings of characters that represents the command line. Please
                note that the first element of the array (indice 0) should point to the name of the process.
    \param argc Number of elements in the array <I>argv</I>.
    \param opt Pointer to the <I>Option_set</I> data structure that contains the option's definitions.
    \param err Pointer to a pointer that will be used to store the address of a zero terminated string of 
               characters. If an error occured, this string contains a literal description of the error.
    \param flag Pointer to a pointer that will be used to store the address of a zero terminated string of
                characters. If an error occured, this string contains the option that generated the error.
    \param arg Pointer to a pointer that will be used to store the address of a zero terminated string of
               characters. If an error occured, this string contains the argument that caused the error.
    \return Upon successful completion, the function returns the value 0. Otherwise the function returns the
            value 1.
    \warning A valid command line should look something like:<P>
             &lt;process name&gt; [-flag1 [arg1]] [-flag2 [arg2]] [...] &lt;required arg 1&gt; ...
 */

int parse_command_line (char *argv[], int argc, Option_set *opt, char **err, char **flag, char **arg)
{
  int         i, oflag, rv;
  static char err1[]     = "Success";
  static char err2[]     = "";
  static char err3[]     = "Unknown option";
  static char err4[]     = "Missing option's argument";
  static char err5[]     = "Duplicated option";
  static char err6[]     = "Argument too long";
  static char err7[]     = "Invalid regular expression";
  static char err8[]     = "Invalid argument (syntax error)";
  static char last_opt[] = "";
  regex_t     reg;
  regmatch_t  pmatch[1];
  Option      *ofound;
  char        *last_option;

  /* ------------------------------------------------------------- */
  /* Initialize variables                                          */
  /* ------------------------------------------------------------- */

  *err            = err1;
  *flag           = err2;
  *arg            = self_option;
  last_option     = last_opt;
  oflag           = 0;
  opt->last_index = 0;

  /* ------------------------------------------------------------- */
  /* Parse command line now                                        */
  /* ------------------------------------------------------------- */

  for (i=1; i<argc; i++)
  {
    if ((*(argv[i]) != '-') && (oflag == 0)) { return 0; }

    /* ----------------------------------------------------------- */
    /* Is it an option ?                                           */
    /* ----------------------------------------------------------- */

    if (*(argv[i]) == '-')
    {
      if (oflag == 1)
      {
        /* ------------------------------------------------------- */
        /* We get an option's flag, but we expected an argument    */
        /* ------------------------------------------------------- */

        *err  = err4;
        *flag = last_option;
        return 1;
      } 

      ofound = find_option (opt, argv[i]);
      if (ofound == NULL)
      {
        /* ------------------------------------------------------- */
        /* This option is bot defined                              */
        /* ------------------------------------------------------- */

        *err  = err3;
        *flag = argv[i];
        return 1;
      }

      if (*(ofound->pointer) != NULL)
      {
        /* ------------------------------------------------------- */
        /* This option is already activated                        */
        /* ------------------------------------------------------- */

        *err  = err5;
        *flag = argv[i];
        return 1;
      }

      /* --------------------------------------------------------- */
      /* Option found. Is it a "self option" ?                     */
      /* --------------------------------------------------------- */

      if (ofound->arg == 0) { oflag = 0; *(ofound->pointer) = self_option; }
      else { oflag = 1; }

      last_option = argv[i];
    }

    /* ----------------------------------------------------------- */
    /* Is it an argument ?                                         */
    /* ----------------------------------------------------------- */

    if (*(argv[i]) != '-')
    {
      if (oflag == 0) { return 0; }

      /* --------------------------------------------------------- */
      /* Testing the size of the argument                          */
      /* --------------------------------------------------------- */

      if (strlen(argv[i]) > ofound->max_size)
      {
        *err  = err6;
        *arg  = argv[i];
        *flag = last_option;
        return 1;
      }

      /* --------------------------------------------------------- */
      /* Check argumment against regular expression                */
      /* --------------------------------------------------------- */

      rv = regcomp (&reg, ofound->reg_exp, REG_EXTENDED);
      if (rv != 0)
      {
        *err  = err7;
        *flag = last_option;
        *arg  = argv[i];
        return 1;
      }

      rv = regexec (&reg, argv[i], 1, pmatch, 0);
      if (rv != 0)
      {
        *err  = err8;
        *flag = last_option;
        *arg  = argv[i];
        regfree (&reg);
        return 1;
      } 

      regfree (&reg);

      *(ofound->pointer) = argv[i];
      oflag = 0;
    }

    opt->last_index = i;
  }

  if (oflag == 1)
  {
    *err  = err4;
    *flag = last_option;
    return 1;
  }

  return 0;
}

/*! \example test_options.c
    This file shows how to use all the function of the library libmy_option.
 */



