/*! \file qksort.c
    This file implements the quick sort algorithm
 */

#include <stdlib.h>
#include <string.h>

#include <stdio.h>

#include "sort.h"

/* ------------------------------------------------------------ */
/* Compare two integers -- Not part of public API               */
/* ------------------------------------------------------------ */

static int compare_int(const void *int1, const void *int2)
{
  if (*(const int *)int1 > *(const int *)int2) { return 1; }
  if (*(const int *)int1 < *(const int *)int2) { return -1; }
  return 0; 
}

/* ------------------------------------------------------------ */
/* Partitions -- Not part of public API                         */
/* ------------------------------------------------------------ */

static int partition (
                       void *donnee,
                       int taille_elt,
                       int i,
                       int k, 
                       int (*compare) (const void *cle1, const void *cle2)
                     )
{
  char    *tab = donnee;
  void    *pivot, *temp;
  int     r[3];

  /* ----------------------------------------------------------- */
  /* Allocated memory                                            */
  /* ----------------------------------------------------------- */

  if ((pivot = malloc(taille_elt)) == NULL) { return -1; }
  if ((temp = malloc(taille_elt))  == NULL) { free(pivot); return -1; }

  /* ----------------------------------------------------------- */
  /* Find the pivot                                              */
  /* ----------------------------------------------------------- */

  r[0] = (rand() % (k - i + 1)) + i;
  r[1] = (rand() % (k - i + 1)) + i;
  r[2] = (rand() % (k - i + 1)) + i;
  issort(r, 3, sizeof(int), compare_int);
  memcpy(pivot, &tab[r[1] * taille_elt], taille_elt);

  /* ----------------------------------------------------------- */
  /* create 2 partitions around this value                       */
  /* ----------------------------------------------------------- */

  i--;
  k++;

  while (1)
  {
    /* --------------------------------------------------------- */
    /* Move to the left                                          */
    /* --------------------------------------------------------- */

   do { k--; } while (compare(&tab[k * taille_elt], pivot) > 0);

    /* --------------------------------------------------------- */
    /* Move to the right                                         */
    /* --------------------------------------------------------- */

   do { i++; } while (compare(&tab[i * taille_elt], pivot) < 0);

   if (i >= k) { break; }
   else {
          /* --------------------------------------------------- */
          /* Swap the two elements                               */
          /* --------------------------------------------------- */

          memcpy(temp, &tab[i * taille_elt], taille_elt);
          memcpy(&tab[i * taille_elt], &tab[k * taille_elt], taille_elt);
          memcpy(&tab[k * taille_elt], temp, taille_elt);
        }
  }

  /* ----------------------------------------------------------- */
  /* Free all allocated memory                                   */
  /* ----------------------------------------------------------- */

  free(pivot);
  free(temp);

  /* ----------------------------------------------------------- */
  /* Return the position                                         */
  /* ----------------------------------------------------------- */

  return k;

}


/*! \brief Sort data using the QuickSort algorithm.
    \param donnee Pointer to the beginning of the array to sort.
    \param nb_elts Number of elements to sort.
    \param taille_elt Size in bytes of an element.
    \param i <i>i</i> <B>MUST BE</B> set to 0.
    \param k <i>k</i> <B>MUST BE</B> set to <i>nb_elts - 1</i>.
    \param compare User defined function used to compare two elements of the array.
           If you want to sort according to increasing values, then the function <i>compare()</i> should return:
           <UL>
              <li>1 if <i>cle1 &gt; cle2</i>
              <li>0 if <i>cle1 == cle2</i>
              <li>-1 if <i>cle1 &lt; cle2</i>
           </UL>
           If you want to sort according to decreasing values, then the function <i>compare()</i> should return:
           <UL>
              <li>1 if <i>cle1 &lt; cle2</i>
              <li>0 if <i>cle1 == cle2</i>
              <li>-1 if <i>cle1 &gt; cle2</i>
           </UL>
     \return Upon successful completion the function returns 0, otherwise it returns -1
 */

int qksort (
             void *donnee,
             int  nb_elts, 
             int  taille_elt,
             int  i,
             int  k, 
             int  (*compare)(const void *cle1, const void *cle2)
           )
{
  int   j;


  while (i < k)
  {
    /* --------------------------------------------------------- */
    /* Find where to make a partition                            */
    /* --------------------------------------------------------- */

    if ((j = partition(donnee, taille_elt, i, k, compare)) < 0) { return -1; }

    /* --------------------------------------------------------- */
    /* Sort the left side of the partition                       */
    /* --------------------------------------------------------- */

    if (qksort(donnee, nb_elts, taille_elt, i, j, compare) < 0) { return -1; }

    /* --------------------------------------------------------- */
    /* Sort the right side of the partition                      */
    /* --------------------------------------------------------- */

    i = j + 1;
  }

  return 0;
}

/*! \example test_sort.c
    This file shows how to use the qksort function.
 */


/* ------------------------------------------------------------ */
/* Partitions -- Not part of public API                         */
/* ------------------------------------------------------------ */

static int extended_partition (
                                void  *donnee,
                                int   taille_elt,
                                int   i,
                                int   k, 
                                int   (*compare) (const void *cle1, const void *cle2),
                                int   (*copy_data)(void *src, void *dest),
                                int   (*init_elem)(void *elm)
                              )
{
  char    *tab = donnee;
  void    *pivot, *temp;
  int     r[3];

  /* ----------------------------------------------------------- */
  /* Allocated memory                                            */
  /* ----------------------------------------------------------- */

  if ((pivot = malloc(taille_elt)) == NULL) { return -1; }
  if ((temp  = malloc(taille_elt)) == NULL) { free(pivot); return -1; }

  if (init_elem != NULL)
  {
    if (init_elem(pivot) == 1) { free(pivot); free(temp); return -1; }
    if (init_elem(temp)  == 1) { free(pivot); free(temp); return -1; }
  }

  /* ----------------------------------------------------------- */
  /* Find the pivot                                              */
  /* ----------------------------------------------------------- */

  r[0] = (rand() % (k - i + 1)) + i;
  r[1] = (rand() % (k - i + 1)) + i;
  r[2] = (rand() % (k - i + 1)) + i;
  issort(r, 3, sizeof(int), compare_int);

  if (copy_data (&tab[r[1] * taille_elt], pivot) == 1) { free(pivot); free(temp); return -1; }

  /* ----------------------------------------------------------- */
  /* create 2 partitions around this value                       */
  /* ----------------------------------------------------------- */

  i--;
  k++;

  while (1)
  {
    /* --------------------------------------------------------- */
    /* Move to the left                                          */
    /* --------------------------------------------------------- */

   do { k--; } while (compare(&tab[k * taille_elt], pivot) > 0);

    /* --------------------------------------------------------- */
    /* Move to the right                                         */
    /* --------------------------------------------------------- */

   do { i++; } while (compare(&tab[i * taille_elt], pivot) < 0);

   if (i >= k) { break; }
   else {
          /* --------------------------------------------------- */
          /* Swap the two elements                               */
          /* --------------------------------------------------- */

          if (copy_data (&tab[i * taille_elt], temp) == 1) { free(pivot); free(temp); return -1; }
          if (copy_data (&tab[k * taille_elt], &tab[i * taille_elt]) == 1)  { free(pivot); free(temp); return -1; }
          if (copy_data (temp, &tab[k * taille_elt]) == 1) { free(pivot); free(temp); return -1; }
        }
  }

  /* ----------------------------------------------------------- */
  /* Free all allocated memory                                   */
  /* ----------------------------------------------------------- */

  free(pivot);
  free(temp);

  /* ----------------------------------------------------------- */
  /* Return the position                                         */
  /* ----------------------------------------------------------- */

  return k;
}


/*! \brief Sort data using the QuickSort algorithm.
    \param donnee Pointer to the beginning of the array to sort.
    \param nb_elts Number of elements to sort.
    \param taille_elt Size in bytes of an element.
    \param i <i>i</i> <B>MUST BE</B> set to 0.
    \param k <i>k</i> <B>MUST BE</B> set to <i>nb_elts - 1</i>.
    \param compare User defined function used to compare two elements of the array.
           If you want to sort according to increasing values, then the function <i>compare()</i> should return:
           <UL>
              <li>1 if <i>cle1 &gt; cle2</i>
              <li>0 if <i>cle1 == cle2</i>
              <li>-1 if <i>cle1 &lt; cle2</i>
           </UL>
           If you want to sort according to decreasing values, then the function <i>compare()</i> should return:
           <UL>
              <li>1 if <i>cle1 &lt; cle2</i>
              <li>0 if <i>cle1 == cle2</i>
              <li>-1 if <i>cle1 &gt; cle2</i>
           </UL>
    \param copy_data User defined function used to copy data. <B>The function must take care of memory management</B>.
                     This function must return:
                     <UL>
                        <li>0 upon successful completion.
                        <li>1 if an error occured.
                     </UL>
    \param init_elem User defined function used to initializen an element. This function must return:
                     <UL>
                        <li>0 upon successful completion.
                        <li>1 if an error occured.
                     </UL>
                     This function is called whenever a new element is allocated.
                     Please note that you can specify a NULL pointer if no initialization is needed.
    \return Upon successful completion the function returns 0, otherwise it returns -1
 */

int extended_qksort (
                      void  *donnee,
                      int   nb_elts, 
                      int   taille_elt,
                      int   i,
                      int   k, 
                      int   (*compare)(const void *cle1, const void *cle2),
                      int   (*copy_data)(void *src, void *dest),
                      int   (*init_elem)(void *elm)
                    )
{
  int   j;


  while (i < k)
  {
    /* --------------------------------------------------------- */
    /* Find where to make a partition                            */
    /* --------------------------------------------------------- */

    if ((j = extended_partition(donnee, taille_elt, i, k, compare, copy_data, init_elem)) < 0) { return -1; }

    /* --------------------------------------------------------- */
    /* Sort the left side of the partition                       */
    /* --------------------------------------------------------- */

    if (extended_qksort(donnee, nb_elts, taille_elt, i, j, compare, copy_data, init_elem) < 0) { return -1; }

    /* --------------------------------------------------------- */
    /* Sort the right side of the partition                      */
    /* --------------------------------------------------------- */

    i = j + 1;
  }

  return 0;
}

/*! \example test_ext_sort.c
    This file shows how to use the extended_qksort() function.
 */

