/*! \file replacer.c
    This file contains the implementation of the Search / Replace API.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "replacer.h"
#include "dstring.h"
#include "chtbl.h"

/*! \brief Array of strings. This variable is defined in the parser's module (see file 'replacer.lex').
           It will contain all the tokens extracted by the parser.
 */

extern char** txt_tokens;

/*! \brief Number of tokens in the array 'txt_tokens'. This variable is defined in the parser's module
           (see file 'replacer.lex').
 */

extern size_t nb_tokens;

/*! \brief Initial size for a dynamic string.
 */

#define INIT_DYNAMIC_STRING 100

/* ----------------------------------------------------- */
/* Hash table management                                 */
/* ----------------------------------------------------- */

/*! \brief Type of the elements strored into the hash table.
 */

struct couple {

                /*! \brief This point points to a zero terminated string of characters that represent a
                           tag. */

                char *key;

                /*! \brief This point points to a zero terminated string of characters that represent the
                           value of a tag. */

                char *value;
              };

/*! \brief Number of linked lists for the hash tables used to store the couples (tags, values).
 */

#define HASH_TABLE_ENTRIES   150

/*! \brief Hash function used by the hash table to hash the tags.
    \param key Pointer to a structure <i>couple</i> that contains the couple (tag, value).
    \return The function returns an integer.
    \warning Make sure that the structure pointed by <i>key</i> has been initialized!
 */

static int hash_function (const void *key)
{
  struct couple  *data;
  int            total, i;

  data = (struct couple*)key;
  total = 1;

  for (i=0; i<strlen(data->key); i++) { total *= (int)((data->key)[i]) ; }

  if (total < 0) { total *= -1; } 

  return total;
}

/*! \brief Function used to free memory allocated for the elements of the hash table.
    \param d Pointer to a structure <i>couple</i> to be freed.
 */

static void free_user_data (void *d)
{
  struct couple  *data;

  data = (struct couple*)d;
  free (data);
  return;
}

/*! \brief Function used to compare two elements of the hash table (two tags).
    \param d1 First element to compare.
    \param d2 Second element to compare.
    \return The function return 1 if d1 == d2, 0 otherwise.
 */

static int compare_user_data (const void *d1, const void *d2)
{
  struct couple *data1, *data2;

  data1 = (struct couple*)d1;
  data2 = (struct couple*)d2;

  return !strcmp(data1->key, data2->key);
}

/* ----------------------------------------------------- */
/* End of hash table management                          */
/* ----------------------------------------------------- */

/*! \brief Replace tags by their values in a text.
    \param text Pointer to a zero terminated string of characters. This string contains the original text. 
    \param tags Pointer to an array of tags (that will be searched in 'text').
    \param values Pointer to an array of values (a tag is replaced by a value). Please note that:
           'tags[i]' will be replaced by 'values[i]'.
    \param nb_tags Number of elements in the arrays 'text' and 'tags'.
    \return The function returns a pointer to a zero terminated string of characters that represents the
            final text. If the function could not allocate memory, then the returned value is NULL.
    \warning <ul>
                <li>The returned string of characters has been allocated within the function. Therefore you must
                    free it yourself (using free()).
                <li>This function is <b>NOT REENTRANT</b>, DO NOT use it with threads!
                <li>This function works in memory, so if the input text is too large, the function may fail
                    because the process is running out of memory. In this case the function returns the value
                    NULL. If you get NULL, you should abort the process execution, because the function does
                    not clean the memory allocated...
             </ul>
 */

char* replace_tags (char *text, char **tags, char **values, int nb_tags)
{
  int           res, i;
  char          *final, *pt, *aux;
  dstring       ds;
  size_t        s;
  CHTbl         htab;
  struct couple *elem, selem;
 
  /* --------------------------------------------------- */
  /* Initialize the hash table                           */
  /* --------------------------------------------------- */

  if (
       chtbl_init (
                    &htab,
                    HASH_TABLE_ENTRIES,
                    hash_function,
                    compare_user_data,
                    free_user_data
                  ) == -1
     )
     { return NULL; }

  /* --------------------------------------------------- */
  /* Load hash table with couples (tags => values)       */
  /* --------------------------------------------------- */

  for (i=0; i<nb_tags; i++)
  {
    elem = (struct couple*)malloc(sizeof(struct couple));
    if (elem == NULL)
    {
      chtbl_destroy (&htab);
      return NULL;
    }

    elem->key   = tags[i];
    elem->value = values[i];

    res = chtbl_insert (&htab, (void*)elem);

    switch (res)
    {
      case -1: {
                  chtbl_destroy (&htab);
                  return NULL;
               }; break;

      case 1: /* tag already in hash table */ break;
    }
  }

  /* --------------------------------------------------- */
  /* Parse the original text.                            */
  /* This operation will fill the array 'txt_tokens'     */
  /* all the tokens.                                     */
  /*                                                     */
  /* o Tags begins with a character '+'.                 */
  /* o Other values begins with a character '-'.         */
  /* --------------------------------------------------- */

  res = yylex_memory(text);

  if (res == RELACER_MEM_ERROR)
  {
    chtbl_destroy (&htab);
    return NULL;
  }

  /* --------------------------------------------------- */
  /* Initialize the dynamic string                       */
  /* --------------------------------------------------- */

  if (dstring_init (&ds, INIT_DYNAMIC_STRING) == 1)
  {
    chtbl_destroy (&htab);
    return NULL;
  }

  /* --------------------------------------------------- */
  /* Look at each token and replace it if necessary.     */
  /* --------------------------------------------------- */

  for (i=0; i<nb_tokens; i++)
  {
    pt = txt_tokens[i]+1;

    if ( *(txt_tokens[i]) == '-')
    {
       /* ---------------------------------------------- */
       /* '-' means that this is NOT a tag, we copy it   */
       /* "as is".                                       */
       /* ---------------------------------------------- */

       if (dstring_add (&ds, pt, strlen(pt)) == 1)
       {
         chtbl_destroy (&htab);
         return NULL;
       }
    }
    else
    {
       /* ---------------------------------------------- */
       /* This is a tag, lets replace it                 */
       /* ---------------------------------------------- */

       selem.key   = pt;
       selem.value = NULL;
       elem        = &selem;

       res = chtbl_lookup (&htab, (void**)&elem);

       switch (res)
       {
         case 0: {
                   /* element found */

                   if (dstring_add (&ds, elem->value, strlen(elem->value)) == 1)
                   {
                     chtbl_destroy (&htab);
                     return NULL;
                   }

                 }; break;

         default: {
                   /* element not found */

                   /* ######################################################## */
                   /* ###                                                  ### */
                   /* ### WARNING !!!!                                     ### */
                   /* ###                                                  ### */
                   /* ### The following code depends on the tag's defini-  ### */
                   /* ### -tion (see file 'replacer.lex'.                  ### */
                   /* ###                                                  ### */
                   /* ######################################################## */

                   aux = (char*)malloc(strlen(pt)+4);
                   if (aux == NULL)
                   {
                     dstring_free(&ds);
                     chtbl_destroy (&htab);
                     return NULL;
                   }

                   sprintf (aux, "${%s}", pt);

                   if (dstring_add (&ds, aux, strlen(aux)) == 1)
                   {
                     free (aux);
                     chtbl_destroy (&htab);
                     return NULL;
                   }

                   free (aux);

                 }; break;
       }
    }
  }

  /* --------------------------------------------------- */
  /* Free hash table                                     */
  /* --------------------------------------------------- */

  chtbl_destroy (&htab);

  /* --------------------------------------------------- */
  /* Free all allocated tokens.                          */
  /* --------------------------------------------------- */

  free_tokens();

  /* --------------------------------------------------- */
  /* Copy the content of the dynamic string              */
  /* --------------------------------------------------- */

  final = dstring_get_data (&ds, &s);
  if (final == NULL) { return NULL; }

  /* --------------------------------------------------- */
  /* Free the dynamic string                             */
  /* --------------------------------------------------- */

  dstring_free(&ds);

  return final;
}

/*! \example test_replacer_final.c
    This file shows how to use the function replace_tags().
 */
