/*! \file sem.c
    This file implements basic synchronosation toolkit.
 */

#include "my_sem.h"

/*! \brief System identifier for the set of semaphores. */
static int      semaphore_set = -1;

/*! \brief Number of semaphores in the set of semaphores. */
static int      set_size      = -1;

/*! \brief Default semaphores' permission used at creation time. */
static int      permission    = SEM_PERMISSION;


/*! \brief Set default access permissions for the set of semaphores.
    \param new_permission New permissions.
    \warning By default, the set of semaphores is created with the persmission mask 0x000001C0, that is, in binary:
             111 000 000. This grants all right to the process' user but nothing for other users (just like file permissions).
 */

void change_sem_access_permission (int new_permission) { permission = new_permission; }

/*! \brief Create a set of semaphores.
    \param ipc_key IPC key for the semaphore set. If you have a well known key (it should be a number greater or equal to
           zero), then you can specify it. Otherwise, if you want the kernel to find the first available key, then you must
           specify the value (key_t)-1.
    \param number_of_sem Number of semaphores in the set.
    \return This function may return one of the followinf values:
            <ul>
              <li>CREATE_SEMAPHORE_OK: The operation was successful.
              <li>CREATE_SEMAPHORE_CREATE_ERROR: A system error occured while creating the set of semaphores.
              <li>CREATE_SEMAPHORE_SET_VAL_ERROR: A system error occured while setting semaphores' values to 1.
            </ul>
    \warning By default, semaphores are created with the value 1 (like mutexes). You can change that later by using the
             function change_sem_access_permission().
 */

int create_semaphore_set (key_t ipc_key, int number_of_sem)
{
    int          i;
    union semun  sem_def;

    set_size = number_of_sem;

    /* =============================================================== */
    /* Create the set of semaphores                                    */
    /* =============================================================== */

    if (ipc_key != (key_t)-1) { semaphore_set = semget (ipc_key, number_of_sem, IPC_CREAT | permission); }
    else { semaphore_set = semget (IPC_PRIVATE, number_of_sem, IPC_CREAT | permission); }

    if (semaphore_set == (key_t)-1) { return CREATE_SEMAPHORE_CREATE_ERROR; }

    /* =============================================================== */
    /* Set all semaphore values to 1 (can be change later)             */
    /* =============================================================== */

    for (i=0; i<number_of_sem; i++)
    {
      sem_def.val = 1;
      if (semctl (semaphore_set, i, SETVAL, sem_def) == -1)
      { return CREATE_SEMAPHORE_SET_VAL_ERROR; }
    }

    return CREATE_SEMAPHORE_OK;
}

/*! \example test_shm.c
    This file shows how to use the function create_semaphore_set().
 */

/*! \brief Get a already created (by another process) set of semaphores.
    \param ipc_key Well known IPC key that identifies the set of semaphores.
    \param number_of_sem Number of semaphores in the set.
    \return The function may return one of the following value:
            <UL>
               <li>GET_SEM_OK: The operation was successful.
               <li>GET_SEM_ALREADY_GET: The process already got the set of semaphores.
               <li>GET_SEM_ERROR: An error occured while getting the semaphores.
            </UL>
    \warning Since you get a semaphores set created by another process, you should know the <b>well known</b> IPC key that
             represents the set of semaphores.
 */

int get_semaphore_set (key_t ipc_key, int number_of_sem)
{

  if (semaphore_set != -1) { return GET_SEM_ALREADY_GET; }

  semaphore_set = semaphore_set = semget (ipc_key, number_of_sem, 0);
  set_size = number_of_sem;

  if (semaphore_set == -1) { return GET_SEM_ERROR; }

  return GET_SEM_OK;
}

/*! \example test_shm.c
    This file shows how to use the function get_semaphore_set().
 */

/*! \brief Set semaphores' values (this update the entire set of semaphores).
    \param values Pointer to an array of integers that represents semaphores values.
    \return The function may return one of the followinf values:
            <ul>
              <li>SET_SEM_VALUES_OK: The operation was sucessful.
              <li>SET_SEM_VALUES_SEM_NOT_CREATED: You did not create (or get) the set of semaphores.
              <li>SET_SEM_VALUES_SET_ERROR: An error occured while setting the semaphores' values.
            </ul>
    \warning The number of elements of the array pointed by 'values' must be equal to the number of semaphores in the
             set.
 */

int set_semaphore_values (int *values)
{
   int          i;
   union semun  sem_def;

   if ((set_size == -1) || (semaphore_set == -1)) { return SET_SEM_VALUES_SEM_NOT_CREATED; }

   for (i=0; i<set_size; i++)
   {
     sem_def.val = values[i];
     if (semctl (semaphore_set, i, SETVAL, sem_def) == -1)
     { return SET_SEM_VALUES_SET_ERROR; }
   }

   return SET_SEM_VALUES_OK;
}

/*! \example test_shm.c
    This file shows how to use the function set_semaphore_values().
 */

/*! \brief Set the value of one semaphore (and only one of the set).
    \param value Integer that represents the semaphore's new value.
    \sem_number Integer that represents the semaphore's index (within the set of semaphores).
    \return The function may return one of the followinf values:
            <ul>
              <li>SET_SEM_VALUES_OK: The operation was sucessful.
              <li>SET_SEM_VALUES_SEM_NOT_CREATED: You did not create (or get) the set of semaphores.
              <li>SET_SEM_VALUES_SET_ERROR: An error occured while setting the semaphores' values.
            </ul>
    Please note that the functions set_semaphore_value() and set_semaphore_values() share the same set of return values.
 */

int set_semaphore_value (int value, int sem_number)
{
   union semun  sem_def;

   if ((set_size == -1) || (semaphore_set == -1)) { return SET_SEM_VALUES_SEM_NOT_CREATED; }

   sem_def.val = value;
   if (semctl (semaphore_set, sem_number, SETVAL, sem_def) == -1)
   { return SET_SEM_VALUES_SET_ERROR; }

   return SET_SEM_VALUES_OK;
}

/*! \example test_shm.c
    This file shows how to use the function set_semaphore_value().
 */

/*! \brief Take one semaphore. This action decrements the semaphore's counter by 1.
    \param sem_idx Index of the semaphore to take. The index of the first semaphore is 0.
    \return Upon successful completion, 0 is returned. Otherwise, -1 is returned.
    \warning If the set of semaphores you created contains N semaphores, then &quot;0 &lt;= sem_idx &lt; N&quot;.
 */

int take_sem (int sem_idx)
{
  struct sembuf   Operation;

  Operation.sem_num = sem_idx;
  Operation.sem_flg = SEM_UNDO;
  Operation.sem_op  = -1;

  return (semop(semaphore_set, &Operation, 1));
}

/*! \example test_shm.c
    This file shows how to use the function take_sem().
 */

/*! \brief Release one semaphore. This action increments the semaphore's counter by 1.
    \param sem_idx Index of the semaphore to take. The index of the first semaphore is 0.
    \return Upon successful completion, 0 is returned. Otherwise, -1 is returned.
    \warning If the set of semaphores you created contains N semaphores, then &quot;0 &lt;= sem_idx &lt; N&quot;.
             Please note that we do <B>NOT</B> specify the semaphore's flag <U>SEM_UNDO</U>. If you specify this flag,
             then you will have the following behaviour: when the program exits, the operation will be canceled. 
             Here the operation is &quot;release the semaphore&quot;. So, if we set SEM_UNDO, and if the process exists,
             then the operation will be canceled: in other words, the semaphore will be taken (the oposite of released). 
 */

int release_sem (int sem_idx)
{
  struct sembuf   Operation;

  Operation.sem_num = sem_idx;
  Operation.sem_op  = +1;

  return (semop(semaphore_set, &Operation, 1));
}

/*! \example test_shm.c
    This file shows how to use the function release_sem().
 */

/*! \brief Destroy the set of semaphores.
    \return Upon successful completion, a positive value is returned. Otherwise, -1 is returned.
 */

int destroy_semaphore_set()
{ return (semctl (semaphore_set, 0, IPC_RMID, 0)); }

/*! \example test_shm.c
    This file shows how to use the function destroy_semaphore_set().
 */
