/*! \file shm.c
    This file implements basic shared memory utilities.
 */

#include "my_shm.h"
#include <stdio.h>

/*! \brief System isentifier for the shared memory segment. This value is set by create_shared_memory_segment() or
           get_shared_memory_segment_sys_id().
 */
static int       shm_id = -1;

/*! \brief Address used by the process to access the top of the shared memory segment. This value is set by the function
           attache_shared_memory_segment().
 */
static SHM_addr  attachement_address = (SHM_addr)-1;

/*! \brief This variable defines the default permission for all newly created SHM. It can be modified using the function
           change_shm_access_permission().
 */
static int       permissions = SHM_PERMISSION;

/*! \brief Data structure that contains information about the shared memory segment (see nam shmctl).
 */
static struct shmid_ds shm_info;


/*! \brief Change the SHM's default permissions.
    \param new_permission New permissions to apply to the shared memory (at creation time).
    \warning By default the permission is 0x000001C0 (grant all rights for the process' user).
 */

void change_shm_access_permission (int new_permission)
{ permissions = new_permission; }

/*! \brief Create the shared memory segment.
    \param size Size in bytes of the segment.
    \param ipc_key This argument represents the IPC identifier of the shared memory segment. If you know the (well known)
           identifier's value, then you can specify it. Otherwise, if you want to let the system choose the first available
           identifier, then you should set 'ipc_key' to (key_t)-1. Please note that an IPC identifier is a value greater or
           equal to 0.
    \return On success, the function may return one of the following values:
            <UL>
              <li>CREATE_SHM_ALREADY_CREATED: The shared memory segment has already been created.
              <li>CREATE_SHM_ERROR: An error occured while creating the segment.
              <li>CREATE_SHM_OK: The shared memory segment has been successfully created.
            </UL>
            Once you called create_shared_memory_segment(), you should call get_shared_memory_segment_id() to get the
            system's identifier assigned to the newly created shared memory segment.
 */

int create_shared_memory_segment (key_t ipc_key, int size)
{
  /* ===================================================== */
  /* Make sure that 'shm_id' is not already associated     */
  /* with a shared memory segment.                         */
  /* ===================================================== */

  if (shm_id != -1) { return CREATE_SHM_ALREADY_CREATED; }

  /* ===================================================== */
  /* Create the shared memory segment                      */
  /* ===================================================== */

  if (ipc_key == (key_t)-1) { shm_id = shmget (IPC_PRIVATE, size, IPC_CREAT | permissions); }
  else { shm_id = shmget (ipc_key, size, IPC_CREAT | permissions); }

  if (shm_id == -1) { return CREATE_SHM_ERROR; }

  return CREATE_SHM_OK;
}

/*! \example test_shm.c
    This example shows how to use the function create_shared_memory_segment().
 */

/*! \brief Get the system's identifier of a shared memory segment that has already been created (by another process).
    \param size Size in bytes of the shared memory segment.
    \param ipc_key <b>Well known</b> IPC key that represents the shared memory segment. This key is shared between all 
           processes that access the shared memory segment.
    \return The function may return one of the following value:
            <UL>
              <li>GET_SHM_ALREADY_GET: You've already got the SHA system's identifier.
              <li>GET_SHM_ERROR: An error occured while getting the SHM system's identifier.
              <li>GET_SHM_OK: The operation was successful.
            </UL>
            Once you've called get_shared_memory_segment_sys_id(), you should call get_shared_memory_segment_id() to get the
            SHM system's identifier.
    \warning If you need to call this function, it means that you want to get the system ID of a SHM that has already been
             created by another process. Then you need to specify a <b>well known</b> IPC key.
 */

int get_shared_memory_segment_sys_id (key_t ipc_key, int size)
{
  /* ===================================================== */
  /* Make sure that the shared memory segement is not      */
  /* already attached.                                     */
  /* ===================================================== */

  if (shm_id != -1) { return GET_SHM_ALREADY_GET; }

  /* ===================================================== */
  /* Get the shared memory segment identifier              */
  /* ===================================================== */

  shm_id = shmget (ipc_key, size, 0);

  if (shm_id == -1) { return GET_SHM_ERROR; }

  return GET_SHM_OK;
}

/*! \brief Tell the IPC library to destroy the SHM on the last detachement.
    \return Upon successful completion, the function return 0, otherwise it returns -1.
    \warning There are 2 very important things to know:
             <UL>
                <li>This function should be called by the process that has created the SHM.
                <li>Once you've called this function, another process can NOT get the shared memory segment (the OS
                    consider that the segment is ready to be destroyed ... so no more process can get it). Therefore 
                    you must call this function when you are sure that no more process will try to get the shared memory
                    segment.
             </UL>
 */

int set_shm_auto_destroy()
{ return shmctl(shm_id, IPC_RMID, NULL); }

/*! \example test_shm.c
    This example shows how to use the function get_shared_memory_segment_sys_id().
 */

/*! \brief Returns the shared memory segment identifier.
    \return The function returns the identifier of the shared memory segment.
    \warning The identifier should be a value greater or equal to 0. If this function returns -1, it means that an error
             occured during the creation of the segment or that the segment has not been created.
 */

int get_shared_memory_segment_id() { return shm_id; }

/*! \example test_shm.c
    This example shows how to use the function get_shared_memory_segment_id().
 */

/*! \brief Returns the address used by the process to access the shared memory segment.
    \return The function returns the address used by the process to access the shared memory segment.
 */

SHM_addr get_shm_address() { return attachement_address; }

/*! \example test_shm.c
    This example shows how to use the function get_shm_address().
 */

/*! \brief Attache a previously created shared memory segment.
    \return Upon successful completion the function returns the process' address of the memory segment (that is, the address
            used by the process to access the memory segment). Otherwise, the function returns (SHM_addr)-1.
 */

SHM_addr attache_shared_memory_segment ()
{
  attachement_address = shmat(shm_id, (SHM_addr)0, 0);
  return attachement_address;
}

/*! \example test_shm.c
    This example shows how to use the function attache_shared_memory_segment().
 */

/*! \brief Return the data structure that describes the Shared Memory Segment (as documented in man shmget).
    \return Upon successful completion, the function returns that address of a structure &quot;struct shmid_ds&quot; that
            describes the SHM (see man shmctl for details about this structure). If an error occured, then the function
            returns NULL.
 */

struct shmid_ds* get_shm_info()
{
  if (shmctl (shm_id, IPC_STAT, &shm_info) == -1)
  { return NULL; }

  return &shm_info;
}

/*! \example test_shm.c
    This example shows how to use the function get_shm_info().
 */

/*! \brief Detache the Shared Memory Segment.
    \return If an error occured, the function returns -1. Otherwise, the SHM attachement address is returned.
 */

int detache_shm()
{ return  shmdt(attachement_address); }

