/*! \file signals.c
    This file contains all signals related functions.
 */

#include "signals.h"

/*! \brief Data structure used to put the new signals mask.
 */

static   sigset_t          new_set;

/*! \brief Signal to manipulate.
 */

static int my_sig = -1;

/*! \brief Set the signal to manipulate.
    \param sig Signal to manipulate.
 */

void set_signal (int sig)
{ my_sig = sig; }

/*! \example test_signals.c
    This file shows how to use the simple signal functions.
 */

/*! \brief Block the signal my_sig only. All other signals are allowed.
    \return If the operation was successful, the function returns 0. Otherwise tha function returns 1. Please note that
            this function should never fail. If it fails, then you are in deep trouble ... check your system.
    \warning You must call set_signal() before you can use this function !.
 */

int block_all_signal()
{
   int cr;

   if (sigemptyset (&new_set) == -1) { return 1; }

   if (sigaddset (&new_set, my_sig) == -1) { return 1; }

   /*****************************************************************/
   /* WARNING: This is weird, but sigprocmask() can be interrupted. */
   /* You must test errno.                                          */
   /*****************************************************************/

   cr = sigprocmask(SIG_BLOCK, &new_set, 0);

   while ((cr == -1) && (errno == EINTR))
   { cr = sigprocmask(SIG_BLOCK, &new_set, 0); }
   if (cr == -1) { return 1; }

   return 0;
}

/*! \example test_signals.c
    This file shows how to use the simple signal functions.
 */

/*! \brief Unblock the signal my_sig.
    \return If the operation was successful, the function returns 0. Otherwise tha function returns 1. Please note that
            this function should never fail. If it fails, then you are in deep trouble ... check your system.
    \warning You must call set_signal() before you can use this function !.
 */ 

int retore_signal()
{
   int cr;

   if (sigemptyset (&new_set) == -1) { return 1; }

   if (sigaddset (&new_set, my_sig) == -1) { return 1; }

   /*****************************************************************/
   /* WARNING: This is weird, but sigprocmask() can be interrupted. */
   /* You must test errno.                                          */
   /*****************************************************************/

   cr = sigprocmask(SIG_UNBLOCK, &new_set, 0);

   while ((cr == -1) && (errno == EINTR))
   { cr = sigprocmask(SIG_UNBLOCK, &new_set, 0); }
   if (cr == -1) { return 1; }

   return 0;
}

/*! \example test_signals.c
    This file shows how to use the simple signal functions.
 */

/*! \brief Look if there are any my_sig pending signals. If it is the case, the pensding signales are activated.
    \return If the operation was successful, the function returns 0. Otherwise tha function returns 1. Please note that
            this function should never fail. If it fails, then you are in deep trouble ... check your system.
    \warning If you look at the man page for the function 'sigsuspend()', you see that 'sigsuspend()' can set 'errno' to
             EINTR. But you must ignore it, otherwise you create an infinite loop. Indeed, every time 'sigsuspend()' raises a
             pending signal, 'sigsuspend()' is interrupted ... And therefore,  EINTR is set!. 
             You must call set_signal() before you can use this function !.
 */

int get_pending_signals()
{
  sigset_t    pendsig;
  sigset_t    sigint_set;

  if (sigfillset(&sigint_set) == -1) { return 1; }

  if (sigdelset(&sigint_set, my_sig) == -1) { return 1; }

  if (sigpending(&pendsig) == -1) { return 1; }

  if (sigismember (&pendsig, my_sig) == 1)
  { sigsuspend (&sigint_set); }

  return 0;
}

/*! \example test_signals.c
    This file shows how to use the simple signal functions.
 */

/*! \brief Set a signal handler on the signal my_sig.
    \param handler Pointer to the signal's handler. The function's prototype should look something like:
                   void function_name(int).
    \return If the operation was successful, the function returns 0. Otherwise tha function returns 1. Please note that
            this function should never fail. If it fails, then you are in deep trouble ... check your system.
    \warning You must call set_signal() before you can use this function !.
 */

int set_signal_handler(void (*handler)(int))
{
  struct sigaction  new_sig_settings;

  new_sig_settings.sa_handler = handler;

  /******************************************************************/
  /* 'sa_mask' represents the set of signals that should be blocked */
  /* while the current signal is being processed. By default the    */
  /* current signal is blocked too (unless SA_NODEFER or SA_NOMASK  */
  /* is set in 'sa_flags').                                         */
  /*                                                                */
  /* Here we say that *only* signals my_sig will be blocked while  */
  /* the current signal is being processed. And, of course, the     */
  /* current signal is blocked too while it is executed. This is    */
  /* the default behaviour and I think it is safe.                  */
  /******************************************************************/

  if (sigemptyset (&(new_sig_settings.sa_mask)) == -1) { return 1; }

  if (sigaddset (&(new_sig_settings.sa_mask), my_sig) == -1) { return 1; }

  /******************************************************************/
  /* No special mask needed.                                        */
  /******************************************************************/

  new_sig_settings.sa_flags = 0;

  /******************************************************************/
  /* We want to catch my_sig.                                      */
  /* I mean only this signal will be linked to an handler.          */
  /******************************************************************/

  if (sigaction(my_sig, &new_sig_settings, 0) == -1) { return 1; }

  /******************************************************************/
  /* OK, now this is the situation:                                 */
  /*                                                                */
  /* - We accept *ALL* signals.                                     */
  /* - When a signal is processed, all other signals are blocked,   */
  /*   including the current signal itself.                         */
  /* - We set a handler to my_sig only.                            */
  /******************************************************************/

  return 0;
}

/*! \example test_signals.c
    This file shows how to use the simple signal functions.
 */



