/*! \file smtp.c
    This file implements the SMTP client API.
 */

#include <unistd.h>
#include "my_smtp.h"

/*! \brief Internet address of the SMTP server.
 */

static struct sockaddr_in tcp_add;

/*! \brief Socket descriptor associated with the TCP connexion.
 */

static int tcp_sock = -1;

/*! \brief This function is the flex parser itself. Please look at the file 'smtp.lex' for details.
 */

extern int smtp_yylex_memory (char *src);

/*! \brief Parse the response from the SMTP server.
    \param response Pointer to a <b>zero terminated</b> string of characters the response from the SMTP server.
    \param data Pointer to a <i>smtp_response</i> data structure that will be used to store information
           about the repsonse.
    \return The function returns one of the following value:
            <UL>
              <li>SMTP_SERVER_IDENTIFICATION_OK = 220
                  ("220 newsletter.tiscali.fr Simple Mail Transfer Service Ready" - follow the connexion)
              <li>SMTP_CLIENT_IDENTIFICATION_OK = 250
                  ("250 newsletter.tiscali.fr" - Follow the command HELO or EHLO)
              <li>SMTP_FROM_OK = 250
                  ("250 OK" - Follow the command MAIL FROM)
              <li>SMTP_RCPT_OK = 250
                  ("250 OK" - Follow the command RCPT TO)
              <li>SMTP_DATA_OK = 354
                  ("354 Start mail input; end with...", Follow the command DATA)
              <li>SMTP_END_OF_DATA_OK = 250
                  ("250 OK" - Follow the &lt;CRLF&gt;.&lr;CRLF&gt; that marks the end of the data part.
              <li>SMTP_INVALID_ANSWER = -1
                  This means that the response is not well formated.
            </UL>
 */

static int response_parser (char *response, struct smtp_response *data)
{
  int rc;

  rc = smtp_yylex_memory(response);

  /* ---------------------------------------------- */
  /* Please look at the file 'smtp.lex': the value  */
  /* of rc can be:                                  */
  /*                                                */
  /*  o 220                                         */
  /*  o 250                                         */
  /*  o 354                                         */
  /*  o -1                                          */
  /*  o 0                                           */
  /* ---------------------------------------------- */

  if ((rc == 0) || (rc == -1)) { return SMTP_INVALID_ANSWER; }

  return rc;
}






/*! \brief Open a connexion to the SMTP server.
    \param conf Pointer to a configuration structure.
    \param result Pointer to <i>smtp_res</i> structure.
    \return Upon successful completion, the function returns a positive number that represents the socket
            descriptor associated with the connexion. In this case, the field <i>ellapsed</i> contains
            the ellapsed time of the connexion in milli-seconds.

            Otherwise the function returns the value -1. In this case the field <i>error_code</i> of
            <i>result</i> contains a value that identifies the failure. This value may be:

            <UL>
              <li>An error code from the function <i>open_tcp_connexion()</i> (of the Common C Library). See the
                  detailed description of <i>open_tcp_connexion()</i> for full details.
              <li>An error code from the function <i>read_from_socket()</i>(of the Common C Library). See the
                  detailed description of <i>read_from_socket()</i> for full details.
              <li>The value SMTP_WRITE_ERROR. This means that the system call write failed.
              <li>The value SMTP_BAD_ANSWER. This means that the response from the SMTP server was not valid.
            </UL>
 */

int client_smtp_connect ( 
                          struct smtp_connexion_config *conf,
                          struct smtp_res *result
                        )
{
  int                   sock, rc;
  char                  iobuff[SMTP_BUFFER_SIZE];
  size_t                read_bytes;
  struct smtp_response  smtp_res;
  long                  ellapsed;
  

  result->error_code = 0;
  result->ellapsed   = 0;
  tcp_sock           = -1;

  /* ----------------------------------------------- */
  /* Open the TCP connexion                          */
  /* ----------------------------------------------- */

  sock = open_tcp_connexion(
                             &tcp_add,
                             conf->host,
                             conf->port,
                             conf->connect_timeout_sec,
                             conf->connect_timeout_micro,
                             &ellapsed
                           );

  if (sock < 0)
  {
    result->error_code = sock;
    return -1;
  }

  result->ellapsed += ellapsed;

  /* ----------------------------------------------- */
  /* Read server's identification                    */
  /* ----------------------------------------------- */

  rc = read_from_socket (
                           sock,
                           iobuff,
                           SMTP_BUFFER_SIZE-1,
                           &read_bytes,
                           conf->read_timeout_sec,
                           conf->read_timeout_micro,
                           &ellapsed
                        );

  if (rc != SCK_READ_OK)
  {
    smtp_disconnect();
    result->error_code = rc;
    return -1;
  }

  result->ellapsed += ellapsed;

  iobuff[read_bytes] = 0;

  /* ----------------------------------------------- */
  /* Check the answer                                */
  /* ----------------------------------------------- */

  rc = response_parser (iobuff, &smtp_res);

  if (rc != SMTP_SERVER_IDENTIFICATION_OK)
  {
    smtp_disconnect();
    result->error_code = SMTP_BAD_ANSWER;
    return -1;
  }

  /* ----------------------------------------------- */
  /* Send client's identificator                     */
  /* ----------------------------------------------- */

  if (write (sock, conf->id, strlen(conf->id)) !=  strlen(conf->id))
  {
    smtp_disconnect();
    result->error_code = SMTP_WRITE_ERROR;
    return -1;
  }

  /* ----------------------------------------------- */
  /* Read the answer from the SMTP server            */
  /* ----------------------------------------------- */

  rc = read_from_socket (
                           sock,
                           iobuff,
                           SMTP_BUFFER_SIZE-1,
                           &read_bytes,
                           conf->read_timeout_sec,
                           conf->read_timeout_micro,
                           &ellapsed
                        );

  if (rc != SCK_READ_OK)
  {
    smtp_disconnect();
    result->error_code = rc;
    return -1;
  }

  result->ellapsed += ellapsed;

  iobuff[read_bytes] = 0;

  /* ----------------------------------------------- */
  /* Check the answer                                */
  /* ----------------------------------------------- */

  rc = response_parser (iobuff, &smtp_res);

  if (rc != SMTP_CLIENT_IDENTIFICATION_OK)
  {
    smtp_disconnect();
    result->error_code = SMTP_BAD_ANSWER;
    return -1;
  }

  /* ----------------------------------------------- */
  /* Everything is OK, set 'tcp_sock' to the socket  */
  /* descriptor.                                     */
  /* ----------------------------------------------- */

  tcp_sock = sock;

  return 0;
}

/*! \brief Close the connexion between the SMTP client ant the server.
 */

void smtp_disconnect() { if (tcp_sock >= 0) { close(tcp_sock); } }

/*! \brief Send an email to the SMTP server.
    \param conf Pointer to a configuration structure.
    \param email Pointer to a <i>smtp_mail</i> structure that contains the mail's data.
    \param result Pointer to <i>smtp_res</i> structure.
    \return Upon successfull completion the function returns 0, therwise it returns -1 and <b>the connexion is
            closed</b>.

            If the returned value is 0 (success), then the field <i>ellapsed</i> of the structure pointed by
            <i>result</i> contains the ellapsed time of the action in milli-seconds.

            If the returned value is -1 (error), then the field <i>error_code</i> of the structure pointed by
            <i>result</i> contains an error code that describes the failure. This value may be:

            <UL>
              <li>An error code from the function <i>read_from_socket()</i>(of the Common C Library). See the
                  detailed description of <i>read_from_socket()</i> for full details.
              <li>The value SMTP_WRITE_ERROR. This means that the system call write failed.
              <li>The value SMTP_BAD_ANSWER. This means that the response from the SMTP server was not valid.
              <li>The value SMTP_BUFFER_OVERFLOW. This means that one of the input arguments "from" or "rcpt"
                  is too long.
            </UL>


 */

int smtp_send_email (
                      struct smtp_connexion_config  *conf,
                      struct smtp_mail              *email,
                      struct smtp_res               *result
                    )
{
  int                   i, rc;
  char                  iobuff[SMTP_BUFFER_SIZE];
  size_t                read_bytes;
  struct smtp_response  smtp_res;
  long                  ellapsed;

  result->error_code = 0;
  result->ellapsed   = 0;

  /* ----------------------------------------------- */
  /* Check for buffer overflows                      */  
  /* ----------------------------------------------- */

  if (
       (
         strlen(email->from)
         +
         strlen(SMTP_FROM_TAG_START)
         +
         strlen(SMTP_FROM_TAG_STOP)
       )
       >
       SMTP_BUFFER_SIZE - 1
     )
     { 
       smtp_disconnect();
       result->error_code = SMTP_BUFFER_OVERFLOW;
       return -1;
     }

  sprintf (iobuff, "%s%s%s", SMTP_FROM_TAG_START, email->from, SMTP_FROM_TAG_STOP); 

  /* ----------------------------------------------- */
  /* Send the FROM command                           */  
  /* ----------------------------------------------- */

  if (write (tcp_sock, iobuff, strlen(iobuff)) !=  strlen(iobuff))
  {
    smtp_disconnect();
    result->error_code = SMTP_WRITE_ERROR;
    return -1;
  }

  /* ----------------------------------------------- */
  /* Read answer from SMTP server                    */
  /* ----------------------------------------------- */

  rc = read_from_socket (
                           tcp_sock,
                           iobuff,
                           SMTP_BUFFER_SIZE-1,
                           &read_bytes,
                           conf->read_timeout_sec,
                           conf->read_timeout_micro,
                           &ellapsed
                        );

  if (rc != SCK_READ_OK)
  {
    smtp_disconnect();
    result->error_code = rc;
    return -1;
  }

  result->ellapsed += ellapsed;

  iobuff[read_bytes] = 0;

  /* ----------------------------------------------- */
  /* Check the answer from the SMTP server           */
  /* ----------------------------------------------- */

  rc = response_parser (iobuff, &smtp_res);

  if (rc != SMTP_FROM_OK)
  {
    smtp_disconnect();
    result->error_code = SMTP_BAD_ANSWER;
    return -1;
  }

  /* ----------------------------------------------- */
  /* Send all RCPT data                              */
  /* ----------------------------------------------- */

  for (i=0; i<email->rcpt_number; i++)
  {

    /* --------------------------------------------- */
    /* Check for buffer overflows                    */
    /* --------------------------------------------- */

    if (
         (
           strlen(email->rcpt)
           +
           strlen(SMTP_RCPT_TAG_START)
           +
           strlen(SMTP_RCPT_TAG_STOP)
         )
         >
         SMTP_BUFFER_SIZE - 1
       )
       {
         smtp_disconnect();
         result->error_code = SMTP_BUFFER_OVERFLOW;
         return -1;
       }

    sprintf (iobuff, "%s%s%s", SMTP_RCPT_TAG_START, email->rcpt, SMTP_RCPT_TAG_STOP);

    /* --------------------------------------------- */
    /* Send the RCPT command                         */
    /* --------------------------------------------- */

    if (write (tcp_sock, iobuff, strlen(iobuff)) !=  strlen(iobuff))
    {
      smtp_disconnect();
      result->error_code = SMTP_WRITE_ERROR;
      return -1;
    }

    /* --------------------------------------------- */
    /* Read the response from the SMTP server        */
    /* --------------------------------------------- */

    rc = read_from_socket (
                             tcp_sock,
                             iobuff,
                             SMTP_BUFFER_SIZE-1,
                             &read_bytes,
                             conf->read_timeout_sec,
                             conf->read_timeout_micro,
                             &ellapsed
                          );

    if (rc != SCK_READ_OK)
    {
      smtp_disconnect();
      result->error_code = rc;
      return -1;
    }

    result->ellapsed += ellapsed;

    iobuff[read_bytes] = 0;

    /* --------------------------------------------- */
    /* Check the answer from the SMTP server         */
    /* --------------------------------------------- */

    rc = response_parser (iobuff, &smtp_res);

    if (rc != SMTP_RCPT_OK)
    {
      smtp_disconnect();
      result->error_code = SMTP_BAD_ANSWER;
      return -1;
    }
  }

  /* ----------------------------------------------- */
  /* Now send the DATA command                       */
  /* ----------------------------------------------- */

  strcpy (iobuff, "DATA\r\n");

  if (write (tcp_sock, iobuff, strlen(iobuff)) !=  strlen(iobuff))
  {
    smtp_disconnect();
    result->error_code = SMTP_WRITE_ERROR;
    return -1;
  }

  /* ----------------------------------------------- */
  /* Read answer from SMTP server                    */
  /* ----------------------------------------------- */

  rc = read_from_socket (
                           tcp_sock,
                           iobuff,
                           SMTP_BUFFER_SIZE-1,
                           &read_bytes,
                           conf->read_timeout_sec,
                           conf->read_timeout_micro,
                           &ellapsed
                        );

  if (rc != SCK_READ_OK)
  {
    smtp_disconnect();
    result->error_code = rc;
    return -1;
  }

  result->ellapsed += ellapsed;

  iobuff[read_bytes] = 0;

  /* ----------------------------------------------- */
  /* Check the answer from the SMTP server           */
  /* ----------------------------------------------- */

  rc = response_parser (iobuff, &smtp_res);

  if (rc != SMTP_DATA_OK)
  {
    smtp_disconnect();
    result->error_code = SMTP_BAD_ANSWER;
    return -1;
  }

  /* ----------------------------------------------- */
  /* Send the body itself                            */
  /* ----------------------------------------------- */

  if (write (tcp_sock, email->data, strlen(email->data)) !=  strlen(email->data))
  {
    smtp_disconnect();
    result->error_code = SMTP_WRITE_ERROR;
    return -1;
  }

  /* ----------------------------------------------- */
  /* Read answer from SMTP server                    */
  /* ----------------------------------------------- */

  rc = read_from_socket (
                           tcp_sock,
                           iobuff,
                           SMTP_BUFFER_SIZE-1,
                           &read_bytes,
                           conf->read_timeout_sec,
                           conf->read_timeout_micro,
                           &ellapsed
                        );

  if (rc != SCK_READ_OK)
  {
    smtp_disconnect();
    result->error_code = rc;
    return -1;
  }

  result->ellapsed += ellapsed;

  iobuff[read_bytes] = 0;

  /* ----------------------------------------------- */
  /* Check the answer from the SMTP server           */
  /* ----------------------------------------------- */

  rc = response_parser (iobuff, &smtp_res);

  if (rc != SMTP_END_OF_DATA_OK)
  {
    smtp_disconnect();
    result->error_code = SMTP_BAD_ANSWER;
    return -1;
  }

  /* ----------------------------------------------- */
  /* Send the QUIT command and close the connexion   */
  /* ----------------------------------------------- */

  strcpy (iobuff, "QUIT\r\n");

  if (write (tcp_sock, iobuff, strlen(iobuff)) !=  strlen(iobuff))
  {
    smtp_disconnect();
    result->error_code = SMTP_WRITE_ERROR;
    return -1;
  }

  smtp_disconnect();

  return 0;
}
