/*! \file sockets.c
    This file implements basic TCP/IP functionalities.
 */ 

#include <sys/wait.h> 
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/time.h>

#include <errno.h>
#include <stdio.h>
#include "my_sockets.h"




/*! \brief Open a TCP connexion to a remote server.
    \param tcp_add Pointer to a structure that will receive the server's IP address. You probably don't need it, but sometimes
           you must have it.
    \param tcp_host Pointer to a zero terminated string of characters that contains the name (or IP address) of the server
           running the host.
    \param tcp_port TCP port used by the server to a listen incoming connections.
    \param tt_sec Number of second for the timeout.
    \param tt_micro Number of micro second for the timeout.
    \param ellapsed_time Ellapsed time - in milli seconds - for the connect() system call.
    \return If the connection succed, the function returns the newly created socket descriptor (which is a positive value).
            Otherwise, if the connection failed, the function returns one of the following codes (negative value):
            <UL>
              <li>SCK_CREATE_ERROR: Can not create the socket (the function socket() failed).
              <li>SCK_CONNECT_ERROR: Can not connect to the remote server (the function connect() failed).
              <li>SCK_GET_IP_ERROR: Can not get the remote server's IP address, the function gethostbyname() failed.
              <li>SCK_TIMEOUT_CONNECT: The connect() call has been broken by a signal - timeout.
              <li>SCK_TIMER_ERROR: An error occured with the high precision timer.
              <li>SCK_CHRONO_ERROR: An error occured with the high precision chronometer.
            </UL>
     \warning Please note that:
              <UL>
                 <li>To use this function you <B>MUST</B> set a handler on the signal SIGALRM. The signal SIGALRM is used to implement the
                     connect() timeout. 
                 <li>This use its own timers, so you can still use the timing library that comes with the Common C Library.
              </UL>
 */

int open_tcp_connexion (struct sockaddr_in *tcp_add, char *tcp_host, int tcp_port, long tt_sec, long tt_micro, unsigned long int* ellapsed_time)
{
  int			desc;
  int			cr;
  struct hostent	*name;
  char			ip[16];
  int			err;
  struct timeval        start, stop;
  struct itimerval      one_shot_timer, old_one_shot_timer;

  *ellapsed_time = 0;

  /* --------------------------------------------------------------------- */
  /* Get the remote server's IP address (using DNS)                        */
  /* --------------------------------------------------------------------- */

  name = gethostbyname (tcp_host);
  if (name == NULL) { return SCK_GET_IP_ERROR; }

  sprintf (ip, "%d.%d.%d.%d", (unsigned char)(name->h_addr_list)[0][0],
                              (unsigned char)(name->h_addr_list)[0][1],
                              (unsigned char)(name->h_addr_list)[0][2],
                              (unsigned char)(name->h_addr_list)[0][3]);

  /* --------------------------------------------------------------------- */
  /* Create client's socket                                                */
  /* --------------------------------------------------------------------- */

  desc = socket(AF_INET, SOCK_STREAM, 0);
  if (desc == -1) { return (SCK_CREATE_ERROR); };
  Get_Sock_Adress (ip, tcp_port, tcp_add);

  /* --------------------------------------------------------------------- */
  /* Get the current timestamp                                             */
  /* --------------------------------------------------------------------- */

  if (gettimeofday(&start, NULL) == -1) { close(desc); return SCK_CHRONO_ERROR; }

  /* --------------------------------------------------------------------- */
  /* Set the one shot timer (that raises SIGALRM)                          */
  /* --------------------------------------------------------------------- */

  one_shot_timer.it_interval.tv_sec   = 0;
  one_shot_timer.it_interval.tv_usec  = 0;
  one_shot_timer.it_value.tv_sec      = tt_sec;
  one_shot_timer.it_value.tv_usec     = tt_micro;
  if (setitimer (ITIMER_REAL, &one_shot_timer, &old_one_shot_timer) == -1) { close(desc); return SCK_TIMER_ERROR; }

  /* --------------------------------------------------------------------- */
  /* Connect to the remote server                                          */
  /* --------------------------------------------------------------------- */

  cr  = connect(desc, (struct sockaddr*)tcp_add, sizeof(struct sockaddr_in));
  err = errno;

  /* --------------------------------------------------------------------- */
  /* Cancel the one shot timer                                             */
  /* --------------------------------------------------------------------- */

  one_shot_timer.it_interval.tv_sec   = 0;
  one_shot_timer.it_interval.tv_usec  = 0;
  one_shot_timer.it_value.tv_sec      = 0;
  one_shot_timer.it_value.tv_usec     = 0;
  if (setitimer (ITIMER_REAL, &one_shot_timer, &old_one_shot_timer) == -1) { close(desc); return SCK_TIMER_ERROR; }

  /* --------------------------------------------------------------------- */
  /* Get the current timestamp                                             */
  /* --------------------------------------------------------------------- */

  if (gettimeofday(&stop, NULL) == -1) { close(desc); return SCK_CHRONO_ERROR; }
  
  if (cr == -1)
  {
    close(desc);
    if (err == EINTR) { return SCK_TIMEOUT_CONNECT; } 
    return SCK_CONNECT_ERROR;
  } 

  /* --------------------------------------------------------------------- */
  /* Calculate the ellapsed time and return                                */
  /* --------------------------------------------------------------------- */

  *ellapsed_time = (stop.tv_sec*1000 + stop.tv_usec/1000) - (start.tv_sec*1000 + start.tv_usec/1000);
  return (desc);
};

/*! \example test_sockets.c
    This file shows hos to use the function open_tcp_connexion().
 */

/*! \brief Get the Internet address of a service.
    \param Machine_serveur Host IP address.
    \param port_serveur TCP port.
    \param TCP_address Pointer to a "sockaddr_in" used to store the internet address.
    \return The function always return SCK_OK.
 */

void Get_Sock_Adress (char *Machine_serveur, int port_serveur, struct sockaddr_in *TCP_address) 
{
  memset ((void*)TCP_address, 0, sizeof(struct sockaddr_in));
  TCP_address->sin_family        = AF_INET;
  TCP_address->sin_port          = htons((u_short)port_serveur);
  (TCP_address->sin_addr).s_addr = inet_addr(Machine_serveur);
}

/*! \brief Read data from an open socket descriptor (read timeout in micro seconds + ellapsed time in milli seconds).
    \param sockfd Socket descriptor (this must be a descriptor to an open socket).
    \param buffer Pointer to that memory location that will receive the data extracted from the socket.
    \param buffer_size Size of the memory location pointed by <i>buffer</i> (in other words, the maximum amount of bytes that can be read).
    \param extracted Pointer to an interger that will be used to store the number of extracted bytes.
    \param tt_sec Number of seconds of the timeout.
    \param tt_micro Number of micro seconds of the timeout.
    \param ellapsed_time Pointer to an unsigned long integer that will be used to store the ellapsed time.
    \return The function may return one of the following values:
            <UL>
              <li>SCK_READ_CHRONO_ERROR: an error occured while manipulating the high precesion chronometer.
              <li>SCK_READ_TIMER_ERROR: an error occured while manipulating the high precesion timer.
              <li>SCK_READ_TIMEOUT: Read timeout.
              <li>SCK_READ_READ_ERROR: The system call <i>read()</i> failed.
              <li>SCK_READ_OK: sucess.
            </UL>
 */

int read_from_socket (int sockfd, char *buffer, size_t buffer_size, size_t* extracted, long tt_sec, long tt_micro, unsigned long int* ellapsed_time)
{
  int                   err;
  size_t                cr;
  struct timeval        start, stop;
  struct itimerval      one_shot_timer, old_one_shot_timer;


  *extracted     = -1;
  *ellapsed_time = -1;

  /* --------------------------------------------------------------------- */
  /* Get the current timestamp                                             */
  /* --------------------------------------------------------------------- */

  if (gettimeofday(&start, NULL) == -1) { return SCK_READ_CHRONO_ERROR; }

  /* --------------------------------------------------------------------- */
  /* Set the one shot timer (that raises SIGALRM)                          */
  /* --------------------------------------------------------------------- */

  one_shot_timer.it_interval.tv_sec   = 0;
  one_shot_timer.it_interval.tv_usec  = 0;
  one_shot_timer.it_value.tv_sec      = tt_sec;
  one_shot_timer.it_value.tv_usec     = tt_micro;
  if (setitimer (ITIMER_REAL, &one_shot_timer, &old_one_shot_timer) == -1) { return SCK_READ_TIMER_ERROR; }

  /* --------------------------------------------------------------------- */
  /* Read data from socket                                                 */
  /* --------------------------------------------------------------------- */

  cr = read (sockfd, (void*)buffer, buffer_size);
  err = errno;

  /* --------------------------------------------------------------------- */
  /* Cancel the one shot timer                                             */
  /* --------------------------------------------------------------------- */

  one_shot_timer.it_interval.tv_sec   = 0;
  one_shot_timer.it_interval.tv_usec  = 0;
  one_shot_timer.it_value.tv_sec      = 0;
  one_shot_timer.it_value.tv_usec     = 0;
  if (setitimer (ITIMER_REAL, &one_shot_timer, &old_one_shot_timer) == -1) { return SCK_READ_TIMER_ERROR; }

  /* --------------------------------------------------------------------- */
  /* Get the current timestamp                                             */
  /* --------------------------------------------------------------------- */

  if (gettimeofday(&stop, NULL) == -1) { return SCK_READ_CHRONO_ERROR; }

  if (cr == (size_t)-1) 
  { 
    if (err == EINTR) { return SCK_READ_TIMEOUT; }
    return SCK_READ_READ_ERROR;
  }

  /* --------------------------------------------------------------------- */
  /* Calculate ellapsed time in milli seconds and exit                     */
  /* --------------------------------------------------------------------- */

  *ellapsed_time = (stop.tv_sec*1000 + stop.tv_usec/1000) - (start.tv_sec*1000 + start.tv_usec/1000);
  *extracted = cr;

  return SCK_READ_OK;
}

/*! \brief Read from a socket, until the connexion is closed.
    \param sockfd Socket descriptor.
    \param buffer Pointer to a pointer that will be used to point to a dynamically allocated memory space. This memory space will contain
                  a zero terminated string of characters which is the peer's reponse.
    \param tt_sec Number of seconds for the timeout.
    \param tt_micro Number of micro seconds for the timeout.
    \param ellapsed_time Pointer to an <i>unsigned long int</i> that will be used to store the ellapsed time.
    \param size Pointer to an <i>unsigned long int</i> that will be used to store the number of bytes read from the peer.
    \return The function may return the following values:
            <UL>
              <li>SCK_READ_OK: the action was succesful. <i>buffer</i> points to a zero terminated string of characters.
              <li>SCK_READ_NO_MEM: the function could not allocate memory.
              <li>SCK_READ_READ_ERROR: The system call read() failed.
              <li>SCK_READ_TIMEOUT: Read timeout.
              <li>SCK_READ_CHRONO_ERROR: A problem occured with the high precision chronometer.
              <li>SCK_READ_TIMER_ERROR: A problem occured with the high precision timer.
            </UL>
    \warning <UL>
              <li>If the returned value is SCK_READ_OK, then the <i>buffer</i> points to a string of characters that has been allocated 
                  within the function. You must free it yourself (using free()).
              <li>This function assumes that the peer closes the connexion when the transfer is finished. This is the case for example if
                  you read the answer from a HTTP server (with no <i>keep alive</i>). 
             </UL>

 */

int read_all_socket (int sockfd, char** buffer, long tt_sec, long tt_micro, unsigned long int* ellapsed_time, unsigned long int* size)
{
  size_t                lu, total_lu, total_allocated;
  char                  buff[SCK_BUFFER_INIT_SIZE];
  int                   cr;
  unsigned long int     et;


  /* --------------------------------------------------------------------- */
  /* Initialize variables                                                  */
  /* --------------------------------------------------------------------- */

  et              = 0;
  lu              = 0;
  total_lu        = 0;
  *ellapsed_time  = 0;
  *size           = 0;
  total_allocated = SCK_BUFFER_INIT_SIZE;
  *buffer         = NULL;

  /* --------------------------------------------------------------------- */
  /* Allocate memory                                                       */
  /* --------------------------------------------------------------------- */

  *buffer = (char*)malloc(total_allocated);
  if (*buffer == NULL) { return SCK_READ_NO_MEM; }

  do {
        /* --------------------------------------------------------------- */
        /* Read data from socket                                           */
        /* --------------------------------------------------------------- */

        cr = read_from_socket (sockfd, buff, SCK_BUFFER_INIT_SIZE, &lu, tt_sec, tt_micro, &et);

        /* --------------------------------------------------------------- */
        /* Please remember that cr could be:                               */
        /*    -  SCK_READ_CHRONO_ERROR.                                    */
        /*    -  SCK_READ_TIMER_ERROR                                      */
        /*    -  SCK_READ_TIMEOUT                                          */
        /*    -  SCK_READ_READ_ERROR                                       */
        /*    -  SCK_READ_OK                                               */
        /*                                                                 */
        /* These values are shared between read_from_socket() and          */
        /* read_all_socket().                                              */
        /*                                                                 */
        /* read_all_socket() may also return SCK_READ_NO_MEM.              */
        /* --------------------------------------------------------------- */

        if (cr != SCK_READ_OK) { free (*buffer); return cr; }

        *ellapsed_time += et;

        /* --------------------------------------------------------------- */
        /* Increase buffer size if necessary and add data to the end of    */
        /* the buffer.                                                     */
        /* --------------------------------------------------------------- */

        while (total_lu+lu > total_allocated)
        {
          total_allocated *= 2;
          *buffer = (char*)realloc((void*)*buffer, total_allocated);
          if (*buffer == NULL)
          {
            free (*buffer);
            return SCK_READ_NO_MEM;
          }
        }

        memcpy ((void*)(*buffer + total_lu), (void*)buff, lu);

        total_lu += lu;

     } while (lu > 0);

  /* --------------------------------------------------------------------- */
  /* Add zero to terminate the string of chatacters                        */
  /* --------------------------------------------------------------------- */


  *size = total_lu;

  /* add the final zero */
  if (total_lu+1 > total_allocated)
  {
    total_allocated += 1;
    *buffer = (char*)realloc( (void*)*buffer, total_allocated);
    if (*buffer == NULL)
    {
       free (*buffer);
       return SCK_READ_NO_MEM;
    }
  }

  *(*buffer + total_lu) = 0;

  return SCK_READ_OK;
}

/*! \example test_sockets.c
    This file shows hos to use the function read_all_socket().
 */

/*! \brief Get the IP address and port number associated with a given connection.
    \param Address_Sender Pointer to a structure sockaddr_in that describes the peer's connection.
    \param IP_Address Pointer to a string (SCK_ADDR_LEN character long) used to store the IP address.
    \param Port Pointer to an integer used to store the port number.
    \return N/A.
 */

void get_ip_addr(struct sockaddr_in *Address_Sender, char *IP_Address, int *Port)
{
  int  i;
  char *a;

  a = inet_ntoa(Address_Sender->sin_addr);
  for (i=0; i<SCK_ADDR_LEN-1; i++) {IP_Address[i] = *(a+i); };
  IP_Address[SCK_ADDR_LEN-1] = 0;

  *Port = ntohs (Address_Sender->sin_port);

  return;
}






