/*! \file strings_utils.c
    This file contains various string utilities.
 */


#include "strings_utils.h"



/*! \brief Replace all series of '\n' by one ';' and all series of spaces by one space.
    \param st Pointer to a zero terminated string of characters that represents the original string.
    \return Upon successful completion the function returns a pointer to a zero terminated string of charaters
            that represents the resulting string. If an error occured (no memory), then the function returns
            NULL.
    \warning The returned pointer points to a memory location allocated within the function. Therefore you
             <B>MUST</B> free it (using free()). Before you free it, make sure that the pointer is not NULL.
 */

char *make_one_line (char *st)
{
  char *res;
  int  size, i, j, space, nl;

  /***************************************************************/
  /* Allocated memory for the result                             */
  /***************************************************************/

  size = strlen(st);
  res = (char*) malloc(sizeof(char) * (size + 1));
  if (res == NULL) { return NULL; }

  /***************************************************************/
  /* Start replacement                                           */
  /***************************************************************/
 
  j     = 0; 
  space = 0;
  nl    = 0;

  for (i=0; i<size; i++)
  {
    if (st[i] == ' ')
    {
      if (space == 0) { res[j++] = ' '; }
      space = 1;
      nl    = 0;
      continue;
    }

    space = 0;

    if (st[i] == '\n')
    {
      if (nl == 0) { res[j++] = ';'; }
      nl = 1;
      continue;
    }

    nl = 0;

    res[j++] = st[i];
  }


  res[j] = 0;
  return res;
}

/*! \example test_make_one_line.c
    This example shows how to use the function make_one_line().
 */

/*! \brief Contains hexadecimal codes for the 256 values from 0 to 255. Values are written using 3 digits (01, 02, ...).
 */

extern char *bin2hex_hexa[256];


/*! \brief Split a string into fields.
    \param st Pointer to a string of characters to split.
    \param delimitor Pointer to a string of characters that defines the delimitor used to separate fieds.
    \param f Pointer to a structure 'fields' that contains the fields extracted from the string 'st'.
    \return The function returns the number of fields extracted. If an error occurs, the the function returns -1.
    \warning <UL>
               <LI>The structure 'f' contains elements that have been allocated within the function. So you must
                   free these elements when you don't need them any more. To do this, you should use the function
                   'free_fields()'.
               <LI>For this function, 2 (or more) delimitors in a row are seen as one and only one delimitor.
                   Example: the string &quot;A;;;B&quot; has 2 fields (&quot;A&quot; and &quot;B&quot;), not 4
                   (assuming that the delimitor is the character &quot;;&quot;).
             </UL>
 */

int s_split(char *st, char* delimitor, struct fields *f)
{
  char*    pt;
  int      nb_field, field_size, i, size;
  char**   tab;   

  /***************************************************************/
  /* Allocate initial number of fields - INITIAL_NUMBER_OF_FIELD */
  /***************************************************************/

  size = INITIAL_NUMBER_OF_FIELD;
  tab = (char**)malloc(size*sizeof(char*));
  if (tab == NULL)
  {
    f->number_of_fields = -1;
    f->size             = -1;
    f->tabs             = NULL;
    return -1;
  }

  nb_field = 0; 
  pt       = strtok(st, delimitor);
  
  while (pt != NULL)
  {
    if (nb_field == size)
    {
      size *= 2;
      tab = (char**)realloc((void*)tab, size*sizeof(char*));
      if (tab == NULL)
      {
        for (i=0; i<nb_field; i++) { free ((void*)(tab[i])); }
        free ((void*)tab);
        f->number_of_fields = -1;
        f->size             = -1;
        f->tabs             = NULL;
        return -1;
      }
    } 

    field_size = strlen(pt)+1;
    tab[nb_field] = (char*)malloc(field_size*sizeof(char));
    if (tab[nb_field] == NULL)
    {
      for (i=0; i<nb_field; i++) { free ((void*)(tab[i])); }
      free ((void*)tab);
      f->number_of_fields = -1;
      f->size             = -1;
      f->tabs             = NULL;
      return -1;
    }

    strcpy (tab[nb_field], pt);
    nb_field++;

    pt = strtok (NULL, delimitor);
  }
  
  f->number_of_fields = nb_field;
  f->size             = size;
  f->tabs             = tab;
  
  return nb_field; 
}

/*! \example test_s_split.c
    This file shows how to use the function s_split().
 */


/*! \brief Split a string into fields.
    \param st Pointer to a string of characters to split.
    \param delimitor Pointer to a string of characters that defines the delimitor used to separate fieds.
    \param f Pointer to a structure 'fields' that contains the fields extracted from the string 'st'.
    \return The function returns the number of fields extracted. If an error occurs, the the function returns -1.
    \warning <UL>
               <LI>The structure 'f' contains elements that have been allocated within the function. So you must
                   free these elements when you don't need them any more. To do this, you should use the function
                   'free_fields()'.
               <LI>This function does NOT behave like the function s_split().
                   Example: the string &quot;A;;;B&quot; has 4 fields (&quot;A&quot;, &quot;&quot;, &quot;&quot;
                   and &quot;B&quot;), not 2 (assuming that the
                   delimitor is the character &quot;;&quot;).
             </UL>
 */

int s_split_exact(char *st, char* delimitor, struct fields *f)
{
  char     *pt, *start, *stop;
  int      nb_field, field_size, i, j, size, pfields;
  char     **tab, **delimitors, **mem;

  /***************************************************************/
  /* Set fields' structure to error values                       */
  /***************************************************************/

  f->number_of_fields = -1;
  f->size             = -1;
  f->tabs             = NULL;

  /***************************************************************/
  /* Allocate initial number of fields - INITIAL_NUMBER_OF_FIELD */
  /***************************************************************/

  size = INITIAL_NUMBER_OF_FIELD;

  delimitors = (char**)malloc(size*sizeof(char*));
  if (delimitors == NULL) { return -1; }

  /***************************************************************/
  /* Search the string for all delimitors                        */
  /***************************************************************/

  nb_field = 0;
  start    = st;

  pt = strstr (start, delimitor);
  while (pt != NULL)
  {
    /*************************************************************/
    /* Re-allocate memory if necessary                           */
    /*************************************************************/

    if (nb_field == size)
    {
      size *= 2;
      mem = (char**)realloc((void*)delimitors, size * sizeof(char*));
      if (mem == NULL) { free(delimitors); return -1; }
      delimitors = mem;
    }

    /*************************************************************/
    /* Save delimitor's position                                 */
    /*************************************************************/

    delimitors[nb_field] = pt;
    nb_field++;

    start = pt + strlen(delimitor);

    pt = strstr (start, delimitor);
  }

  /***************************************************************/
  /* note: "A;B;"   => 2 fields                                  */
  /*       "A;B"    => 2 fields                                  */
  /***************************************************************/

  if (*start != 0)
  {
    if (nb_field == size)
    {
      size *= 2;
      mem = (char**)realloc((void*)delimitors, size * sizeof(char*));
      if (mem == NULL) { free(delimitors); return -1; }
      delimitors = mem;
    }

    delimitors[nb_field] = st + strlen(st);

    nb_field++;
  }

  /***************************************************************/
  /* Allocate fields' structure                                  */
  /***************************************************************/

  tab = (char**)malloc(nb_field * sizeof(char*));
  if (tab == NULL)
  {
    free (delimitors);
    return -1;
  }

  f->number_of_fields = nb_field;
  f->size             = nb_field;
  f->tabs             = tab;

  /***************************************************************/
  /* Now extract fields and store them into fields' structure    */
  /***************************************************************/

  pfields = 0;
  start   = st;
  for (j=0; j<nb_field; j++)
  {
    stop       = delimitors[j];
    field_size = (int)(stop-start);

    pt         = (char*) malloc ((field_size+1) * sizeof(char));
    if (pt == NULL)
    {
      f->number_of_fields = -1;
      f->size             = -1;
      free (delimitors);
      for (i=0; i<pfields; i++) { free ((void*)tab[i]); }
      return -1;
    } 

    strncpy (pt, start, field_size);
    pt[field_size] = 0;

    tab[pfields] = pt;
    pfields += 1;

    start = stop + strlen(delimitor);
  }


  free (delimitors);
  return nb_field;
}

/*! \example test_s_split_exact.c
    This file shows how to use the function s_split_exact().
 */

/*! \brief Free allocated memory used to put fields extracted from a string of characters.
    \param fd Pointer to a structure 'fields'.
    \warning You should call this function after a call to s_split(). Please note that if s_split() failes,
             then you don't have to free the resulting structure 'fields'. But you can do it, it won't harm.
 */

void free_fields (struct fields *fd)
{
  int i;

  for (i=0; i<fd->number_of_fields; i++) { free ((void*)((fd->tabs)[i])); }
  free (fd->tabs);
  fd->number_of_fields = -1;
  fd->size             = -1;
  fd->tabs             = NULL;
}

/*! \brief Extract everything between the string " = " and the end of the string.
    \param src Pointer to the string to parse.
    \param dst Pointer to a memory location used to put the extracted string.
    \param nbcar Maximum number of characters the function can put into 'dst'.
    \return The function return a pointer to 'dst', or NULL if an error occured.
 */

char* get_end_of_string (char *src, char *dst, int nbcar)
{
  char *pt;
  int  size;
 
  if (nbcar <= 0) { return NULL; }  /* be paranoid */
 
  pt = strstr (src, " = ");
  if (pt == NULL) { dst[0]=0; return dst; }
  pt += 3;

  size = strlen(pt);
  if (nbcar < size+1) { dst[0]=0; return NULL; }

  strcpy (dst, pt);

  return dst;
}

/*! \example test_get_end_of_string.c
    This file shows how to use the function get_end_of_string().
 */

/*! \brief Transform a IP address "XXX.XXX.XXX.XXX" into a hexadecimal representation "HHHHHHHH".
    \param src Pointer to a string of characters that contains an IP address written as "XXX.XXX.XXX.XXX".
    \param dst Pointer to a memory location used to put the resulting IP address in hexadecimal notation.
    \param nbcar Maximum number of characters the function can put into 'dst'.
    \return The function returns a pointer to 'dst', or NULL if an error occured.
 */

char* IP2hex (char *src, char *dst, int nbcar)
{
  char          *pt;
  struct fields fd;
  int           n;
  int           ip1, ip2, ip3, ip4;
  
  if (nbcar < 12) { return NULL; }  /* be paranoid */
  memset ((void*)dst, 0, 12);

  pt = (char*)malloc((strlen(src)+1)*sizeof(char));
  if (pt == NULL) { return NULL; }
  strcpy (pt, src);

  n = s_split(pt, ".", &fd);
  if (n != 4)
  {
    free (pt);
    return NULL;
  }
  free (pt);

  ip1 = atoi (fd.tabs[0]);
  ip2 = atoi (fd.tabs[1]);
  ip3 = atoi (fd.tabs[2]);
  ip4 = atoi (fd.tabs[3]);

  if ((ip1 < 0) || (ip1 > 255)) { free_fields(&fd); return NULL; } 
  if ((ip2 < 0) || (ip2 > 255)) { free_fields(&fd); return NULL; } 
  if ((ip3 < 0) || (ip3 > 255)) { free_fields(&fd); return NULL; } 
  if ((ip4 < 0) || (ip4 > 255)) { free_fields(&fd); return NULL; } 

  sprintf (dst, "%s%s%s%s", bin2hex_hexa[ip1], bin2hex_hexa[ip2], bin2hex_hexa[ip3], bin2hex_hexa[ip4]);
 
  free_fields(&fd);
  return dst;
}

/*! \example test_IP2hex.c
    This file shows how to ude the function IP2hex().
 */


/*! \brief Delete spaces (or tabulations) at the end of a string. 
    \param src pointer to the string to parser.
    \return The function returns a pointer to 'src'.
    \warning The string 'src' is modified.
 */

char* post_chop(char *src)
{
  int  last_idx;
  int  i;

  last_idx = strlen(src);
  if (last_idx == 0) { return src; }

  last_idx--;
  for (i=last_idx; i>=0; i--)
  { 
     if ((src[i]==' ') || (src[i]==9)) { src[i]=0; }
     else { return src; }
  }
  return src;
}

/*! \brief Delete spaces (or tabulations) at the begining of a string.
    \param src Pointer to a string of of characters that contains the string to parser.
    \return value the function returns a pointer to 'src', or NULL if an error occured.
    \warning the string 'src' is modified.
 */

char* pre_chop(char *src)
{
  int  size;
  int  i, passed, count;
  char *aux;

  size = strlen(src);
  if (size == 0) { return src; }

  aux = (char*)malloc((size+1)*sizeof(char));
  if (aux == NULL) { return NULL; }

  passed = 0;
  count  = 0;
  for (i=0; i<size; i++)
  {
     if (passed == 0)
     {
       if ((src[i] != ' ') && (src[i] != 9))
       { aux[count++]=src[i]; passed = 1; }
     }
     else { aux[count++]=src[i]; }
  }
  aux[count]=0;
  strcpy (src, aux);
  free (aux);
  return src;
}

/*! \brief Delete spaces (or tabulations) at the begining and at the end of a string.
    \param src Pointer to the string to parse.
    \return The function returns a pointer to 'src', or NULL if an error occured.
    \warning The string 'src' is modified.
 */


char* chop(char *src)
{ return pre_chop(post_chop(src)); }

/*! \example test_chop.c
    This file shows how to use the function chop().
 */

/*! \brief Replace all charater 'new line' (that is '\n'), by the sequence '\n#'.
    \param src Pointer to the string to convert into comment.
    \param dst Pointer to a pointer that will be used to store the resulting string.
    \return The function returns a pointer to the resulting string, or NULL if an error occured.
    \warning The returned pointer has been allocated using 'malloc()'. So you must free the allocated memory yourself
             (using 'free()').
 */ 

char* add_comment (char *src, char **dst)
{
  int   src_size, i, j, nnline, dst_size;
  char  c;


  src_size = strlen(src);

  /* Count the number of new lines */
  nnline = 0;
  for (i=0; i<src_size; i++) { if (src[i] == '\n') { nnline++; } }

  /* Allocate memory for the destination */
  dst_size = src_size + nnline;
  *dst = (char*)malloc((dst_size+1)*sizeof(char));
  if (*dst == NULL) { return NULL; }
 
  /* Replace all '\n' by "\n#" */
  j = 0;
  for (i=0; i<src_size; i++)
  {
    c = src[i];
    (*dst)[j++] = c;
    if (c == '\n') { (*dst)[j++] = '#'; }
  }

  (*dst)[j] = 0; 

  return *dst;
}

/*! \example test_add_comment.c
    This file shows how to use the function add_comment().
 */

/*! \brief Extract a field's value from a configuration file.
    \param buffer Pointer to a NULL terminated string of characters that contains the line extracted from the configuration
           file.
    \param delim Pointer to a NULL terminated string of characters that defines the delimitor between the field' label and
           the field's value.
    \return The function returns a pointer to the field's value. If an error occured, the function returns NULL.
 */


char* get_config_value (char *buffer, char *delim)
{
  char *res, *pt;

  pt = strstr (buffer, delim);
  if (pt == NULL) { return NULL; }
  pt += 1;

  res = (char*)malloc((strlen(pt)+1)*sizeof(char));
  if (res == NULL) { return NULL; }
  strcpy (res, pt);

  if (chop(res) == NULL) {
                           free(res);
                           return NULL;
                         };
  return res;
}

/*! \brief The maximum number of characters of an interger (in decimal). For example the value "134" takes 3 characters.
 */

#define INTEGER_STR_MAX_SIZE 32

/*! \brief Replace a tag by an integer in a string.
    \param src Pointer to NULL terminated string of characters that contains the tag to replace.
    \param tag Tag to replace.
    \param value Integer value that will replace the tag.
    \return The function returns a pointer to the resulting string of characters. If an error occured, then the function
            returns a NULL pointer (this means that the program is running out of memory).
 */

char* replace_tag_by_integer (char* src, char *tag, int value)
{
  char *res, *start, *stop, c;
  int  tag_size, size;
  char buff[INTEGER_STR_MAX_SIZE];

  /* Look for the tag to replace */
  start = strstr(src, tag);
  if (start == NULL)
  {
    size = strlen(src)+1;
    res = (char*)malloc(size * sizeof(char)); 
    if (res == NULL) { return NULL; }
    strcpy (res, src);
    return res;
  }

  tag_size = strlen(tag);
  stop = start + tag_size;

  sprintf (buff, "%d", value);
  size = (strlen(src) - tag_size) + strlen(buff) + 1;

  res = (char*)malloc(size * sizeof(char));
  if (res == NULL) { return NULL; }
  c = *start;
  *start = 0;
  sprintf (res, "%s%s%s", src, buff, stop);
  *start = c;

  return res;
}

/*! \brief The array is used to make the relation between characters and intergers ('0'->0, '1'->1, ...).
 */

extern int char2int[256];

/*! \brief The array is used to make the relation between characters and hexadecimal ('0'->0, '1'->1, ..., 'a'->10, ...)
 */

extern int char2hex[256];




/*! \brief Convert a string of characters into an 'unsigned int'.
    \param st Pointer to a NULL terminated string of characters that will be converted into "u_int'.
    \param error Pointer to an interger used to signal an error.
    \return The function returns the value represented bu the string 'st' (if sucess, then 'error' is equal to 0). Please
            note that if an error occures, then the returned value is equal to -1 (all in binary - in this case 'error'
            is equal to 1).
 */

unsigned int string2unsigned_int (char *st, int *error)
{
  int           last_idx, i, digit;
  unsigned int  value, factor;


  value    = 0;
  factor   = 1;
  *error   = 0;

  last_idx = strlen(st)-1;
  for (i=last_idx; i>=0; i--)
  {
    digit = char2int[(unsigned int)st[i]];
    if (digit == -1) { *error = 1; return (unsigned int)-1; }
    value += (unsigned int)digit*factor;
    factor *= 10;
  }

  return value;
}


/*! \brief Convert a string of characters into a 'time_t'.
    \param st Pointer to a NULL terminated string of characters that will be converted into "time_t".
    \param error Pointer to an interger used to signal an error.
    \return The function returns the value represented bu the string 'st' (if sucess, then 'error' is equal to 0). Please
            note that if an error occures, then the returned value is equal to -1 (all in binary - in this case 'error'
            is equal to 1).
 */

time_t string2time_t (char *st, int *error)
{
  int     last_idx, i, digit;
  time_t  value, factor;


  value    = 0;
  factor   = 1;
  *error   = 0;

  last_idx = strlen(st)-1;
  for (i=last_idx; i>=0; i--)
  {
    digit = char2int[(unsigned int)st[i]];
    if (digit == -1) { *error = 1; return (unsigned int)-1; }
    value += (unsigned int)digit*factor;
    factor *= 10;
  }

  return value;
}

/*! \brief Convert a string of characters into an 'int'.
    \param st Pointer to a NULL terminated string of characters that will be converted into "int'.
    \param error Pointer to an interger used to signal an error.
    \return The function returns the value represented bu the string 'st' (if sucess, then 'error' is equal to 0). Please
            note that if an error occures, then the returned value is equal to -1 (in this case 'error' is equal to 1).
 */

int string2int (char *st, int *error)
{
  int    last_idx, i, digit;
  int    value, factor;


  value    = 0;
  factor   = 1;
  *error   = 0;

  last_idx = strlen(st)-1;
  for (i=last_idx; i>=0; i--)
  {
    digit = char2int[(unsigned int)st[i]];
    if (digit == -1) { *error = 1; return -1; }
    value += digit*factor;
    factor *= 10;
  }

  return value;
}

/*! \example test_string2int.c
    This file shows how to use the function string2int().
 */


/*! \brief Test if a string of characters represents an interger.
    \param st Pointer to a NULL terminated string of characters that should represent an integer.
    \return The function returns 0 if the string 'st' represents an interger. Otherwise, it returns 1.
 */

int is_integer (char *st)
{
  int i, size, digit;

  if (st == NULL) { return 1; }
  if (st[0] == 0) { return 1; } 

  size = strlen (st);

  for (i=0; i<size; i++)
  {
    digit = char2int[(unsigned int)st[i]];
    if (digit == -1) { return 1; }
  }

  return 0;
}

/*! \example test_is_integer.c
    This file shows how to use the function is_integer().
 */

/*! \brief Test if a string of characters represents a hexadecimal.
    \param st Pointer to a NULL terminated string of characters that should represent a hexadecimal.
    \return The function returns 0 if the string 'st' represents an interger. Otherwise, it returns 1.
 */

int is_hexa (char *st)
{
  int i, size, digit;

  if (st == NULL) { return 1; }
  if (st[0] == 0) { return 1; }

  size = strlen (st);

  for (i=0; i<size; i++)
  {
    digit = char2hex[(unsigned int)st[i]];
    if (digit == -1) { return 1; }
  }

  return 0;
}

/*! \example test_is_hexa.c
    This file shows how to use the function is_hexa().
 */

/*! \brief Test if a string of characters represents an IP address.
    \param st Pointer to a zero terminated string of characters that might represents an IP address.
    \return The function return 0 is 'st' represents an IP address. Otherwise, the function returns 1.
    \warning An error code 1 could also mean &quot;Out of memory&quot;. But it should not happen.
 */

int is_ip_address (char *st)
{
  struct fields fd;
  int           n, i, err;
  char          *aux; 

  if (st == NULL) { return 1; }
  if (st[0] == 0) { return 1; }

  aux = (char*)malloc((strlen(st)+1)*sizeof(char));
  if (aux == NULL) { return 1; }
  strcpy (aux, st);

  n = s_split(aux, ".", &fd);
  if (n == -1)
  { 
    free(aux);
    return 1;
  }

  if (fd.number_of_fields != 4)
  {
    free (aux);
    free_fields(&fd);
    return 1;
  }  

  for (i=0; i<fd.number_of_fields; i++)
  {
    n = string2int(fd.tabs[i], &err);
    if (err == 1)
    {
      free (aux);
      free_fields(&fd);
      return 1;
    }
    if ((n<0) || (n>255))
    {
      free (aux);
      free_fields(&fd);
      return 1;
    }
  }

  free(aux);
  free_fields(&fd);
  return 0;
}

/*! \example test_is_ip_address.c
    This file shows how to use the function is_ip_address()
 */


/*! \brief Convert an IPV4 address to its numerical equivalent.
    \param st Pointer to a zero terminated string of characters that represents the IP address (X.X.X.X).
    \param num Pointer to a array of - at least - 4 unsigned char.
    \return The function returns 0 if the operation was successfull, 1 otherwise.
 */

int ip_address_to_num (char *st, unsigned char *num)
{
  struct fields fd;
  int           n;


  if (st == NULL) { return 1; }
  if (st[0] == 0) { return 1; }

  if (is_ip_address(st) == 1) { return 1; }

  /********************************************/
  /* Split the IP address                     */
  /********************************************/

  n = s_split(st, ".", &fd);
  if (n == -1) { return 1; }

  if (fd.number_of_fields != 4)
  {
    free_fields(&fd);
    return 1;
  }

  /********************************************/
  /* First byte X.x.x.x                       */
  /********************************************/

  num[0] = (unsigned char) (atoi(fd.tabs[0]));

  /********************************************/
  /* Second byte x.X.x.x                      */
  /********************************************/

  num[1] = (unsigned char) (atoi(fd.tabs[1]));

  /********************************************/
  /* Third byte x.x.X.x                       */
  /********************************************/

  num[2] = (unsigned char) (atoi(fd.tabs[2]));

  /********************************************/
  /* Fourth byte x.x.x.X                      */
  /********************************************/

  num[3] = (unsigned char) (atoi(fd.tabs[3]));

  free_fields(&fd);
  return 0;
}

/*! \example test_ip_address_to_num.c
    This file shows how to use the function ip_address_to_num().
 */

/*! \brief Convert a string of ASCII bytes into hexa representation.
    \param src Pointer to a zero terminated string of characters to convert.
    \param dst Pointer to a memory buffer used to store the converted string.
    \param dst_size Size, in bytes, of the destination buffer.
    \return Upon successful completion the function returns the value 1 (the destination buffer is not big
            enough). Otherwize, the function returns the value 0.
 */

int bin2hex (char *src, char *dst, int dst_size)
{
  int i;

  if ((strlen(src) * 2) >= dst_size) { return 0; } 

  memset ((void*)dst, 0, dst_size);

  for (i=0; i<strlen(src); i++)
  {
    unsigned int c;
    c = (unsigned int)src[i];
    strcat (dst, bin2hex_hexa[c]); 
  }

  return 1;
}

/*! \brief Convert a binary buffer into hexa representation.
    \param src Pointer to a list of bytes to convert.
    \param dst Pointer to a memory buffer used to store the converted string.
    \param src_size Size, in bytes, of the source buffer.
    \param dst_size Size, in bytes, of the destination buffer.
    \return Upon successful completion the function returns the value 1 (the destination buffer is not big
            enough). Otherwize, the function returns the value 0.
 */

int bin2hex_binary (unsigned char *src, unsigned char *dst, int src_size, int dst_size)
{
  int i;

  if ((src_size * 2) >= dst_size) { return 0; }

  memset ((void*)dst, 0, dst_size);

  for (i=0; i<src_size; i++)
  {
    unsigned int c;
    c = (unsigned int)src[i];
    strcat (dst, bin2hex_hexa[c]);
  }

  return 1;
}






