/*! \file timing.c
    This file contains functions for time management.
 */

#include "timing.h"
#include <stdio.h>

/*! \brief Static variable used by start_chrono() to save the starting timestamp.
 */
static struct timeval   start;

/*! \brief Static variable used by stop_chrono() to save the stoping timestamp.
 */
static struct timeval   stop;

/*! \brief Start the stopwatch.
    \return Upon successful completion, the function returns 0. Otherwise the function returns -1;
 */
int start_chrono() { return (gettimeofday(&start, NULL)); }

/*! \example test_stopwatch.c
    This example shows how to use the function start_chrono()
 */

/*! \brief Stop the  stopwatch.
    \return Upon successful completion, the function returns 0. Otherwise the function returns -1;
 */
int stop_chrono() { return (gettimeofday(&stop, NULL)); } 

/*! \example test_stopwatch.c
    This example shows how to use the function stop_chrono()
 */

/*! \brief Get the measured time in ms.
    \return The function returns the number of mili seceond.
 */
unsigned long int get_ms()
{
  long   t;

  t = (stop.tv_sec*1000 + stop.tv_usec/1000) - (start.tv_sec*1000 + start.tv_usec/1000);

  return (unsigned long int)t;
}

/*! \example test_stopwatch.c
    This example shows how to use the function get_ms()
 */  

/*! \brief This structure is used to specify the timer's delay (for a one shot timer).
 */
static struct itimerval   one_shot_timer;

/*! \brief This variable is used by the system call setitimer(). This system call requires a structure itimerval to save
           the current timer configuration (it is not used here).
 */
static struct itimerval   old_one_shot_timer;

/*! \brief This variable is used by the functions pause_one_shot_timer() and restart_one_shot_timer(). It is used to save the current
           timer's state (i.e: the remaining time).
 */
static struct itimerval   current_one_shot_timer;


/*! \brief Create a one shot timer which precision is the micro second. The timer raises the signal SIGALRM when the timout
           expired.
    \param sec Number of second
    \param micro_second Number of micro seconds.
    \return The function returns one of the following values:
            <UL>
                <li>SET_ONE_SHOT_TIMER_ERROR: An error occured while creating the timer.
                <li>SET_ONE_SHOT_TIMER_OK: The timer is successfuly created.
            </UL>
 */

int set_one_shot_timer (long sec, long micro_second)
{
  one_shot_timer.it_interval.tv_sec   = 0;
  one_shot_timer.it_interval.tv_usec  = 0;
  one_shot_timer.it_value.tv_sec      = sec;
  one_shot_timer.it_value.tv_usec     = micro_second;
  if (setitimer (ITIMER_REAL, &one_shot_timer, &old_one_shot_timer) == -1)
  { return SET_ONE_SHOT_TIMER_ERROR; }

  return SET_ONE_SHOT_TIMER_OK; 
}

/*! \example test_one_shot_timer.c
    This file shows how to use the function set_one_shot_timer().
 */

/*! \brief Cancel the previously activated one shot timer.
        \return The function returns one of the following values:
            <UL>
                <li>CANCEL_ONE_SHOT_TIMER_ERROR: An error occured while canceling the timer.
                <li>CANCEL_ONE_SHOT_TIMER_OK: The timer is successfuly canceled.
            </UL>
 */

int cancel_one_shot_timer ()
{
  one_shot_timer.it_interval.tv_sec   = 0;
  one_shot_timer.it_interval.tv_usec  = 0;
  one_shot_timer.it_value.tv_sec      = 0;
  one_shot_timer.it_value.tv_usec     = 0;
  if (setitimer (ITIMER_REAL, &one_shot_timer, &old_one_shot_timer) == -1)
  { return CANCEL_ONE_SHOT_TIMER_ERROR; }

  return CANCEL_ONE_SHOT_TIMER_OK; 
}

/*! \example test_one_shot_timer.c
    This file shows how to use the function cancel_one_shot_timer().
 */

/*! \brief Pause the one shot timer.
    \return The function return one of the following values:
            <UL>
              <li>PAUSE_ONE_SHOT_TIMER_ERROR: An error occured with the high precision timer.
              <li>PAUSE_ONE_SHOT_TIMER_OK: The operation was successful.
            </UL>
    \warning You can restart the timer with restart_one_shot_timer().
 */

int pause_one_shot_timer ()
{
  if (getitimer (ITIMER_REAL, &current_one_shot_timer) == -1)
  { return PAUSE_ONE_SHOT_TIMER_ERROR; }

  if (cancel_one_shot_timer() != CANCEL_ONE_SHOT_TIMER_OK)
  { return PAUSE_ONE_SHOT_TIMER_ERROR; }

  return PAUSE_ONE_SHOT_TIMER_OK;
}

/*! \example test_one_shot_timer.c
    This file shows how to use the function pause_one_shot_timer().
 */

/*! \brief Restart the one shot timer.
    \return The function return one of the following values:
            <UL>
              <li>RESTART_ONE_SHOT_TIMER_ERROR: An error occured with the high precision timer.
              <li>RESTART_ONE_SHOT_TIMER_OK: The operation was successful.
            </UL>
    \warning The timer should have been paused with the function pause_one_shot_timer().
 */

int restart_one_shot_timer ()
{
  if (setitimer (ITIMER_REAL, &current_one_shot_timer, &old_one_shot_timer) == -1)
  { return RESTART_ONE_SHOT_TIMER_ERROR; }

  return RESTART_ONE_SHOT_TIMER_OK;
}

/*! \example test_one_shot_timer.c
    This file shows how to use the function restart_one_shot_timer().
 */



